/* av3d.c
 * 3d avatar viewer, to be linked with s3dt.y or inventor.y for different
 * type of 3d graphics
 */
#include "sim.h"
#include "client.h"
#include "graphics.h"
#include "3d.h"
#include "jx/jx.h"

static void print_usage(char *av0);
static int init_av_ui(void);
static void run_av_ui(double xcen, double ycen, double zcen);
static void clean_up(int s);

int fd_debug		= 0;
int is_sim_server	= 0;
int parse_debug         = 0;

int main(int argc, char **argv)
{
int i;
char *avsource;		/* file or server */
int avport = SIM_PORT;
int fd = -1, cppfd;	/* file descriptors */
userprefs cp;
double cx,cy,cz;	/* point to rotate about */
int gotcentre = 0;	/* centre point given in command line args? */
int rv;

/* Parse command line args
 */
get_prefs(&cp);
for(i=1; i<argc; i++) {
	if (strequ(argv[i],"-3d") && i < argc-1) {
		cp.default_3d = atoi(argv[++i]);
		if (cp.default_3d < 0 || cp.default_3d > 2)
			print_usage(argv[0]);
		}
	else if (strequ(argv[i],"-parse_debug"))
		parse_debug = 1;
	else if (strequ(argv[i],"-help"))
		print_usage(argv[0]);
	else if (strequ(argv[i],"-port") && i < argc-1) {
		avport = atoi(argv[++i]);
		}
	else if (strequ(argv[i],"-centre") && i < argc-3) {
		cx = atof(argv[++i]);
		cy = atof(argv[++i]);
		cz = atof(argv[++i]);
		gotcentre = 1;
		}
	else
		avsource = argv[i];
	}
if (!avsource)
	print_usage(argv[0]);
put_prefs(&cp);

/* Create main window
 */
if (!init_av_ui()) {
	logerror("Error initializing user interface");
	exit(2);
	}

/* Open and parse avatar shape
 */
if ((fd = open(avsource,O_RDONLY)) != -1) {
	/* Read from file
	 */
	cppfd = pipethrough(fd,NULL,CPPCMD);
	rv = parse_3ds(cppfd);
	pipeclose(cppfd);
	close(fd);
	}
else if (strequ(avsource,"-")) {
	/* Read from stdin
	 */
	fd = dup(0);
	cppfd = pipethrough(fd,NULL,CPPCMD);
	rv = parse_3ds(cppfd);
	pipeclose(cppfd);
	close(fd);
	}
else if ((fd = open_tcp_send(avsource,avport)) != -1) {
	/* Read from server
	 */
	if (!send_version(fd)) {
		logerror("Version mismatch with server");
		exit(3);
		}
	fdprintf(fd,"%d\n",con_getshapes);
	fdprintf(fd,"%d\n",supported_3d());
	rv = parse_3ds(fd);
	close(fd);
	}
else {
	logerror("Could find file or server %s",avsource);
	exit(3);
	}

if (rv) {
	logerror("Error parsing shape data");
	exit(3);
	}

signal(SIGINT,clean_up);
signal(SIGTERM,clean_up);

/* Compute scene midpoint
 */
if (!gotcentre) {
	double xmin,ymin,zmin;	/* scene extents */
	double xmax,ymax,zmax;

	boundingbox_3d(&xmin,&ymin,&zmin,&xmax,&ymax,&zmax);
	cx = (xmax+xmin)/2;
	cy = (ymax+ymin)/2;
	cz = (zmax+zmin)/2;
	}
run_av_ui(cx,cy,cz);
}


/* print_usage
 * Display correct command line args and exit
 */
static void print_usage(char *av0)
{
printf("usage : %s <[file] | [server]> [-port num] [-3d 0,1,2] [-parse_debug]\n",av0);
exit(0);
}


static Display *display;	/* connection to X server */
static int screen_num;		/* screen number windows are on */
static Window root_win;		/* root window of this screen */
static Window avwin;		/* main window for gadgets */
static int width=400,height=400;/* size of main window */
static long black,white;	/* colours for gadgets */
static long grey,red;

static JXgadget *horiz,*vert;	/* rotation scroll bars */
static JXgadget *zoom;		/* zoom factor scroll bar */
static JXgadget *quit;		/* quit button */ 

/* init_av_ui
 * Create user interface and 3d view window
 */
static int init_av_ui(void)
{
XSizeHints hints;

if ((display = XOpenDisplay(NULL)) == NULL) {
	logerror("Couldn't open display %s",XDisplayName(NULL));
	return 0;
	}
screen_num = DefaultScreen(display);
root_win   = RootWindow(display,screen_num);

/* Create top level window
 */
avwin = XCreateSimpleWindow(display,root_win,100,100,width,height,0,0,0);
hints.flags = PMinSize;
hints.min_width  = 200;
hints.min_height = 200;
XSetStandardProperties(display,avwin,"av3d","av3d",None,NULL,0,&hints);
XSelectInput(display,avwin,StructureNotifyMask | KeyPressMask);
XMapWindow(display,avwin);
XFlush(display);

init_3d(NULL,avwin,0,0,width-20,height-40);

black = best_colour(0,0,0);
white = best_colour(255,255,255);
grey  = best_colour(100,100,100);
red   = best_colour(255,0,0);

/* Create gadgets in main window
 */
JXStartup(display,screen_num);
horiz   = JXNewGadget(display,avwin,white,black,grey,scroller,0,height-40,
		      width,20,0.0);
JXSetOption(display,horiz,WrapSlider,1,ScrollStep,0.05,SmoothMove,1,0);
vert    = JXNewGadget(display,avwin,white,black,grey,scroller,width-20,0,
		      20,height-40,0.5);
JXSetOption(display,vert,ScrollStep,0.05,SmoothMove,1,0);
zoom    = JXNewGadget(display,avwin,white,black,grey,scroller,0,height-20,
		      width-100,20,0.5);
JXSetOption(display,zoom,ScrollStep,0.05,SmoothMove,1,0);
quit    = JXNewGadget(display,avwin,white,black,grey,boolean,width-100,
		      height-20,100,20,"Quit");
JXSetOptionAll(display,HighlightCol,red,0);
return 1;
}


/* run_av_ui
 * Endlessly loop polling for events and rendering the 3d view
 */
static void run_av_ui(double xcen, double ycen, double zcen)
{
need_redraw();
while(1) {
	double ang,tilt,range;
	plview vw;
	fd_set readfds;
	int dfd = ConnectionNumber(display);
	int tfd = threed_fd();

	FD_ZERO(&readfds);
	FD_SET(dfd,&readfds);
	FD_SET(tfd,&readfds);
	select(max(dfd,tfd)+1,&readfds,NULL,NULL,NULL);

	if (FD_ISSET(dfd,&readfds)) {
		while(XPending(display)) {
			XEvent event;

			XNextEvent(display,&event);
			if (!JXGadgetEvent(display,&event)) {
				switch(event.type) {
				case ConfigureNotify:
					if (event.xconfigure.width == width &&
					    event.xconfigure.height == height)
						continue;	/* only moved */
					/* Window resized
					 */
					width  = event.xconfigure.width;
					height = event.xconfigure.height;
					JXMoveGadget(display,horiz,0,height-40,
						     width,20);
					JXMoveGadget(display,vert,width-20,0,
						     20,height-40);
					JXMoveGadget(display,zoom,0,height-20,
						     width-100,20);
					JXMoveGadget(display,quit,width-100,
						     height-20,100,20);
					resize_3d(0,0,width-20,height-40);
					need_redraw();
					break;
				case ClientMessage:
					/* Gadget moved / pressed
					 */
					if (JXGMatch(quit,&event)) {
						shutdown_3d();
						exit(0);
						}
					need_redraw();
					break;
					}
				}
			}

		/* Redraw the 3d view window
		 */
		ang   = NPosFromScroller(horiz)*360;
		tilt  = NPosFromScroller(vert)*178 - 89;
		range = sqr(NPosFromScroller(zoom)*7);
		vw.ang = fixang(ang + 180);
		vw.tilt = -tilt;
		vw.x2 = xcen + cos(dtor(ang))*cos(dtor(tilt))*range;
		vw.y2 = ycen + sin(dtor(ang))*cos(dtor(tilt))*range;
		vw.z2 = zcen + sin(dtor(tilt))*range;
		vw.range = vw.dst = 0;
		render_3d(&vw);
		}

	if (FD_ISSET(tfd,&readfds)) {
		/* Let the 3d view window check for events
		 */
		check_3d();
		}
	}
}


/* clean_up
 * Clean up on ctrl-c
 */
static void clean_up(int s)
{
shutdown_3d();
exit(1);
}

