/* client.c
 *
 * Various functions used only by clients
 */
#include "sim.h"
#include "client.h"

extern int parse_debug;
extern int fd_debug;

static void usage_error(char *av0);

/* parse_client_args
 * Parse and check command line arguments for a sim client
 */
int parse_client_args(int argc, char **argv, char **name)
{
userprefs cp;
userservprefs csp;
int i;

get_prefs(&cp);
get_sprefs(&csp);
for(i=1; i<argc; i++) {
	if (strequ(argv[i],"-name") && i<argc-1) {
		*name = argv[++i];
		}
	else if (strequ(argv[i],"-parse_debug")) {
		parse_debug = 1;
		}
	else if (strequ(argv[i],"-fd_debug")) {
		fd_debug = 1;
		}
	else if ((strequ(argv[i],"-3d") || strequ(argv[i],"-default_3d"))
		 && i<argc-1) {
		cp.default_3d = atoi(argv[++i]);
		if (cp.default_3d < 0 || cp.default_3d > 2) {
			logerror("Invalid 3d level");
			usage_error(argv[0]); return 0;
			}
		}
	else if (strequ(argv[i],"-recved_data") && i<argc-1) {
		cp.recved_data = atoi(argv[++i]);
		if (cp.recved_data < 0 || cp.recved_data > 2) {
			logerror("Invalid received-data preference");
			usage_error(argv[0]); return 0;
			}
		}
	else if (strequ(argv[i],"-move_speed") && i<argc-1) {
		cp.move_speed = atof(argv[++i]);
		if (cp.move_speed < 0.0 || cp.move_speed > 5.0) {
			logerror("Invalid move speed");
			usage_error(argv[0]); return 0;
			}
		}
	else if (strequ(argv[i],"-password") && i<argc-1) {
		strcpy(cp.password,argv[++i]);
		}
	else if (strequ(argv[i],"-size") && i<argc-2) {
		cp.w = atoi(argv[++i]);
		cp.h = atoi(argv[++i]);
		if (cp.w < 400 || cp.h < 400) {
			logerror("Window size is too small");
			usage_error(argv[0]); return 0;
			}
		}
	else if (strequ(argv[i],"-position") && i<argc-2) {
		cp.x = atoi(argv[++i]);
		cp.y = atoi(argv[++i]);
		}
	else if (strequ(argv[i],"-port") && i<argc-1) {
		csp.port = atoi(argv[++i]);
		if (!csp.port) {
			logerror("Invalid port number");
			usage_error(argv[0]); return 0;
			}
		}
	else if (argv[i][0] != '-') {
		strcpy(csp.server,argv[i]);
		}
	else {
		usage_error(argv[0]); return 0;
		}
	}

put_prefs(&cp);
put_sprefs(&csp);
save_sprefs();
return 1;
}


/* usage_error
 * Output correct command line args
 */
static void usage_error(char *av0)
{
logerror(
"usage : %s [server] [-port <number>] [-name <playername>]\n"
"        [-default_3d <level 0,1 or 2>] [-recved_data <mode 0,1,2 or 3>]\n"
"        [-move_speed <speed 0 to 5>] [-password <password>]\n"
"        [-size w h] [-position x y]\n",av0);
}


/* cm
 * Store names, numbers and parameter counts for each command
 */
static struct {
	char *name;
	int num;
	int argc;
	} cm[] = {	"go",cmd_go,2, "say",cmd_say,-1, "yell",cmd_yell,-1,
			"get",cmd_get,2, "look",cmd_look,-1, "do",cmd_do,-1,
			"exam",cmd_exam,2, "quit",cmd_quit,1,"help",cmd_help,-1,
			"drop",cmd_drop,2, "inv",cmd_inv,1, "face",cmd_face,2,
			"give",cmd_give,3, "upload",cmd_upload,2,
			"send",cmd_send,3, "ls",cmd_ls,1, "rm",cmd_rm,2,
			"what",cmd_what,2, "set",cmd_set,3, "prefs",cmd_prefs,1,
			"suicide",cmd_suicide,1, "avatar",cmd_avatar,2,
			"email",cmd_email,2, "fullname",cmd_fullname,2,
			"about",cmd_about,2
			};


/* identify_command
 * Returns the command number for a given command string
 */
int identify_command(char *cmd)
{
int i;

for(i=0; i<sizeof(cm)/sizeof(cm[0]); i++) 
	if (strequ(cm[i].name,cmd))
		return cm[i].num;
return cmd_unknown;
}


/* validate_command
 * Checks if a command has the corrent number of arguments and format
 * arguments.
 */
int validate_command(int opc, char **vec, int c)
{
int i;

/* Check for right number of args
 */
for(i=0; i<sizeof(cm)/sizeof(cm[0]); i++)
	if (cm[i].num == opc) {
		if (cm[i].argc != -1 && cm[i].argc != c)
			return 0;
		break;
		}

/* Check for correct argument format
 */
switch(opc) {
case cmd_go:
	if (atoi(vec[1]) < 0)
		return 0;
	break;

case cmd_look:
	if (c > 2)
		return 0;
	break;

case cmd_do:
	if (c < 3)
		return 0;
	break;

case cmd_help:
	if (c > 2)
		return 0;
	break;

	} /* end switch */

return 1;
}


/* list_commands
 * Print a list of all known command names
 */
void list_commands(void)
{
int i;

for(i=0; i<sizeof(cm)/sizeof(cm[0]); i++) {
	fdprintf(1,"%-15s%s",cm[i].name,(i+1)%5 ? "" : "\n");
	}
fdprintf(1,"\n");
}


/* command_help
 * Display help on a particular command
 */
void command_help(char *cmd)
{
switch(identify_command(cmd)) {
case cmd_go:
	fdputs(1,"go - move from your current position to a neighbouring");
	fdputs(1,"     position.");
	fdputs(1,"formats - go <position name>");
	fdputs(1,"          go <path number>");
	break;
case cmd_say:
	fdputs(1,"say - send a message to nearby players.");
	fdputs(1,"format  - say <word> <word> ...");
	break;
case cmd_yell:
	fdputs(1,"yell - send a message to everybody on this server.");
	fdputs(1,"format  - yell <word> <word> ...");
	break;
case cmd_get:
	fdputs(1,"get - pick up an object at you current location.");
	fdputs(1,"format  - get <object>");
	break;
case cmd_look:
	fdputs(1,"look - display information on your current position or a");
	fdputs(1,"       neighbouring position.");
	fdputs(1,"formats - look");
	fdputs(1,"          look <position name>");
	fdputs(1,"          look <path number>");
	break;
case cmd_do:
	fdputs(1,"do - do something to an object at your position or in your");
	fdputs(1,"     inventory.");
	fdputs(1,"formats - do <action> <object>");
	fdputs(1,"        - do <action> <object> <parameters>");
	break;
case cmd_exam:
	fdputs(1,"exam - display the description of an object in your");
	fdputs(1,"       inventory or at your location.");
	fdputs(1,"format  - exam <object>");
	break;
case cmd_quit:
	fdputs(1,"quit - exit the client and leave the simulation.");
	fdputs(1,"format  - quit");
	break;
case cmd_help:
	fdputs(1,"help - display a list of commands or help on an individual");
	fdputs(1,"       command.");
	fdputs(1,"formats - help");
	fdputs(1,"        - help <command>");
	break;
case cmd_drop:
	fdputs(1,"drop - drop an object from your inventory at your current");
	fdputs(1,"       location.");
	fdputs(1,"format  - drop <object>");
	break;
case cmd_inv:
	fdputs(1,"inv - list the objects in your inventory.");
	fdputs(1,"format  - inv");
	break;
case cmd_face:
	fdputs(1,"face - turn your 3d shape to face a player nearby.");
	fdputs(1,"format  - face <player>");
	break;
case cmd_give:
	fdputs(1,"give - give an object from your inventory to a another");
	fdputs(1,"       nearby player.");
	fdputs(1,"format  - give <object> <player>");
	break;
case cmd_upload:
	fdputs(1,"upload - transfer a data file to the server to be have");
	fdputs(1,"         associated with you.");
	fdputs(1,"format  - upload <filename>");
	break;
case cmd_send:
	fdputs(1,"send - send a data file associated with you to another");
	fdputs(1,"       nearby player or yourself.");
	fdputs(1,"format  - send <data> <player>");
	break;
case cmd_ls:
	fdputs(1,"ls - list all data associated with you on the server.");
	fdputs(1,"format  - ls");
	break;
case cmd_rm:
	fdputs(1,"rm - remove data associated with you from the server.");
	fdputs(1,"format  - rm <data>");
	break;
case cmd_what:
	fdputs(1,"what - display a list of all actions that can be done");
	fdputs(1,"       to an object with do.");
	fdputs(1,"format  - what <object>");
	break;
case cmd_set:
	fdputs(1,"set - set the value of a preference.");
	fdputs(1,"formats - set default_3d <initial 3d level>");
	fdputs(1,"                          0 = worst");
	fdputs(1,"                          1 = intermediate");
	fdputs(1,"                          2 = best");
	fdputs(1,"          set recved_data <received data option>");
	fdputs(1,"                           0 = always save");
	fdputs(1,"                           1 = save or display");
	fdputs(1,"                           2 = always display");
	fdputs(1,"                           3 = always ignore");
	fdputs(1,"          set move_speed <movement speed>");
	fdputs(1,"                           A speed from 0 - 5 m/s");
	fdputs(1,"          set password <password>");
	fdputs(1,"                           A password needed to connect");
	fdputs(1,"                           to a server.");
	fdputs(1,"          set display_transfer <mode>");
	fdputs(1,"                           1 = display transfer status");
	fdputs(1,"                           0 = don't display");
	fdputs(1,"          set transfer_block <size>");
	fdputs(1,"                           Size of blocks read for client-");
	fdputs(1,"                           server transfer");
	break;
case cmd_prefs:
	fdputs(1,"prefs - display current preferences state.");
	fdputs(1,"format  - prefs");
	break;
case cmd_suicide:
	fdputs(1,"suicide - destroy your player.");
	fdputs(1,"format - suicide");
	break;
case cmd_avatar:
	fdputs(1,"avatar - set your avatar to a new shape, read from a file.");
	fdputs(1,"format - avatar <filename>");
	break;
case cmd_email:
	fdputs(1,"email - set your email address for others to see.");
	fdputs(1,"format - email <name@host.domain>");
	break;
case cmd_fullname:
	fdputs(1,"fullname - set your full name for others to see.");
	fdputs(1,"format - fullname <name>");
	break;
case cmd_about:
	fdputs(1,"about - get information about a person.");
	fdputs(1,"format - about <person>");
case cmd_unknown:
	fdprintf(1,"Unknown command %s\n",cmd);
	break;
	} /* end switch */
}


/* cfunc
 * Compare function for sorting angles
 */
static int cfunc(const void *d1, const void *d2)
{
return (int)(*(double *)d1 - *(double *)d2);
}


/* path_angles
 * Returns a list of angles to connecting nodes.
 */
double *path_angles(node *pos)
{
static double ang[MAX_LINKS];
int i;

for(i=0; i<pos->linkc; i++) {
	node *end = other_end(pos->link[i],pos);

	ang[i] = rtod(atan2(end->y - pos->y, end->x - pos->x));
	ang[i] = fixang(ang[i]);
	}
return ang;
}


/* sorted_path_angles
 * Return a sorted array of angles to connecting nodes
 */
double *sorted_path_angles(node *pos)
{
double *ang;

ang = path_angles(pos);
qsort(ang,pos->linkc,sizeof(double),cfunc);
return ang;
}


/* next_angle
 * Given a sorted array of angles, find the index of the first one greater
 * than a.
 */
int next_angle(double *angs, int angc, double a)
{
int i;

for(i=0; i<angc && angs[i] < a; i++)
	;
return i;
}


/* upload_data
 * Upload a named file to the server using a concurrent process.
 */
int upload_data(int fd, char *file)
{
struct stat stbuf;	/* for finding file size */
int f;			/* fd for file to read from */
int port = get_sprefs(NULL)->port + alloc_port();
int i;

if ((f = open(file,O_RDONLY)) == -1)
	return 0;
stat(file,&stbuf);
if ((stbuf.st_mode & S_IFMT) == S_IFDIR || basenam(file)[0] == 0)
	return 0;	/* is a directory! */

client_request(fd,req_upload,basenam(file),(int)stbuf.st_size,port);
if (!sfork()) {
	/* Send data to server concurrently
	 */
	set_signals(SIG_DFL);
	send_data(port,f,stbuf.st_size);
	close(f);
	exit(0);
	}
close(f);
return 1;
}


/* move_along
 * Move forward or backwards along the closest path
 */
void move_along(int fd, person *pr, plview *plv, int fwd)
{
double *ang = path_angles(pr->pos);
double closest = 360.0;		/* degrees to closest path */
int  pnum;			/* number of closest path */
double fa;			/* direction to move */
int i;

fa = fwd ? plv->ang : fixang(plv->ang+180);
for(i=0; i<pr->pos->linkc; i++) {
	double diff = abs(fa - ang[i]);

	if (diff > 180.0)
		diff = 360.0 - diff;
	if (diff < closest) {
		closest = diff;
		pnum = i;
		}
	}
if (closest < 45.0)
	client_request(fd,req_move,pnum);
}


/* rotate_left
 * Rotate viewpoint left by 45 degrees or to the next path, whichever is less.
 */
void rotate_left(person *pr, plview *plv)
{
double *ang = sorted_path_angles(pr->pos);
int lft = next_angle(ang,pr->pos->linkc,fixang(plv->ang-4.0));
double nang,diff;

if (lft == 0) nang = ang[pr->pos->linkc-1];
else          nang = ang[lft-1];
diff = fixang(plv->ang - nang);
if (diff > 45 || pr->pos->linkc <= 1) diff = 45;
plv->ang = fixang(plv->ang - diff);
}


/* rotate_right
 * Rotate viewpoint right by 45 degrees or to the next path, whichever is less.
 */
void rotate_right(person *pr, plview *plv)
{
double *ang = sorted_path_angles(pr->pos);
int lft = next_angle(ang,pr->pos->linkc,fixang(plv->ang+4.0));
double nang,diff;

if (lft == pr->pos->linkc) nang = ang[0];
else                       nang = ang[lft];
diff = fixang(nang - plv->ang); 
if (diff > 45 || pr->pos->linkc <= 1) diff = 45;
plv->ang = fixang(plv->ang + diff);
}

