/* fdio.c
 *
 * Functions for doing stdio-like input and output on file descriptors,
 */
#include "sim.h"

extern int fd_debug;
extern int errno;

/* fdgets
 * Read a line from a file descriptor
 */
char *fdgets(int fd, char *buf, int len)
{
int i;

for(i=0; i<len-1; i++) {
	int rv = fdread(fd,&buf[i],1,1);

	if (rv == 0)
		break;			/* end of file */
	else if (rv < 0) {
		logerror("fdgets : \"%s\"",strerror(errno));
		break;
		}
	else if (buf[i] == '\n') {
		i++;			/* end of line */
		break;
		}
	}
buf[i] = '\0';
if (fd_debug)
	printf("read \"%s\"\n",buf);
if (buf[0] == '\0')
	return NULL;		/* didn't read anything */
return buf;
}


/* fdprintf
 * Write a formatted string to a file descriptor
 */
int fdprintf(int fd, char *fmt, ...)
{
char buf[2048];
int len;
va_list ap;

va_start(ap,fmt);
#ifdef ODD_VSPRINTF
	vsprintf(buf,fmt,ap);
	len = strlen(buf);
#else
	len = vsprintf(buf,fmt,ap);
#endif
if (fdwrite(fd,buf,1,len) != len) 
	logerror("fdprintf : \"%s\"",strerror(errno));
va_end(ap);
if (fd_debug)
	printf("wrote \"%s\"\n",buf);
}


/* fdgetc
 * Get a char from a file descriptor
 */
int fdgetc(int fd)
{
unsigned char c;
int rv;

rv = fdread(fd,&c,1,1);
if (rv == 0)
	return -1;
else if (rv < 0) {
	logerror("fdgetc : \"%s\"",strerror(errno));
	return -1;
	}
else
	return c;
}


/* fdputc
 * Write a char to a file descriptor
 */
void fdputc(int fd, char c)
{
if (fdwrite(fd,&c,1,1) != 1)
	logerror("fdputc : \"%s\"",strerror(errno));
}


/* fdputs
 * Write a line to a file descriptor
 */
void fdputs(int fd, char *s)
{
int len = strlen(s);

if (fdwrite(fd,s,1,len) != len)
	logerror("fdputs : \"%s\"",strerror(errno));
fdputc(fd,'\n');
if (fd_debug)
	printf("Wrote \"%s\"\n",s);
}


/* fdgetw
 * Read one word from a file descriptor
 */
char *fdgetw(int fd, char *buf, int len)
{
int i;
unsigned char c;

/* skip whitespace
 */
do {
	int rv = fdread(fd,&c,1,1);

	if (rv == 0)
		return NULL;
	else if (rv < 0) {
		logerror("fdgetw : \"%s\"",strerror(errno));
		return NULL;
		}
	} while(c == ' ' || c == '\t' || c == '\n');

/* Read word till whitespace
 */
buf[0] = c;
for(i=1; i<len-1; i++) {
	int rv = fdread(fd,&buf[i],1,1);

	if (rv == 0)
		break;		/* end of file */
	else if (rv < 0) {
		logerror("fdgetw : \"%s\"",strerror(errno));
		break;
		}
	else if (buf[i] == ' ' || buf[i] == '\t' || buf[i] == '\n')
		break;		/* end of word */
	}
buf[i] = '\0';
if (fd_debug)
	printf("read \"%s\"\n",buf);
if (buf[0] == '\0')
	return NULL;		/* didn'd read anything */
return buf;
}


/* fdread
 * Reads n objects of size bytes into buf from fd. Returns the number of
 * bytes actually read.
 */
int fdread(int fd, void *buf, int size, int n)
{
int i=0,sz = size*n;

while(i<sz) {
	int got;

	while((got = read(fd,(char *)buf+i,sz-i)) < 0)
		if (errno != EINTR) return -1;
	if (got == 0) return i;
	i+=got;
	}
return i;
}


/* fdwrite
 * Writes n objects of size bytes from buf to fd. Returns the number of
 * bytes actually written.
 */
int fdwrite(int fd, void *buf, int size, int n)
{
int i=0,sz=size*n;

while(i<sz) {
	int put;

	while((put = write(fd,(char *)buf+i,sz-i)) < 0)
		if (errno != EINTR) return -1;
	if (put == 0) return 0;
	i+=put;
	}
return i;
}


/* pipethrough
 * Takes a file descriptor, runs everything that comes out of it through
 * a given command and returns a new fd for the output from that command.
 */
int pipethrough(int fd, int *pid, char *cmd)
{
char **argv;		/* for extracting args */
int argc;
int tube[2];
int status;
int npid;

/* Extract args from command
 */
parse_string(cmd,&argv,&argc);

pipe(tube);
if (!(npid = sfork())) {
	int status;

	set_signals(SIG_DFL);
	close(0);
	dup2(fd,0);
	close(1);
	dup2(tube[1],1);
	execv(argv[0],argv);
	logerror("pipethrough : command %s not found",argv[0]);
	exit(10);
	}
close(tube[1]);
if (pid) *pid = npid;
return tube[0];
}


/* pipeclose
 * Clean up a file descriptor created by pipethrough
 */
int pipeclose(int fd)
{
int status;

close(fd);
wait(&status);
return status;
}


/* pipeinto
 * Takes a file descriptor, outputs everything from a given command to that
 * file desciptor and returns a new fd for input to that command
 */
int pipeinto(int fd, int *pid, char *cmd)
{
char **argv;
int argc;
int tube[2];
int npid;

parse_string(cmd,&argv,&argc);

pipe(tube);
if (!(npid = sfork())) {
	int status;

	set_signals(SIG_DFL);
	close(1);
	dup2(fd,1);
	close(0);
	dup2(tube[0],0);
	execv(argv[0],argv);
	logerror("pipeinto : command %s not found",argv[0]);
	exit(10);
	}
close(tube[0]);
if (pid) *pid = npid;
return tube[0];
}


/* copy_data
 * Copy len bytes from one file descriptor to another. Returns 0 on error,
 * or 1 if successful.
 */
int copy_data(int fd1, int fd2, int len)
{
char buf[BUFSIZ];
int i=0;

while(i<len) {
	int n = (len-i < BUFSIZ ? len-i : BUFSIZ);

	if (fdread(fd1,buf,1,n) == -1) {
		logerror("copy_data : \"%s\"",strerror(errno));
		return 0;
		}
	if (fdwrite(fd2,buf,1,n) == -1) {
		logerror("copy_data : \"%s\"",strerror(errno));
		return 0;
		}
	i += n;
	}
return 1;
}


/* copy_to_eof
 * Copy data from one fd1 to fd2, until the EOF condition is met on fd1
 */
int copy_to_eof(int fd1, int fd2)
{
char buf[BUFSIZ];
int n;

do {
	if ((n = fdread(fd1,buf,1,BUFSIZ)) == -1) {
		logerror("copy_to_eof : \"%s\"",strerror(errno));
		return 0;
		}
	if (fdwrite(fd2,buf,1,n) == -1) {
		logerror("copy_to_eof : \"%s\"",strerror(errno));
		return 0;
		}
	} while(n != 0);
return 1;
}

	
/* reset_fds
 * Close all file descriptors, and reset stdin/out/err
 */
void reset_fds(void)
{
/* int maxfds = getdtablesize(); */
int maxfds = sysconf(_SC_OPEN_MAX);
int i;

for(i=0; i<maxfds; i++)
	close(i);
open("/dev/tty",O_RDONLY);
open("/dev/tty",O_WRONLY);
open("/dev/tty",O_WRONLY);
}

