/* hash.c
 *
 * Functions for managing generic hash tables by name.
 */
#include "sim.h"

static int hash_string(char *s);

/* entry
 * A generic hash table entry. All pointers to be hashed into the table
 * should be to structures starting like this
 */
typedef struct entry {
	char name[256];		/* unique name */
	struct entry *next;	/* next entry with same hash val */
	struct entry *ne;	/* next in list of all entrys */
	} entry;

/* add_entry
 * Insert an entry into a hash table. Entries with a 0-length name are
 * not added to the table, but only to the list of entries.
 */
void add_entry(hashtable *ht, void *e)
{
entry *en = e;
int hash = hash_string(en->name);
entry **ep;

if (en->name[0]) {
	/* Hash to correct position
	 */
	for(ep=(entry **)&ht->tab[hash]; *ep; ep=&((*ep)->next))
		if (strequ((*ep)->name,en->name)) {
			logerror("add_entry : duplicate entry to table : %s",
				 en->name);
			return;
			}
	*ep = en;
	en->next = NULL;
	}

/* Add to list in hash table
 */
if (!ht->head) ht->tailp = &ht->head;
*(ht->tailp) = e;
en->ne = NULL;
ht->tailp = (void **)&en->ne;
}


/* lookup_entry
 * Find a named entry in a hash table, or return NULL
 */
void *lookup_entry(hashtable *ht, char *name)
{
int hash = hash_string(name);
entry *en;

for(en=ht->tab[hash]; en; en=en->next)
	if (strequ(en->name,name))
		return en;
return NULL;
}


/* remove_entry
 * Unlink (but don't free) an entry in a hash table
 */
void remove_entry(hashtable *ht, void *e)
{
entry *en = e;
int hash = hash_string(en->name);
entry **ep;

/* Unlink from list with same hash val
 */
for(ep=(entry **)&ht->tab[hash]; *ep != en; ep=&((*ep)->next))
	;
*ep = en->next;

/* Unlink from list of all entrys
 */
for(ep=(entry **)&ht->head; *ep != en; ep=&((*ep)->ne))
	;
*ep = en->ne;
if (en->ne == NULL) {
	/* This was last one in list
	 */
	ht->tailp = (void **)ep;
	}
}


/* reset_position
 * Resets the step-through pointer to the first entry in the hash table
 */
void reset_position(hashtable *ht)
{
ht->np = ht->head;
}


/* next_position
 * Returns the step-through pointer for the table, and steps in along to 
 * the next entry.
 */
void *next_position(hashtable *ht)
{
entry *curr = ht->np;

if (curr) {
	ht->np = ((entry *)(ht->np))->ne;
	}
return curr;
}


/* clear_entries
 * Remove and free every entry in the hash table.
 */
void clear_entries(hashtable *ht, freefunc ff)
{
entry *en;
int i;

reset_position(ht);
while((en = next_position(ht)))
	ff(en);
for(i=0; i<HASH_TABLE_SIZE; i++)
	ht->tab[i] = NULL;
ht->head  = NULL;
ht->tailp = NULL;
ht->np    = NULL;
}


/* hash_string
 * Hash a string to an integer
 */
static int hash_string(char *s)
{
int sum = 0;

while(*s)
	sum += *(s++);
return sum % HASH_TABLE_SIZE;
}

