%{
/* inventor.y
 *
 * Yacc script for parsing sim 3d descriptions into an Inventor scene.
 * Functions for rendering and manipulating the scene.
 */
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoFaceSet.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoTexture2Transform.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/SbColor.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoOffscreenRenderer.h>

#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <X11/Xlib.h>
extern "C" {
#include "sim.h"
#include "3d.h"
#include "graphics.h"
#include "client.h"

typedef struct {
	double x,y,z;
	} triple;
}

static void new_shape(void);
static void rerender_3d(void);
static void setup_camera(plview *plv);
static char *fix_name(char *n);
static char *unfix_name(char *n);
%}

%union {
	double num;
	char string[256];
	}
%token <num> NUM
%token <string> STRING
%token END,POLY,SPHERE,BOX,CONE,TUBE,COLL,POS,NAME,COLOUR,DISC
%token ROTATE,CENTRE,RADIUS,HEIGHT,POINTS,CORNERS,SURFACE,SCALE,TEXTURE,CMT
%type <num> surftype
%%
input		:	END
			{ return -1;
			}
		|	shape_trans
			{ return 0;
			}
		;

shape_in_coll	:	{ /* Push current shape (collection) onto stack
			   */
			  coll[collp].sha = sha;
			  coll[collp++].coll = sep;
			}
			shape_trans
			{ /* Add new shape to collection
			   */
			  coll[collp-1].coll->addChild(sep);

			  /* Extract collection and attribs from stack
			   */
			  sep = coll[--collp].coll;
			  sha = coll[collp].sha;

			  /* Get pointers to attrib nodes for collection
			   */
			  mat = (SoMaterial *)sep->getChild(0);
			  ttr = (SoTexture2Transform *)sep->getChild(1);
			  tex = (SoTexture2 *)sep->getChild(2);
			  rot = (SoTransform *)sep->getChild(3);
			  sca = (SoTransform *)sep->getChild(4);
			  tra = (SoTransform *)sep->getChild(5);
			  dst = (SoDrawStyle *)sep->getChild(6);
			  pst = (SoPickStyle *)sep->getChild(7);
			  stype = 7;
			}
		;

shape_trans	:	shape
			{ tra->translation.setValue(sha.cx, sha.cy, sha.cz);
			  sca->scaleFactor.setValue(sha.sx, sha.sy, sha.sz);
			  SbVec3f axis(sha.rx, sha.ry, sha.rz);
			  rot->rotation.setValue(axis,-dtor(sha.rang));
			  sca->center.setValue(sha.cx, sha.cy, sha.cz);
			  rot->center.setValue(sha.cx, sha.cy, sha.cz);
			}

shape		:	poly
		|	sphere
		|	box
		|	cone
		|	tube
		|	disc
		|	coll
		;


/*********************** polygon **********************
 */
poly		:	POLY 
			{ new_shape();
			  stype = 1;
			}
			'{' polyattribs '}'
		;

polyattribs	:	polyattribs polyattrib
		|	polyattribs shapeattrib
		|
		;

polyattrib	:	POINTS
			{ ptc=0;
			}
			polypoints
			{ SoShapeHints *phts = new SoShapeHints;
			  SoCoordinate3 *pcor = new SoCoordinate3;
			  SoFaceSet *pfac = new SoFaceSet;
			  long nv[1];

			  /* ensure both sides are lit. */
			  phts->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;

			  pcor->point.setValues(0,ptc,ppt);
			  nv[0] = ptc;
			  pfac->numVertices.setValues(0,1,nv);
			  sep->addChild(phts);
			  sep->addChild(pcor);
			  sep->addChild(pfac);
			}
		;

polypoints	:	polypoints polypoint
		|
		;

polypoint	:	NUM NUM NUM
			{ ppt[ptc][0] = $1;
			  ppt[ptc][1] = $3;	/* inventor thinks y is up */
			  ppt[ptc][2] = $2;
			  ptc++;
			}


/*********************** sphere **********************
 */
sphere		:	SPHERE
			{ new_shape();
			  stype = 2;
			  sep->addChild(new SoSphere);
			  sha.sx *= 0.5; sha.sy *= 0.5; sha.sz *= 0.5;
			}
			'{' sphereattribs '}'
		;

sphereattribs	:	sphereattribs sphereattrib
		|	sphereattribs shapeattrib
		|
		;

sphereattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $4; sha.cz += $3;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sy *= $2*2; sha.sz *= $2*2;
			}
		;


/*********************** box **********************
 */
box		:	BOX 
			{ new_shape();
			  stype = 3;
			  sep->addChild(new SoCube);
			  sha.sx = 1.0/2; sha.sy = 1.0/2; sha.sz = 1.0/2;
			}
			'{' boxattribs '}'
		;

boxattribs	:	boxattribs boxattrib
		|	boxattribs shapeattrib
		|
		;

boxattrib	:	CORNERS NUM NUM NUM  NUM NUM NUM
			{ double x1=$2,y1=$3,z1=$4, x2=$5,y2=$6,z2=$7;

			  if (x2 < x1) swapdbl(&x1,&x2);
			  if (y2 < y1) swapdbl(&y1,&y2);
			  if (z2 < z1) swapdbl(&z1,&z2);
			  sha.cx += (x1+x2)/2; sha.cy += (z1+z2)/2;
			  sha.cz += (y1+y2)/2;
			  sha.sx *= (x2-x1); sha.sy *= (z2-z1);
			  sha.sz *= (y2-y1);
			}
		;


/*********************** cone **********************
 */
cone		:	CONE 
			{ new_shape();
			  stype = 4;
			  sep->addChild(new SoCone);
			  sha.sx = 1.0/2; sha.sz = 1.0/2; sha.sy = 1.0/2;
			}
			'{' coneattribs '}'
		;

coneattribs	:	coneattribs coneattrib
		|	coneattribs shapeattrib
		|
		;

coneattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $4; sha.cz += $3;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sz *= $2*2;
			}
		|	HEIGHT NUM
			{ sha.sy *= $2;
			}
		;


/*********************** tube **********************
 */
tube		:	TUBE
			{ new_shape();
			  stype = 5;
			  sep->addChild(new SoCylinder);
			  sha.sy = 1.0/2; sha.sx = 1.0/2; sha.sz = 1.0/2;
			}
			'{' tubeattribs '}'
		;

tubeattribs	:	tubeattribs tubeattrib
		|	tubeattribs shapeattrib
		|
		;

tubeattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $4; sha.cz += $3;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sz *= $2*2;
			}
		|	HEIGHT NUM
			{ sha.sy *= $2;
			}


/*********************** disc **********************
 */
disc		:	DISC
			{ new_shape();
			  stype = 6;
			  SoSeparator *fsep = new SoSeparator;
			  SoShapeHints *fhts = new SoShapeHints;
			  SoCoordinate3 *fcor = new SoCoordinate3;
			  SoFaceSet *ffac = new SoFaceSet;
			  long nv[1];
			  float fpoints[18][3];
			  int i;

			  fhts->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
			  for(i=0; i<18; i++) {
			    fpoints[i][0] = sin(dtor(i*20))*0.5;
			    fpoints[i][1] = cos(dtor(i*20))*0.5;
			    fpoints[i][2] = 0.0;
			    }
			  fcor->point.setValues(0,18,fpoints);
			  nv[0] = 18;
			  ffac->numVertices.setValues(0,1,nv);
			  fsep->addChild(fhts);
			  fsep->addChild(fcor);
			  fsep->addChild(ffac);
			  sep->addChild(fsep);
			}
			'{' discattribs '}'
		;

discattribs	:	discattribs discattrib
		|	discattribs shapeattrib
		|
		;

discattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $4; sha.cz += $3;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sy *= $2*2;
			}
		;


/*********************** coll **********************
 */
coll		:	COLL
			{ new_shape();
			  stype = 7;
			}
			'{' collattribs '}'
		;

collattribs	:	collattribs shape_in_coll
		|	collattribs shapeattrib
		|
		;


shapeattrib	:	NAME STRING
			{ sep->setName(fix_name($2));
			}
		|	POS NUM NUM NUM
			{ sha.cx += $2; sha.cy += $4; sha.cz += $3;
			}
		|	COLOUR NUM NUM NUM
			{ mat->diffuseColor.setValue($2/255,$3/255,$4/255);
			}
		|	ROTATE NUM  NUM NUM NUM
			{ sha.rx = $3; sha.ry = $5; sha.rz = $4; sha.rang = $2;
			}
		|	SURFACE surftype
			{ if ($2 == sft_flat) {
			    mat->transparency = 0.0;
			    mat->shininess    = 0.0;
			    mat->specularColor.setValue(0,0,0);
			    }
			  else if ($2 == sft_glass) {
			    mat->transparency = 0.5;
			    }
			  else if ($2 == sft_shiny) {
			    mat->shininess = 1.0;
			    mat->specularColor.setValue(1.0,1.0,1.0);
			    }
			}
		|	SCALE NUM NUM NUM
			{ sha.sx *= $2; sha.sz *= $3; sha.sy *= $4;
			}
		|	TEXTURE NUM NUM
			{ texw = (int)$2; texh = (int)$3;
			  texdat = (unsigned char *)calloc(1,texw*texh*3);
			  texc = 0;
			}
			'{' texture_strings '}'
			{ int y;		/* for inverting textures */
			  void *textmp;

			  if (texc != texw*texh*3) {
			    yyerror("Incorrect amount of texture data read");
			    return 1;
			    }

			  /* Invert texture for inventor 
			   */
			  textmp = malloc(texw*3);
			  for(y=0; y<texh/2; y++) {
			    void *ulp = &texdat[y*texw*3];
			    void *llp = &texdat[(texh-y-1)*texw*3];

			    memcpy(textmp,ulp,texw*3);
			    memcpy(ulp,llp,texw*3);
			    memcpy(llp,textmp,texw*3);
			    }
			  free(textmp);

			  tex->image.setValue(SbVec2s(texw,texh),3,texdat);
			  // tex->model = SoTexture2::DECAL;
			  free(texdat);
			  if (stype == 1) {
			    /* fix bug in texture mapping to planes
			     */
			    ttr->scaleFactor.setValue(1.0,texw/(float)texh);
			    }
			  mat->diffuseColor.setValue(1.0, 1.0, 1.0);
			}
		|	CMT STRING
		;

surftype	:	NUM
			{ $$ = $1; }
		|	STRING
			{ if (strequ($1,"flat"))
			    $$ = sft_flat;
			  else if (strequ($1,"glass"))
			    $$ = sft_glass;
			  else if (strequ($1,"shiny"))
			    $$ = sft_shiny;
			  else {
			    yyerror("Unknown surface type");
			    return 1;
			    }
			}
		;

texture_strings	:	texture_strings texture_string
		|
		;

texture_string	:	STRING
			{ int len = strlen($1);
			  decode_rgb($1,&texdat[texc],len);
			  texc += (len/4)*3;
			}

%%
static Display *display;		/* display for 3d rendering */
static Window vwin;			/* window for rendering into */
static Colormap ccmap = None;		/* custom colourmap for ui */
static SoSeparator *scene_root;
static SoPerspectiveCamera *camera;
static SoPointLight *light;		/* light at player's positon */
static SoDirectionalLight *light2;	/* general lights on scene */
static SoRayPickAction *picker;		/* action for getting object clicked */
static SoComplexity *complex;		/* complex / simple control node */
static SoEnvironment *environ;		/* scene environment */
static SoSeparator *sel;		/* base of highlighting shape */
static Colormap cmap = None;		/* colourmap for main window */
static SoGLRenderAction *raction;	/* render action for GL */
static SbViewportRegion *vport;		/* viewport for raction */
static SoGetBoundingBoxAction *bbaction;/* action for highlighting */

static SoSeparator *sep;		/* base of new shape */
static SoMaterial *mat;			/* colour and texture of new shape */
static SoTexture2Transform *ttr;	/* transform applied to texture */
static SoTexture2 *tex;			/* texture of new shape */
static SoTransform *sca;		/* scaling of new shape */
static SoTransform *rot;		/* rotation of new shape */
static SoTransform *tra;		/* translation of new shape */
static SoDrawStyle *dst;		/* invisible/solid/wire for shape */
static SoPickStyle *pst;		/* invisible shapes are unpickable */

static int ptc;			/* no. of points in polygon */
static float ppt[253][3];	/* coords of polygon points */
static shapeattribs sha;	/* scale,rotation and pos of shape and coll */
int stype;			/* type of current shape */

static struct {			/* stack of collections attribs and pointers */
	shapeattribs sha;
	SoSeparator *coll;
	} coll[256];
static int collp;		/* position in stack */

static unsigned char *texdat;	/* texture data */
static int texw, texh;		/* size of texture */
static int texc;		/* number of rgb triples read */

static char *last_click_name;   /* name of last object clicked on */
static int last_click_button;   /* last button clicked */

static SoSeparator *hs = NULL;	/* currently highlighted object */

static reserved rwtab[] = {	"end",END,"poly",POLY,
				"sphere",SPHERE,"box",BOX,"cone",CONE,
				"tube",TUBE,"disc",DISC,"coll",COLL,
				"pos",POS,"name",NAME,"colour",COLOUR,
				"rotate",ROTATE,"centre",CENTRE,
				"radius",RADIUS,"height",HEIGHT,
				"points",POINTS,"corners",CORNERS,
				"surface",SURFACE,"scale",SCALE,
				"texture",TEXTURE,"cmt",CMT };
#define error yyerror
#define lex yylex
#define lval yylval

#include "lexer.h"


/* new_shape
 * Common stuff for all new shapes
 */
static void new_shape(void)
{
sep = new SoSeparator;
mat = new SoMaterial;
ttr = new SoTexture2Transform;
tex = new SoTexture2;
sca = new SoTransform;
rot = new SoTransform;
tra = new SoTransform;
dst = new SoDrawStyle;
pst = new SoPickStyle;
sep->addChild(mat);
sep->addChild(ttr);
sep->addChild(tex);
sep->addChild(rot);
sep->addChild(sca);
sep->addChild(tra);
sep->addChild(dst);
sep->addChild(pst);
sha.cx = sha.cy = sha.cz = 0.0;
sha.rx = sha.ry = sha.rz = 0.0; sha.rang = 0.0;
sha.sx = sha.sy = sha.sz = 1.0;
}


/* supported_3d
 * Returns the supported 3d level.
 */
int supported_3d(void)
{
userprefs *cp = get_prefs(NULL);

if (cp->default_3d == 2)
	return lvl_textured;
else
	return lvl_flat;
}


/* parse_3d
 * Parse one 3d object and add it to the object table. 
 */
int parse_3d(int filed)
{
int rv;

init_lexer(filed);
collp = 0;
rv = yyparse();
if (rv == 0) {
	char *oname = (char *)sep->getName().getString();
	if (!oname || unfix_name(oname)[0] != '_')
		scene_root->addChild(sep);
	}
return rv;
}


/* parse_3ds
 * Parse 3d objects and add them to the hash table until an 'end' is read.
 */
int parse_3ds(int filed)
{
int rv;

init_lexer(filed);
collp = 0;
while((rv = yyparse()) == 0) {
	char *oname = (char *)sep->getName().getString();
	if (!oname || unfix_name(oname)[0] != '_')
		scene_root->addChild(sep);
	}
if (rv == 1)
	return 1;
else
	return 0;
}


/* init_3d
 * Create a window for the 3d view.
 * Initialize the inventor scene.
 */
void init_3d(char *display_name, Window par, int x, int y, int w, int h)
{
XVisualInfo *vi;		/* visuals usable by GL */
static int attributeList[] = {
		GLX_RGBA, GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1, GLX_DEPTH_SIZE, 1, GLX_DOUBLEBUFFER, 
		None,
		};		/* GL visual type needed */
GLXContext cx;			/* context for GL */
XSetWindowAttributes swa;	/* attribs for new window */
int errorb,eventb;		/* used by glXQueryExtension */

w--; h--;
setup_scenegraph_3d();

/* Use GL to create a window for rendering into
 */
display = XOpenDisplay(display_name);
if (!glXQueryExtension(display,&errorb,&eventb)) {
	fprintf(stderr,"GL not available on X server %s\n",
		XDisplayName(display_name));
	exit(2);
	}
vi = glXChooseVisual(display,DefaultScreen(display),attributeList);
cx = glXCreateContext(display, vi, 0, GL_TRUE);
cmap = XCreateColormap(display,RootWindow(display,vi->screen),vi->visual,
		       AllocNone);
swa.colormap = cmap;
swa.border_pixel = 0;
swa.event_mask = StructureNotifyMask | ExposureMask | ButtonPressMask;
vwin = XCreateWindow(display,par,x,y,w,h,0,vi->depth,InputOutput,vi->visual,
	             (CWBorderPixel | CWColormap | CWEventMask),&swa);
XMapWindow(display,vwin);
XFlush(display);
glXMakeCurrent(display,vwin,cx);
glEnable(GL_DEPTH_TEST);

/* Tell inventor to use this window
 */
vport = new SbViewportRegion(w,h);
raction = new SoGLRenderAction(*vport);
picker = new SoRayPickAction(vport->getWindowSize());
bbaction = new SoGetBoundingBoxAction(vport->getWindowSize());

/* Create shape for highlighting things
 */
sel = new SoSeparator;
SoMaterial *selm = new SoMaterial;
sel->addChild(selm);
sel->addChild(new SoTexture2Transform);
sel->addChild(new SoTexture2);
sel->addChild(new SoTransform);
sel->addChild(new SoTransform);
sel->addChild(new SoTransform);
SoPickStyle *psty = new SoPickStyle;
sel->addChild(psty);
selm->diffuseColor.setValue(1.0, 0.0, 0.0);
selm->transparency = 0.5;
psty->style = SoPickStyle::UNPICKABLE;
scene_root->addChild(sel);

int sn = DefaultScreen(display);
if (DefaultDepth(display,sn) <= 8) {
	/* Alloc custom colourmap for best_colour()
	 */
	ccmap = XCreateColormap(display,par,DefaultVisual(display,sn),AllocAll);
	XSetWindowColormap(display,par,ccmap);
	XFlush(display);
	}
}


/* setup_scenegraph_3d
 * Create the base scene graph. This function is called instead of init_3d()
 * for programs that output scenes to a file.
 */
void setup_scenegraph_3d(void)
{
/* Initialize 3d scene
 */
SoDB::init();
scene_root = new SoSeparator;
camera = new SoPerspectiveCamera;
camera->farDistance = 10000000.0;
camera->nearDistance = 0.2;
light = new SoPointLight;
light->intensity = 0.7;
light2 = new SoDirectionalLight;
light2->intensity = 0.3;
light2->direction.setValue(-1,-1,-1);
complex = new SoComplexity;
environ = new SoEnvironment;
environ->ambientIntensity = 0.5;
environ->ambientColor.setValue(1.0, 1.0, 1.0);
scene_root->ref();
scene_root->addChild(complex);
scene_root->addChild(camera);
scene_root->addChild(light);
scene_root->addChild(light2);
scene_root->addChild(environ);

/* set draw style for different graphics levels
 */
userprefs *cp = get_prefs(NULL);
if (cp->default_3d == 0) {
	complex->value = 0.2;
	}
else {
	complex->value = 0.5;
	glClearColor(0.2,0.2,1.0,0);
	}
}


/* new_3d_level
 * Change the 3d graphics level
 */
void new_3d_level(void)
{
switch(get_prefs(NULL)->default_3d) {
case 0:
	complex->value = 0.2;
	glClearColor(0.0, 0.0, 0.0, 0);
	break;
case 1:
case 2:
	complex->value = 0.5;
	glClearColor(0.2, 0.2, 1.0, 0);
	break;
	}
rerender_3d();
}


/* threed_fd
 * Returns the display used by the 3d render area widget
 */
int threed_fd(void)
{
return ConnectionNumber(display);
}


/* move_3d
 * Move an s3dt object by dx,dy,dz
 */
void move_3d(char *name, double dx, double dy, double dz)
{
SoSeparator *ms = (SoSeparator *)SoNode::getByName(fix_name(name));
SoTransform *tr = (SoTransform *)ms->getChild(5);
SoTransform *sc = (SoTransform *)ms->getChild(4);
SoTransform *ro = (SoTransform *)ms->getChild(3);
SbVec3f pos = tr->translation.getValue();

pos[0] += dx; pos[1] += dz; pos[2] += dy;
tr->translation.setValue(pos);
sc->center.setValue(pos);
ro->center.setValue(pos);
}


/* del_3d
 * Remove a 3d object and all it's parts
 */
void del_3d(char *name)
{
SoNode *ms = SoNode::getByName(fix_name(name));

if ((SoSeparator *)ms == hs)
	unhighlight_3d();
scene_root->removeChild(ms);
}


/* clear_3d
 * Delete all 3d objects in the scene (but not cameras, lights, etc..)
 */
void clear_3d(void)
{
int i;

/* Remove children of the root node that are separators, except for the
 * box used for highlighting.
 */
i=0;
while(i < scene_root->getNumChildren()) {
	SoNode *ch = scene_root->getChild(i);

	if (ch->isOfType(SoSeparator::getClassTypeId()) && ch != sel)
		scene_root->removeChild(i);
	else
		i++;
	}
}


static int redraw_needed = 0;	/* should render_3d actually render? */

/* render_3d
 * Tell inventor to draw the current scene
 */
int render_3d(plview *plv)
{
if (redraw_needed) {
	setup_camera(plv);
	rerender_3d();
	redraw_needed = 0;
	return 1;
	}
return 0;
}


/* output_rgb_3d
 * Write a view of the current scene to a FILE, as a 300 x 200 rgb file
 */
void output_rgb_3d(plview *plv, FILE *fp)
{
SbViewportRegion rgbvp(300,200);
SoOffscreenRenderer offscr(rgbvp);

if (get_prefs(NULL)->default_3d == 0)
	offscr.setBackgroundColor(SbColor(0.0, 0.0, 0.0));
else
	offscr.setBackgroundColor(SbColor(0.2, 0.2, 1.0));
setup_camera(plv);
offscr.render(scene_root);
offscr.writeToRGB(fp);
}


/* setup_camera
 * Set the camera node from a plview
 */
static void setup_camera(plview *plv)
{
double x,y,z;		/* view position */
double nx, ny, nz;	/* view direction */

current_position(plv,&x,&y,&z);
camera->position.setValue(x,z,y);
light->location.setValue(x,z,y);

nx = cos(dtor(plv->ang))*cos(dtor(plv->tilt));
ny = sin(dtor(plv->ang))*cos(dtor(plv->tilt));
nz = sin(dtor(plv->tilt));
SbVec3f tar(x+nx,z+nz,y+ny);
camera->pointAt(tar);
}


/* need_redraw
 * Indicate that a redraw should be one on the next call to render_3d()
 */
void need_redraw(void)
{
redraw_needed = 1;
}


/* rerender_3d
 * Redraw display with current point of view
 */
static void rerender_3d(void)
{
glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
if (get_prefs(NULL)->default_3d == 2)
	glShadeModel(GL_SMOOTH);
else
	glShadeModel(GL_FLAT);
raction->apply(scene_root);
glXSwapBuffers(display,vwin);
XSync(display,False);
}


/* check_3d
 * Poll for and handle events for the 3d view window
 */
void check_3d(void)
{
while(XPending(display)) {
	XEvent event;

	XNextEvent(display,&event);
	if (event.type == Expose && !event.xexpose.count &&
	    event.xexpose.window == vwin) {
		rerender_3d();
		}
	else if (event.type == ButtonPress && event.xbutton.window == vwin) {
		/* Find out what was clicked */
		last_click_button = event.xbutton.button;
		last_click_name   = onscreen_3d(event.xbutton.x,
						event.xbutton.y);
		}
		
	}
}


/* resize_3d
 * Resize the render area used by inventor
 */
void resize_3d(int x, int y, int w, int h)
{
vport->setWindowSize(w,h);
raction->setViewportRegion(*vport);
XMoveResizeWindow(display,vwin,x,y,w,h);
XRaiseWindow(display,vwin);
rerender_3d();

delete picker;	/* need to chuck away and re-create because of size change */
delete bbaction;
picker = new SoRayPickAction(vport->getWindowSize());
bbaction = new SoGetBoundingBoxAction(vport->getWindowSize());
}


/* best_colour
 * Returns the pixel value of the colour which best matches the rgb values 
 * given.
 */
long best_colour(int red, int green, int blue)
{
XColor col;
static int pv = 0;

col.red = red*256;
col.green = green*256;
col.blue = blue*256;
col.flags = DoRed | DoGreen | DoBlue;
if (ccmap == None) {
	/* Use the default colourmap
	 */
	Colormap defcm = DefaultColormap(display,DefaultScreen(display));
	XAllocColor(display,defcm,&col);
	}
else {
	col.pixel = pv++;
	XStoreColor(display,ccmap,&col);
	}
return col.pixel;
}


/* rotate_3d
 * Rotate a shape to the given angle about the given axis.
 */
void rotate_3d(char *name, double ang, double x, double y, double z)
{
SoSeparator *ms = (SoSeparator *)SoNode::getByName(fix_name(name));
SbVec3f axis(x,z,y);

((SoTransform *)ms->getChild(3))->rotation.setValue(axis,-dtor(ang));
}


/* onscreen_3d
 * Returns the name of the shape at a given position in the 3d view 
 * window, or NULL
 */
char *onscreen_3d(int x, int y)
{
SoPickedPoint *ppoint;
SoPath *pth;		/* path to picked object */
SoNode *pnd;		/* separator of object clicked */
static char name[256];	/* name of clicked object */

/* Use inventor's reversed y coordinates
 */
y = (raction->getViewportRegion().getWindowSize())[1] - y;
SbVec2s pt(x,y);	/* screen point clicked */

/* Find the node clicked.
 */
picker->setPickAll(FALSE);
picker->setPoint(pt);
picker->apply(scene_root);
ppoint = picker->getPickedPoint();
if (!ppoint) return NULL;
pth = ppoint->getPath();
pnd = pth->getNode(1);
strcpy(name,unfix_name((char *)pnd->getName().getString()));
return name;
}


/* highlight_3d
 * Set the currently highlighted shape
 */
void highlight_3d(char *name)
{
SbBox3f bbox;
double xs,ys,zs, cx,cy,cz;

unhighlight_3d();

/* Find bounding box
 */
hs = (SoSeparator *)SoNode::getByName(fix_name(name));
bbaction->setInCameraSpace(FALSE);
bbaction->apply(hs);
bbox = bbaction->getBoundingBox();
xs = (bbox.getMax()[0] - bbox.getMin()[0])/2.0 + .02;
ys = (bbox.getMax()[1] - bbox.getMin()[1])/2.0 + .02;
zs = (bbox.getMax()[2] - bbox.getMin()[2])/2.0 + .02;
cx = bbox.getCenter()[0];
cy = bbox.getCenter()[1];
cz = bbox.getCenter()[2];

/* Move bounding cube into place
 */
sel->addChild(new SoCube);
((SoTransform *)sel->getChild(4))->scaleFactor.setValue(xs,ys,zs);
((SoTransform *)sel->getChild(4))->center.setValue(cx,cy,cz);
((SoTransform *)sel->getChild(5))->translation.setValue(cx,cy,cz);
}


/* unhighlight_3d
 * Cancel highlighting on the current shape
 */
void unhighlight_3d(void)
{
if (hs) {
	sel->removeChild(7);	/* remove the bounding cube */
	hs = NULL;
	}
}


/* real_colour_3d
 * Change the colour of a shape and all it's child shapes
 */
static void real_colour_3d(SoSeparator *ms, int r, int g, int b)
{
SoMaterial *mt = (SoMaterial *)ms->getChild(0);
int i;

mt->diffuseColor.setValue(r/256.0,g/256.0,b/256.0);
for(i=0; i<ms->getNumChildren(); i++) {
	SoNode *ch = ms->getChild(i);

	if (ch->isOfType(SoSeparator::getClassTypeId()))
		real_colour_3d((SoSeparator *)ch,r,g,b);
	}
}

/* colour_3d
 * Change the colour of a named 3d shape, and all it's children
 */
void colour_3d(char *name, int r, int g, int b)
{
SoSeparator *ms = (SoSeparator *)SoNode::getByName(fix_name(name));

real_colour_3d(ms,r,g,b);
}


/* real_shade_3d
 * Change the surface type of a shape and it's children
 */
static void real_shade_3d(SoSeparator *ms, int shd)
{
SoMaterial *mt = (SoMaterial *)ms->getChild(0);
int i;

if (shd == sft_flat) {
	mt->transparency = 0.0;
	mt->shininess    = 0.0;
	mt->specularColor.setValue(0,0,0);
	}
else if (shd == sft_glass) {
	mt->transparency = 0.5;
	}
else if (shd == sft_shiny) {
	mt->shininess    = 1.0;
	mt->specularColor.setValue(1.0,1.0,1.0);
	}
for(i=0; i<ms->getNumChildren(); i++) {
	SoNode *ch = ms->getChild(i);

	if (ch->isOfType(SoSeparator::getClassTypeId()))
		real_shade_3d((SoSeparator *)ch,shd);
	}
}

/* shade_3d 
 * Change the surface type of a named 3d shape
 */
void shade_3d(char *name, int shd)
{
SoSeparator *ms = (SoSeparator *)SoNode::getByName(fix_name(name));

real_shade_3d(ms,shd);
}


/* get_3d_click
 * Get the name of the last shape clicked on, plus the button used to click it
 */
int get_3d_click(char **n, int *b)
{
if (last_click_name) {
	*n = last_click_name;
	*b = last_click_button;
	last_click_name = NULL;
	return 1;
	}
return 0;

}


/* shutdown_3d
 * Called just before exit() to free up stuff
 */
void shutdown_3d(void)
{
delete picker;
scene_root->unref();
}


/* boundingbox_3d
 * Compute the bounding box of the entire scene
 */
void boundingbox_3d(double *x1, double *y1, double *z1,
		    double *x2, double *y2, double *z2)
{
SbBox3f bbox;

bbaction->setInCameraSpace(FALSE);
bbaction->apply(scene_root);
bbox = bbaction->getBoundingBox();
*x1 = bbox.getMin()[0]; *y1 = bbox.getMin()[2]; *z1 = bbox.getMin()[1];
*x2 = bbox.getMax()[0]; *y2 = bbox.getMax()[2]; *z2 = bbox.getMax()[1];
}


/* real_visible_3d
 * Change the visibility of a shape and all it's child shapes
 */
static void real_visible_3d(SoSeparator *ms, int vis)
{
SoDrawStyle *ds = (SoDrawStyle *)ms->getChild(6);
SoPickStyle *ps = (SoPickStyle *)ms->getChild(7);
int i;

ds->style = vis ? SoDrawStyle::FILLED : SoDrawStyle::INVISIBLE;
ps->style = vis ? SoPickStyle::SHAPE : SoPickStyle::UNPICKABLE;
for(i=0; i<ms->getNumChildren(); i++) {
	SoNode *ch = ms->getChild(i);

	if (ch->isOfType(SoSeparator::getClassTypeId()))
		real_visible_3d((SoSeparator *)ch,vis);
	}
}


/* visible_3d
 * Make a shape visible or invisible
 */
void visible_3d(char *name, int vis)
{
SoSeparator *ms = (SoSeparator *)SoNode::getByName(fix_name(name));

real_visible_3d(ms,vis);
}


/* fix_name
 * Convert a name from a general string to something inventor can live with
 * as a name.
 */
static char *fix_name(char *n)
{
static char buf[256];
char *bufp = buf;

for(; *n; n++) {
	if (isalnum(*n))
		*(bufp++) = *n;
	else {
		*(bufp) = '_';
		sprintf(bufp+1,"%.2x",*n);
		bufp += 3;
		}
	}
*bufp = 0;
return buf;
}


/* unfix_name
 * Convert a name from the format output by fix_name back to the original
 */
static char *unfix_name(char *n)
{
static char buf[256];
char *bufp = buf;

while(*n) {
	if (*n == '_') {
		char nb[3] = {0};
		int hex;

		nb[0] = *(n+1);
		nb[1] = *(n+2);
		sscanf(nb,"%x",&hex);
		*bufp = hex;
		n += 3;
		}
	else
		*bufp= *(n++);
	bufp++;
	}
*bufp = 0;
return buf;
}

