/* lexer.c
 *
 * yylex() function and supporting code for use in parsers.
 * This file is included by the various parsers. Before inclusion, the
 * .y file should have defines changing error and lex to whatever functions
 * are called in that parser for error reporting and lexing.
 */
#include "sim.h"

extern int parse_debug;

typedef struct {
	char *word;
	int tok;
	} reserved;

static void error(char *s);
static void nextchar(void);
int yylex(void);
static int reservedword(char *w);

static char currch;		/* character just read by nextchar() */
static char filename[256];	/* file currently being processed */
static int lineno;		/* line up to in file */
static int fd;			/* file descriptor being read from */

/* error
 * Called by yacc when a syntactic error occurs
 */
static void error(char *s)
{
logerror("Parse error : %s\nNear line %d in %s",s,lineno,filename);
}


/* nextchar
 * Get one more char from the input source
 */
static void nextchar(void)
{
currch = fdgetc(fd);
if (currch == '\n')
	lineno++;
}


/* lex
 * Called by yacc to read on token
 */
int lex(void)
{
char tok[256] = {0};
char *tp = tok;

/* skip whitespace
 */
while(currch == ' ' || currch == '\t' || currch == '\n') {
	nextchar();
	}

if (isdigit(currch) || currch == '.' || currch == '-') {
	/* A number
	 */
	double num;

	while(isdigit(currch) || currch == '.' || currch == '-') {
		*tp++ = currch;
		nextchar();
		}
	yylval.num = atof(tok);
	if (parse_debug)
		printf("num : %f\n",yylval.num);
	return NUM;
	}
else if (currch == '\"') {
	/* A quoted string
	 */
	nextchar();
	while(currch != '\"') {
		*tp++ = currch;
		nextchar();
		}
	nextchar();
	strcpy(yylval.string,tok);
	if (parse_debug)
		printf("string : %s\n",yylval.string);
	return STRING;
	}
else if (isalpha(currch)) {
	/* An alphanumeric string
	 */
	int rw;

	while(isdigit(currch) || isalpha(currch) || currch == '_' ||
	      currch == '.') { 
		*tp++ = currch;
		nextchar();
		}
	if ((rw = reservedword(tok)) != -1) {
		if (parse_debug)
			printf("reserved : %d = %s\n",rw,tok);
		return rw;
		}
	strcpy(yylval.string,tok);
	if (parse_debug)
		printf("string : %s\n",yylval.string);
	return STRING;
	}
else if (currch == '#') {
	/* cpp file info line
	 */
	char line[256];		/* line starting with '#' */
	char *lnum;		/* real line number */
	char *fnam;		/* real file name */
	int i=0;

	while(currch != '\n') {
		nextchar();
		line[i++] = currch;
		}
	lnum = strtok(line," \n");
	fnam = strtok(NULL," \n");
	lineno = atoi(lnum);
	strcpy(filename,fnam);
	nextchar();
	return yylex();		/* call again to get next token */
	}
else if (currch == -1) {
	/* end of file
	 */
	return END;
	}
else {
	/* some other character.
	 */
	int ch = currch;

	nextchar();
	if (parse_debug)
		printf("char : %c\n",ch);
	return ch;
	}
}


/* init_lexer
 * Reset lexer variables
 */
void init_lexer(int filed)
{
fd = filed;
currch = ' ';
lineno = 0;
strcpy(filename,"unknown");
}


/* reservedword
 * Looks up a string in the reserved word table, rwtab.
 * This table must be defined in the parser into which this file is included.
 */
static int reservedword(char *s)
{
int i;

for(i=0; i<sizeof(rwtab)/sizeof(rwtab[0]); i++)
	if (strequ(rwtab[i].word,s))
		return rwtab[i].tok;
return -1;
}

