/* lexer.h
 *
 * lex() function and supporting code for use in parsers.
 */
#include "sim.h"

extern int parse_debug;

static int reservedword(char *s);

static int currch;		/* character just read by nextchar() */
static char filename[256];	/* file currently being processed */
static int lineno;		/* line up to in file */
static int fd;			/* file descriptor being read from */

/* error
 * Called by yacc when a syntactic error occurs
 */
static void error(char *s)
{
logerror("Parse error : %s\nNear line %d in %s",s,lineno,filename);
}


/* nextchar
 * Get one more char from the input source
 */
static void nextchar(void)
{
char c;

if (read(fd,&c,1) != 1)
	currch = END;
else
	currch = c;
if (currch == '\n')
	lineno++;
}


/* lex
 * Called by yacc to read on token
 */
int lex(void)
{
char tok[256] = {0};
char *tp = tok;

/* skip whitespace
 */
while(currch == ' ' || currch == '\t' || currch == '\n') {
	nextchar();
	}

if (currch == END)
	return END;
else if (isdigit(currch) || currch == '.' || currch == '-') {
	/* A number
	 */
	while(isdigit(currch) || currch == '.' || currch == '-') {
		*tp++ = currch;
		nextchar();
		}
	lval.num = atof(tok);
	if (parse_debug)
		printf("num : %f\n",lval.num);
	return NUM;
	}
else if (currch == '\"') {
	/* A quoted string
	 */
	nextchar();
	while(currch != '\"') {
		*tp++ = currch;
		nextchar();
		}
	nextchar();
	strcpy(lval.string,tok);
	if (parse_debug)
		printf("string : %s\n",lval.string);
	return STRING;
	}
else if (isalpha(currch) || currch == '_') {
	/* An alphanumeric string
	 */
	int rw;

	while(currch != ' ' && currch != '\n' && currch != '\t') {
		*tp++ = currch;
		nextchar();
		}
	if ((rw = reservedword(tok)) != -1) {
		if (parse_debug)
			printf("reserved : %d = %s\n",rw,tok);
		return rw;
		}
	strcpy(lval.string,tok);
	if (parse_debug)
		printf("string : %s\n",lval.string);
	return STRING;
	}
else if (currch == '#') {
	/* cpp file info line
	 */
	char line[256];		/* line starting with '#' */
	char *lnum;		/* real line number */
	char *fnam;		/* real file name */
	int i=0;

	while(currch != '\n') {
		nextchar();
		line[i++] = currch;
		}
	lnum = strtok(line," \n");
	fnam = strtok(NULL," \n");
	lineno = atoi(lnum);
	strcpy(filename,fnam);
	nextchar();
	return lex();		/* call again to get next token */
	}
else {
	/* some other character.
	 */
	int ch = currch;

	nextchar();
	if (parse_debug)
		printf("char : %c\n",ch);
	return ch;
	}
}


/* init_lexer
 * Reset lexer variables
 */
static void init_lexer(int filed)
{
fd = filed;
currch = ' ';
lineno = 0;
strcpy(filename,"unknown");
}


/* reservedword
 * Looks up a string in the reserved word table, rwtab.
 * This table must be defined in the parser into which this file is included.
 */
static int reservedword(char *s)
{
int i;

for(i=0; i<sizeof(rwtab)/sizeof(rwtab[0]); i++)
	if (strequ(rwtab[i].word,s))
		return rwtab[i].tok;
return -1;
}

