/* misc.c
 *
 * Assorted functions used by both the client and server
 */
#include "sim.h"

static char *time_string(void);

/* strdupe
 * Duplicate a string (for systems that don't have strdup)
 */
char *strdupe(char *s)
{
char *ns = smalloc(strlen(s)+1);

return strcpy(ns,s);
}


static FILE *logfile = NULL;

/* open_log_file
 * Open a file to record errors and messages to
 */
int open_log_file(char *filename)
{
if ((logfile = fopen(filename,"a")) == NULL)
	return 0;
fprintf(logfile,"[%s] *** STARTING LOG ***\n",time_string());
return 1;
}


/* logerror
 * General error-reporting function
 */
void logerror(char *fmt, ...)
{
va_list ap;
char buf[1000];

va_start(ap,fmt);
vsprintf(buf,fmt,ap);
va_end(ap);
fprintf(stderr,"Error : %s\n",buf);
fflush(stderr);
if (logfile) {
	fprintf(logfile,"[%s] Error : %s\n",time_string(),buf);
	fflush(logfile);
	}
}


/* logmessage
 * General purpose output and logging function
 */
void logmessage(char *fmt, ...)
{
va_list ap;
char buf[1000];

va_start(ap,fmt);
vsprintf(buf,fmt,ap);
va_end(ap);
fprintf(stdout,"%s\n",buf);
fflush(stdout);
if (logfile) {
	fprintf(logfile,"[%s] %s\n",time_string(),buf);
	fflush(logfile);
	}
}


/* time_string
 * Returns a string containing the current date and time, for logging purposes
 */
static char *time_string(void)
{
static char tbuf[256];
time_t tm = time(NULL);

strftime(tbuf,256,"%d/%b/%Y:%H:%M:%S",localtime(&tm));
return tbuf;
}


/* smalloc
 * malloc() memory or die trying
 */
void *smalloc(size_t n)
{
void *mem = malloc(n);
if (!mem && n) {
	logerror("malloc(%d) failed! Shutting down.",n);
	exit(127);
	}
return mem;
}


/* scalloc
 * calloc() memory or die trying
 */
void *scalloc(size_t c, size_t n)
{
void *mem = calloc(c,n);
if (!mem && c && n) {
	logerror("calloc(%d,%d) failed! Exiting.",c,n);
	exit(127);
	}
return mem;
}


/* srealloc
 * realloc() memory or die trying
 */
void *srealloc(void *m, size_t n)
{
void *mem = realloc(m,n);
if (!mem && n) {
	logerror("realloc(%p,%d) failed! Exiting.",m,n);
	exit(127);
	}
return mem;
}


/* sfork
 * fork() or die trying
 */
pid_t sfork(void)
{
pid_t p = fork();

if (p == -1) {
	logerror("fork() failed - %s! Exiting.",strerror(errno));
	exit(127);
	}
return p;
}


/* memdupe
 * Duplicate a chunk of memory of given size
 */
void *memdupe(void *mem, int size)
{
void *nm = smalloc(size);

return memcpy(nm,mem,size);
}


/* timenow
 * Returns the current time with microsecond precision
 */
double timenow(void)
{
struct timeval tv;
struct timezone tz;

gettimeofday(&tv,&tz);
return tv.tv_sec + tv.tv_usec/1000000.0;
}


/* error_string
 * Returns a short error message for an error number
 */
char *error_string(int ern)
{
switch(ern) {
case err_nopath:
	return "Path does not exist";
case err_blocked:
	return "Path is blocked";
case err_notfound:
	return "Item not found";
case err_down:
	return "Remote server down";
case err_invfull:
	return "Inventory full";
case err_donthave:
	return "Object not in inventory";
case err_nothere:
	return "Person not nearby";
case err_nofile:
	return "Data not found";
case err_noaction:
	return "Cannot do that";
case err_execfailed:
	return "Error executing action!";
case err_notonline:
	return "Player not logged in";
case err_cantget:
	return "Cannot get object";
case err_parse:
	return "Parse error";
case err_maxdata:
	return "Data quota exceeded";
case err_avatar:
	return "Invalid avatar shape";
default:
	return "Unknown error";
	}
}


/* parse_string
 * Break a string up into words. Handles long words with spaces surrounded
 * by quotes, as long as there are spaces between words.
 */
void parse_string(char *cmd, char ***cmdv, int *cmdc)
{
static char *vec[256];
int i = 0;

while(1) {
	int quote = 0;

	while(*cmd == ' ' || *cmd == '\t' || *cmd == '\n')
		cmd++;		/* skip spaces before a word */
	if (*cmd == 0) break;	/* done */
	if (*cmd == '\"') { cmd++; quote = 1; }	/* found a quote */
	vec[i++] = cmd;		/* record start of word */
	if (quote)
		while(*cmd != '\"')
			cmd++;	/* skip until end quote */
	else
		while(*cmd && *cmd != ' ' && *cmd != '\t' && *cmd != '\n')
			cmd++;	/* skip until whitespace */
	if (*cmd == 0) break;	/* this was the last word */
	*cmd = 0;		/* mark end of word */
	cmd++;
	}
vec[i] = NULL;
*cmdv = vec;
*cmdc = i;
}


/* itoa
 * Converts an integer to a string
 */
char *itoa(int i)
{
static char buf[30];

sprintf(buf,"%d",i);
return buf;
}


/* ftoa
 * Converts a double to a string
 */
char *ftoa(double f)
{
static char buf[400];

sprintf(buf,"%f",f);
return buf;
}


/* count_chars
 * Counts the number of times a character occurs in a string
 */
int count_chars(char *txt, char c)
{
int total=0;

for(; *txt; txt++)
	if (*txt == c) total++;
return total;
}


/* isinteger
 * Returns non-zero if a string consists entirely of digits
 */
int isinteger(char *s)
{
while(*s)
	if (!isdigit(*(s++)))
		return 0;
return 1;
}


/* basenam
 * Returns the non-directory part of a file path 
 */
char *basenam(char *fn)
{
char *bn = fn+strlen(fn)-1;

while(bn >= fn && *bn != '/')
	bn--;
if (*bn == '/')
	return bn+1;
return fn;
}


/* dirnam
 * returns the directory part of a file path, or "."
 */
char *dirnam(char *fn)
{
static char buf[256];
int slpos;		/* position of last / */
int i;

for(slpos=strlen(fn)-1; fn[slpos] != '/'; slpos--)
	if (slpos < 0)
		return ".";
for(i=0; i<slpos; i++)
	buf[i] = fn[i];
buf[i] = 0;
return buf;
}


/* swapdbl
 * Swap two doubles
 */
void swapdbl(double *a, double *b)
{
double tmp = *a;
*a = *b;
*b = tmp;
}


/* current_time
 * Returns the current time with (supposed) microsecond accuracy
 */
double current_time(void)
{
struct timeval tv;
struct timezone tz;

gettimeofday(&tv,&tz);
return tv.tv_sec + tv.tv_usec/1000000.0;
}


/* delete_dir
 * Deletes everything in a directory. 
 */
void delete_dir(char *dir)
{
DIR *dlist;
struct dirent *dent;

if ((dlist = opendir(dir)) == NULL)
	return;
while((dent = readdir(dlist)))
	if (dent->d_ino && dent->d_name[0] != '.') {
		char fullname[256];

		sprintf(fullname,"%s/%s",dir,dent->d_name);
		if (unlink(fullname) < 0)
			logerror("Error deleting %s",fullname);
		}
closedir(dlist);
}


/* open_data_file
 * Create a file for storing data, making any necessary directories on the
 * way. Returns the fd for writing to the file. 
 */
int open_data_file(char *file)
{
char *dirbuf = dirnam(file);
struct stat stbuf;
int fd;

/* Create data dir if necessary
 */
if (stat(dirbuf,&stbuf) == -1) {
	/* Doesn't exist at all
	 */
	mkdir(dirbuf,0700);
	}
else if ((stbuf.st_mode & S_IFMT) != S_IFDIR) {
	/* Is a file!
	 */
	if (unlink(dirbuf) == -1) {
		logerror("Couldn't remove file %s for creating data dir",
			 dirbuf);
		return -1;
		}
	mkdir(dirbuf,0700);
	}
/* Create file
 */
if ((fd = creat(file,0700)) == -1) {
	logerror("Couldn't create file %s for storing data",file);
	return -1;
	}
return fd;
}


/* expandstr
 * Convert special characters (tabs, returns, etc..) in a string to their
 * \x notation.
 */
char *expandstr(char *s)
{
static char buf[1024];
int i;

for(i=0; *s; s++) {
	if (*s == '\\')      { strcpy(buf+i,"\\\\"); i+=2; }
	else if (*s == '\n') { strcpy(buf+i,"\\n"); i+=2; }
	else if (*s == '\t') { strcpy(buf+i,"\\t"); i+=2; }
	else if (*s == '\"') { strcpy(buf+i,"\\\""); i+=2; }
	else if (*s == '\'') { strcpy(buf+i,"\\\'"); i+=2; }
	else buf[i++] = *s;  
	}
buf[i] = 0;
return buf;
}


/* need_quotes
 * Returns 1 if a string has spaces in it or is 0 length
 */
int need_quotes(char *s)
{
if (*s == 0) return 1;
for(;*s;s++)
	if (*s==' ' || *s=='\n' || *s=='\t' || *s=='\r')
		return 1;
return 0;
}


/* get_home_dir
 * returns the user's home directory
 */
char *get_home_dir(void)
{
char *home;
char *login;
struct passwd *pe;

if ((home = getenv("HOME")))
	return home;
else if ((login = getlogin()) && (pe = getpwnam(login)))
	return pe->pw_dir;
else
	return ".";
}


/* tabs
 * Returns a string of t tabs
 */
char *tabs(int t)
{
static char tbuf[256];
int i;

for(i=0; i<t; i++)
	tbuf[i] = '\t';
tbuf[i] = 0;
return tbuf;
}


/* strip_return
 * Remove the return char from a string
 */
void strip_return(char *s)
{
char *last = s + strlen(s) - 1;

if (*last == '\n') *last = 0;
}


/* nrand
 * Returns a random number between 0 and 1
 */
double nrand(void)
{
return (rand()%10001)/10000.0;
}


/* set_signals
 * Set handlers for the interrupt & terminate signals. 
 */
void set_signals(void (*sfunc)(int s))
{
signal(SIGINT,sfunc);
signal(SIGTERM,sfunc);
}


/* concat_vec
 * Concatenate a vector of strings into one large string, separated by spaces
 */
char *concat_vec(char **vecv, int vecc)
{
static char buf[1024], *bufp = buf;
int i;

if (vecc == 0) return "";
for(i=0; i<vecc; i++) {
	sprintf(bufp," %s",vecv[i]);
	bufp += strlen(vecv[i]+1);
	}
return buf+1;	/* skip leading space */
}


/* send_message
 * General purpose function for sending client requests or server commands
 */
void send_message(int fd, int type, va_list ap, mesgargs ma[], int mac)
{
char mbuf[2048], *mptr = mbuf;
int i;
int *args = NULL;

sprintf(mptr,"%d",type);  mptr += strlen(mptr);

for(i=0; i<mac; i++)
	if (ma[i].mtype == type)
		args = ma[i].margs;
if (!args) {
	logerror("No arg list for message of type %d!",type);
	return;
	}

for(i=0; args[i]; i++) {
	switch(args[i]) {
	case arg_str: {
		char *strarg = va_arg(ap,char *);
		if (need_quotes(strarg)) sprintf(mptr," \"%s\"",strarg);
		else			 sprintf(mptr," %s",strarg);
		mptr += strlen(mptr);
		break;
		}
	case arg_dbl: {
		double dblarg = va_arg(ap,double);
		sprintf(mptr," %f",dblarg);
		mptr += strlen(mptr);
		break;
		}
	case arg_int: {
		int intarg = va_arg(ap,int);
		sprintf(mptr," %d",intarg);
		mptr += strlen(mptr);
		break;
		}
	case arg_strs: {
		char **wordarg = va_arg(ap,char **);
		int wordc;
		int j;

		if (args[++i] != arg_int) {
			logerror("No word count after arg of type arg_strs!");
			return;
			}
		wordc = va_arg(ap,int);
		for(j=0; j<wordc; j++) {
			if (need_quotes(wordarg[j]))
				sprintf(mptr," \"%s\"",wordarg[j]);
			else
				sprintf(mptr," %s",wordarg[j]);
			mptr += strlen(mptr);
			}
		break;
		}

		} /* end switch */
	} /* end for */
fdputs(fd,mbuf);
}


/* directory_size
 * Count up the total size of all files in a directory (and it's subdirs)
 */
long directory_size(char *d)
{
long c = 0;
DIR *dlist;
struct dirent *dent;

if ((dlist = opendir(d))) {
	while((dent = readdir(dlist)))
		if (dent->d_ino && dent->d_name[0] != '.') {
			char filename[256];
			struct stat stbuf;

			sprintf(filename,"%s/%s",d,dent->d_name);
			if (stat(filename,&stbuf) != -1) {
				if ((stbuf.st_mode & S_IFMT) == S_IFDIR)
					c += directory_size(filename);
				else 
					c += stbuf.st_size;
				}
			else
				logerror("Could check size of %s",filename);
			}
	closedir(dlist);
	}
return c;
}

#ifdef NO_STRERROR
/* strerror
 * Create an error message string, for systems that don't have a proper
 * strerror() function  (are you listening, Sun?)
 */
char *strerror(int en)
{
static char buf[256];
sprintf(buf,"error %d",en);
return buf;
}
#endif
