/* net.c
 *
 * Functions for creating and using tcp/ip network sockets.
 */
#include "sim.h"

extern int is_sim_server;

/* findmyhostname
 * Find the full internet address of this host.
 * BUG - this doesn't work properly yet. For some reason, on alphas running
 * OSF I only get back the host name, not the full name & domain.
 */
char *findmyhostname(void)
{
static char namebuf[256] = "";
struct hostent *addr, *name;

if (namebuf[0] != '\0') {
	/* already found.
	 */
	return namebuf;
	}
gethostname(namebuf,256);
if (strchr(namebuf,'.')) {
	/* Full, sane host name returned
	 */
	return namebuf;
	}
else {
	/* For some strange reason, only the first part of the host name
	 * was returned. Call nslookup to get the full name.
	 */
	char cmdbuf[256];	/* nslookup command string */
	FILE *nsfp;		/* output from nslookup */
	char nsline[256];	/* line from nslookup output */

	sprintf(cmdbuf,"nslookup %s",namebuf);
	if ((nsfp = popen(cmdbuf,"r")) == NULL) {
		logerror("nslookup not found - error getting this machine's "
			 "host name.");
		return namebuf;
		}

	while(fgets(nsline,256,nsfp) != NULL) {
		char *item;
		char *val;

		if ((item = strtok(nsline," \t\n")) == NULL ||
		    (val  = strtok(NULL," \t\n")) == NULL)
			continue;	/* empty or strange line */
		if (strequ(item,"Name:")) {
			pclose(nsfp);
			strcpy(namebuf,val);
			return namebuf;
			}
		}
	pclose(nsfp);
	logerror("nslookup didn't find this machine's full address");
	return namebuf;
	}
}


/* findmyip
 * Returns the ip number of the host this program is running on
 */
struct in_addr findmyip()
{
char *nm = findmyhostname();
struct in_addr ip = *(struct in_addr *)gethostbyname(nm)->h_addr;

return ip;
}


/* open_tcp_dest
 * Open a tcp socket for accepting new connections on
 */
int open_tcp_dest(int port)
{
int sock;
struct sockaddr_in addr;
int one=1;

/* Create a new tcp/ip socket
 */
sock = socket(AF_INET,SOCK_STREAM,0);
if (sock < 0)
	return -1;

/* Assign a port to the new socket
 */
addr.sin_family      = AF_INET;
addr.sin_port        = htons(port);
addr.sin_addr.s_addr = INADDR_ANY;

if (bind(sock,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
	close(sock);
	return -1;
	}

/* Allow re-use of socket after program ends
 */
/* setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(one)); */

listen(sock,8);

/*logmessage("opened socket on port %d",port);*/
return sock;
}


/* new_connection
 * Accept a new connection on a master socket, and find out the name
 * of the host it is from.
 */
int new_connection(int sfd, char *hostname, struct in_addr *addr)
{
struct sockaddr_in iaddr;	/* address of connecting host */
int size = sizeof(iaddr);
struct hostent *hent;		/* info on name of connecting host */
int fd;				/* fd of new connection */

fd = accept(sfd,(struct sockaddr *)&iaddr,&size);
*addr = iaddr.sin_addr;

if (hostname) {
	/* Look up name of connecting host
	 */
	hent = gethostbyaddr((char *)&addr->s_addr,sizeof(struct in_addr),
			     AF_INET);
	if (hent)
		strcpy(hostname,hent->h_name);
	else {
		int ip = htonl(addr->s_addr);
		sprintf(hostname,"%d.%d.%d.%d",(ip&0xff000000)>>24,
			(ip&0xff0000)>>16,(ip&0xff00)>>8,(ip&0xff));
		}
	}

/*logmessage("got connection from %s",dotted_notation(*addr));*/
return fd;
}


/* open_tcp_send
 * Open a tcp connection to a specified host and port
 */
int open_tcp_send(char *host, int port)
{
struct hostent *dest;		/* ip address of host */

/* Look up ip address of host
 */
dest = gethostbyname(host);
if (!dest)
	return -1;

return open_tcp_send_ip(*(struct in_addr *)dest->h_addr,port);
}


/* open_tcp_send_ip
 * Open a tcp connection to a specified ip address and port
 */
int open_tcp_send_ip(struct in_addr ip, int port)
{
int sock;			/* fd of new connection */
struct sockaddr_in addr;


/* Create a new tcp/ip socket
 */
sock = socket(AF_INET,SOCK_STREAM,0);
if (sock < 0)
	return -1;

/* Bind address and port to socket
 */
addr.sin_family = AF_INET;
addr.sin_port   = htons(port);
addr.sin_addr   = ip;

/* Connect to host
 */
if (connect(sock,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
	close(sock);
	return -1;
	}

/*logmessage("connected to host %s port %d",dotted_notation(ip),port);*/
return sock;
}


/* send_version
 * Send this programs version number to a server, and check that it matches.
 * Returns 1 on success, 0 otherwise.
 */
int send_version(int fd)
{
char resp[30];

fdprintf(fd,"%s\n",SIM_VERSION);
if (fdgetw(fd,resp,30) == NULL) {
	logerror("Error reading version response");
	return 0;
	}
if (!strequ(resp,"ok")) {
	logerror("Version mismatch");
	return 0;
	}
return 1;
}


/* check_version
 * Receives a version number, checks if it matches, and sends back an
 * appropriate response.
 * Returns 1 if ok, 0 otherwise
 */
int check_version(int fd)
{
char vers[30];

if (fdgetw(fd,vers,30) == NULL) {
	logerror("Error reading version");
	return 0;
	}
if (!strequ(vers,SIM_VERSION)) {
	logerror("Version mismatch");
	fdprintf(fd,"fail\n");
	return 0;
	}
fdprintf(fd,"ok\n");
return 1;
}


/* send_data
 * Used for concurrent data transfer, so should be called from a forked
 * process. Opens a new socket on port, waits for a connection, then
 * transfers size bytes from fd to the new connection.
 * BUG - should only wait around a few mins before dying
 */
int send_data(int port, int fd, int size)
{
int ifd;		/* initial port fd */
int cfd;		/* new connection fd */
struct in_addr addr;	/* address of connector */
fd_set cfds;		/* for waiting on connection */
struct timeval tv;	/* time to wait for connection */
int rv;

if ((ifd=open_tcp_dest(port)) == -1) {
	logerror("Couldn't create socket for sending data");
	return 0;
	}
FD_ZERO(&cfds);
FD_SET(ifd,&cfds);
tv.tv_usec = 0; tv.tv_sec = 60;
if (select(ifd+1,&cfds,NULL,NULL,&tv) != 1) {
	logerror("Didn't get connection on port %d to send data after %d secs",
		 port,tv.tv_sec);
	return 0;
	}
cfd = new_connection(ifd,NULL,&addr);
rv = copy_data(fd,cfd,size);
close(cfd);
close(ifd);
return rv;
}


/* recv_data
 * Opens a connection to the given address and port, then reads size
 * bytes of data and writes them to fd. Returns 0 on error, or 1 if successful
 */
int recv_data(struct in_addr addr, int port, int fd, int size)
{
int cfd;		/* fd for connection to other host */
int trys=0;		/* number of failed attempts to connect */
int rv;

while((cfd = open_tcp_send_ip(addr,port)) == -1) {
	sleep(2);
	if (trys++ > 10) return 0;
	}
rv = copy_data(cfd,fd,size);
close(cfd);
return rv;
}


/* alloc_port
 * Silly function for cycling through port numbers for sending data. Returns
 * a number from 1-10
 */
int alloc_port(void)
{
static int i=0;

i = (i+1)%10;
/*
return SIM_DATA_PORT+(is_sim_server ? 10 : 0)+i;
*/
return i+1;
}


/* dotted_notation
 * Converts an ip-number to a string in dotted-decimal notation
 */
char *dotted_notation(struct in_addr ip)
{
static char buf[20];
u_long n = ntohl(ip.s_addr);

sprintf(buf,"%d.%d.%d.%d",(n&0xff000000)>>24,(n&0x00ff0000)>>16,
			  (n&0x0000ff00)>>8, (n&0x000000ff));
return buf;
}


/* send_password
 * Send password, and check that an 'ok' was sent back.
 */
int send_password(int fd, char *pass)
{
char resp[20];

fdprintf(fd,"%s\n",pass);
if (fdgetw(fd,resp,20) == NULL) {
	logerror("Error reading password response");
	return 0;
	}
if (!strequ(resp,"ok")) {
	logerror("Password rejected!");
	return 0;
	}
return 1;
}


/* check_password
 * Reads and checks a password from another server 
 */
int check_password(int fd, char *pass)
{
char other[256];		/* password from other server */

if (fdgetw(fd,other,256) == NULL) {
	logerror("Error reading password");
	return 0;
	}
if (!strequ(other,pass)) {
	fdprintf(fd,"fail\n");
	logerror("Rejecting server with wrong password");
	return 0;
	}
fdprintf(fd,"ok\n");
return 1;
}

