%{
/* s3dt.y
 *
 * Yacc script for parsing sim 3d descriptions into s3dt objects.
 * Functions for rendering and manipulating s3dt objects.
 */
#include "s3dt/s3dt.h"
#include "s3dt/s3view.h"
#include "sim.h"
#include "3d.h"
#include "graphics.h"
#include "client.h"

static void new_shape(void);

#define SCALE_FACTOR 4.0
%}

%union {
	double num;
	char string[256];
	}
%token <num> NUM
%token <string> STRING
%token END,POLY,SPHERE,BOX,CONE,TUBE,COLL,POS,NAME,COLOUR,DISC
%token ROTATE,CENTRE,RADIUS,HEIGHT,POINTS,CORNERS,SURFACE,SCALE,TEXTURE,CMT
%type <num> surftype
%%
input		:	END
			{ return -1;
			}
		|	shape_trans
			{ return 0;
			}
		;

shape_in_coll	:	{ /* Push current collection shape onto stack
			   */
			  coll[collp].sha = sha;
			  coll[collp++].coll = no;
			}
			shape_trans
			{ /* Add new shape to collection
			   */
			  S3AddChildObject(no,coll[collp-1].coll);

			  /* Pop collection and attribs from stack
			   */
			  no = coll[--collp].coll;
			  sha = coll[collp].sha;
			}
		;

shape_trans	:	shape
			{ S3ScaleObject3(no, sha.sx, sha.sy, sha.sz);
			  S3SetRotateAbout(no, sha.rang, sha.rx, sha.ry, sha.rz,
					   0,0,0);
			  S3SetObjectPos(no, sha.cx, sha.cy, sha.cz);
			}

shape		:	poly
		|	sphere
		|	box
		|	cone
		|	tube
		|	disc
		|	coll
		;


/*********************** polygon **********************
 */
poly		:	POLY 
			{ no = S3CreateObject("",0);
			  S3SetObjectName(no,"");
			  new_shape();
			}
			'{' polyattribs '}'
		;

polyattribs	:	polyattribs polyattrib
		|	polyattribs shapeattrib
		|
		;

polyattrib	:	POINTS
			{ S3GetObjectPointC(no,&init_ptc);
			  ptc = 0;
			}
			polypoints
			{ int pt[256];
			  int i;

			  for(i=0; i<ptc; i++)
			    pt[i] = i+init_ptc;
			  S3AddPolygon(no,pt,ptc,31,def_ftype);
			}
		;

polypoints	:	polypoints polypoint
		|
		;

polypoint	:	NUM NUM NUM
			{ S3AddPoint(no,$1,$2,$3);
			  ptc++;
			}


/*********************** sphere **********************
 */
sphere		:	SPHERE 
			{ no = S3CreateSphere(31,def_ftype | pt_oneside);
			  S3SetObjectName(no,"");
			  new_shape();
			}
			'{' sphereattribs '}'
		;

sphereattribs	:	sphereattribs sphereattrib
		|	sphereattribs shapeattrib
		|
		;

sphereattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $3; sha.cz += $4;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sy *= $2*2; sha.sz *= $2*2;
			}
		;


/*********************** box **********************
 */
box		:	BOX 
			{ no = S3CreateCube(31,def_ftype | pt_oneside);
			  S3SetObjectName(no,"");
			  new_shape();
			}
			'{' boxattribs '}'
		;

boxattribs	:	boxattribs boxattrib
		|	boxattribs shapeattrib
		|
		;

boxattrib	:	CORNERS NUM NUM NUM  NUM NUM NUM
			{ double x1=$2,y1=$3,z1=$4, x2=$5,y2=$6,z2=$7;

			  if (x2 < x1) swapdbl(&x1,&x2);
			  if (y2 < y1) swapdbl(&y1,&y2);
			  if (z2 < z1) swapdbl(&z1,&z2);
			  sha.sx *= x2-x1; sha.sy *= y2-y1; sha.sz *= z2-z1;
			  sha.cx += (x2+x1)/2; sha.cy += (y2+y1)/2;
			  sha.cz += (z2+z1)/2;
			}
		;


/*********************** cone **********************
 */
cone		:	CONE 
			{ no = S3CreateCone(31,def_ftype | pt_oneside);
			  S3SetObjectName(no,"");
			  new_shape();
			}
			'{' coneattribs '}'
		;

coneattribs	:	coneattribs coneattrib
		|	coneattribs shapeattrib
		|
		;

coneattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $3; sha.cz += $4;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sy *= $2*2;
			}
		|	HEIGHT NUM
			{ sha.sz *= $2;
			}
		;


/*********************** tube **********************
 */
tube		:	TUBE 
			{ no = S3CreateTube(31,def_ftype | pt_oneside);
			  S3SetObjectName(no,"");
			  S3MovePoints(no,0,0,-0.5);
			  new_shape();
			}
			'{' tubeattribs '}'
		;

tubeattribs	:	tubeattribs tubeattrib
		|	tubeattribs shapeattrib
		|
		;

tubeattrib	:	CENTRE NUM NUM NUM
			{ sha.cx = $2; sha.cy = $3; sha.cz = $4;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sy *= $2*2;
			}
		|	HEIGHT NUM
			{ sha.sz *= $2;
			}


/*********************** disc **********************
 */
disc		:	DISC 
			{ no = S3CreateDisc(31,def_ftype);
			  S3SetObjectName(no,"");
			  new_shape();
			}
			'{' discattribs '}'
		;

discattribs	:	discattribs discattrib
		|	discattribs shapeattrib
		|
		;

discattrib	:	CENTRE NUM NUM NUM
			{ sha.cx += $2; sha.cy += $3; sha.cz += $4;
			}
		|	RADIUS NUM
			{ sha.sx *= $2*2; sha.sz *= $2*2;
			}
		;


/*********************** coll **********************
 */
coll		:	COLL 
			{ no = S3CreateObject("",0);
			  S3SetObjectName(no,"");
			  new_shape();
			}
			'{' collattribs '}'
		;

collattribs	:	collattribs shape_in_coll
		|	collattribs shapeattrib
		|
		;


shapeattrib	:	NAME STRING
			{ S3SetObjectName(no,$2);
			}
		|	POS NUM NUM NUM
			{ sha.cx += $2; sha.cy += $3; sha.cz += $4;
			}
		|	COLOUR NUM NUM NUM
			{ int col = S3ClosestColour($2*256,$3*256,$4*256);
			  S3SetObjectColour(no,col);
			}
		|	ROTATE NUM  NUM NUM NUM
			{ sha.rx = $3; sha.ry = $4; sha.rz = $5;
			  sha.rang = $2;
			}
		|	SURFACE surftype
			{ if ($2 == sft_glass) {
			    S3SetObjectPType(no,pt_glass,pt_glass);
			    }
			  else if ($2 == sft_shiny) {
			    S3SetObjectPType(no,pt_shiny,pt_shiny);
			    }
			}
		|	SCALE NUM NUM NUM
			{ sha.sx *= $2; sha.sy *= $3; sha.sz *= $4;
			}
		|	TEXTURE NUM NUM '{' texture_strings '}'
		|	CMT STRING
		;

texture_strings	:	texture_strings STRING
		|
		;

surftype	:	NUM
			{ $$ = $1; }
		|	STRING
			{ if (strequ($1,"flat"))
			    $$ = sft_flat;
			  else if (strequ($1,"glass"))
			    $$ = sft_glass;
			  else if (strequ($1,"shiny"))
			    $$ = sft_shiny;
			  else {
			    yyerror("Unknown surface type");
			    return 1;
			    }
			}
		;

%%
static Display *display;	/* connection to X server */

static S3objtab *ot;		/* hash table of s3dt objects */
static S3object *no;		/* object currently being read */
static int init_ptc;		/* point count at start of polygon */
static int ptc;			/* no. of points in polygon */
static shapeattribs sha;	/* attribs of current shape and collection */

static struct {			/* stack of collection attribs and pointers */
	shapeattribs sha;
	S3object *coll;
	} coll[256];
static int collp;		/* position in stack */

S3view *vw;			/* direction and position of 3d view */
int def_ftype;			/* default face type */

static char *last_click_name;	/* name of last object clicked on */
static int last_click_button;	/* last button clicked */

static S3object *hs = NULL;	/* currently highlighted shape */

static reserved rwtab[] = {	"end",END,"poly",POLY,
				"sphere",SPHERE,"box",BOX,"cone",CONE,
				"tube",TUBE,"disc",DISC,"coll",COLL,
				"pos",POS,"name",NAME,"colour",COLOUR,
				"rotate",ROTATE,"centre",CENTRE,
				"radius",RADIUS,"height",HEIGHT,
				"points",POINTS,"corners",CORNERS,
				"surface",SURFACE,"scale",SCALE,
				"texture",TEXTURE,"cmt",CMT };
#define error yyerror
#define lex yylex
#define lval yylval

#include "lexer.h"


/* new_shape
 * Reset shape attributes
 */
static void new_shape(void)
{
sha.cx = sha.cy = sha.cz = 0.0;
sha.sx = sha.sy = sha.sz = 1.0;
sha.rx = sha.ry = sha.rz = 0.0;
sha.rang = 0.0;
}


/* supported_3d
 * Returns the supported 3d level.
 */
int supported_3d(void)
{
return lvl_flat;
}


/* parse_3d
 * Parse one 3d object and add it to the object table. 
 */
int parse_3d(int filed)
{
int rv;

init_lexer(filed);
collp = 0;
rv = yyparse();
if (rv == 0) {
	char *oname;
	S3GetObjectName(no,&oname);
	if (!oname || oname[0] != '_')
		S3AddObjectAndChildren(ot,no);
	}
return rv;
}


/* parse_3ds
 * Parse 3d objects and add them to the hash table until an 'end' is read.
 */
int parse_3ds(int filed)
{
int rv;

init_lexer(filed);
collp = 0;
while((rv = yyparse()) == 0) {
	char *oname;
	S3GetObjectName(no,&oname);
	if (!oname || oname[0] != '_')
		S3AddObjectAndChildren(ot,no);
	}
if (rv == 1)
	return 1;
else
	return 0;
}


/* init_3d
 * Create a window of the given size, and init the s3dt view in it.
 * Initialize the s3dt object table.
 */
void init_3d(char *display_name, Window par, int x, int y, int w, int h)
{
Colormap cmap;
XColor def[8];
Window win;
userprefs *cp;

display = XOpenDisplay(display_name);
win = XCreateSimpleWindow(display,par,x,y,w,h,0,0,0);
XSelectInput(display,win,ButtonPressMask | ExposureMask);
XMapWindow(display,win);
XFlush(display);

def[0].red = 65535; def[0].green = 65535; def[0].blue = 65535;	/* white */
def[1].red = 0    ; def[1].green = 65535; def[1].blue = 0;	/* green */
def[2].red = 20000; def[2].green = 20000; def[2].blue = 65535;	/* sky blue */
def[3].red = 65535; def[3].green = 0    ; def[3].blue = 0;	/* red */
def[4].red = 65535; def[4].green = 57343; def[4].blue = 49151;	/* brown */
def[5].red = 65535; def[5].green = 65535; def[5].blue = 0;	/* yellow */
def[6].red = 65535; def[6].green = 0    ; def[6].blue = 65535;	/* purple */
def[7].red = 65535; def[7].green = 56320; def[7].blue = 56320;	/* pink */
cmap = S3CreateSimpleColourmap(display,def);
cp = get_prefs(NULL);
if (cp->default_3d == 0) {
	def_ftype = pt_plane | pt_lighted | pt_shaded;
	vw = S3CreateView(display,win,rm_wire);
	}
else if (cp->default_3d == 1) {
	def_ftype = pt_plane | pt_lighted | pt_shaded;
	vw = S3CreateView(display,win,rm_plane);
	S3SetViewBackColour(vw,95);
	}
else {
	def_ftype = pt_plane | pt_lighted | pt_shaded;
	vw = S3CreateView(display,win,rm_zbufshm);
	S3SetViewBackColour(vw,95);
	}
S3SetViewEyeDistance(vw,0.1);
S3SetViewScale(vw,w*SCALE_FACTOR);
S3AddViewLight(vw,0,0,0,0.7);			/* light at player's pos */
S3AddViewLight(vw,10000,10000,10000,0.3);	/* constant light */
XSetWindowColormap(display,win,cmap);
ot = S3CreateObjectTable(tt_name);
XSetWindowColormap(display,par,cmap);
XFlush(display);
}


/* new_3d_level
 * Change the 3d graphics level 
 */
void new_3d_level(void)
{
switch(get_prefs(NULL)->default_3d) {
case 0:
	S3ChangeView(display,vw,rm_wire);
	S3SetViewBackColour(vw,0);
	break;
case 1:
	S3ChangeView(display,vw,rm_plane);
	S3SetViewBackColour(vw,95);
	break;
case 2:
	S3ChangeView(display,vw,rm_zbufshm);
	S3SetViewBackColour(vw,95);
	break;
	}
S3RenderView(display,vw,ot);
}


/* threed_fd
 * Returns the fd for the 3d view's display
 */
int threed_fd(void)
{
return ConnectionNumber(display);
}


/* move_3d
 * Move an s3dt object by dx,dy,dz
 */
void move_3d(char *name, double dx, double dy, double dz)
{
S3object *obj = S3LookupObject(ot,name);

S3DeltaObjectPos(obj,dx,dy,dz);
}


/* del_3d
 * Remove a 3d object and all it's parts
 */
void del_3d(char *name)
{
S3object *obj = S3LookupObject(ot,name);

if (obj == hs)
	unhighlight_3d();
S3DeleteObjectAndChildren(ot,obj);
}


/* clear_3d
 * Delete all 3d objects in the scene
 */
void clear_3d(void)
{
S3ClearObjectTable(ot);
}


static int redraw_needed = 0;	/* should render_3d actually render? */

/* render_3d
 * Tell s3dt to draw the current scene
 */
int render_3d(plview *plv)
{
double x, y, z;		/* position of view */
double nx, ny, nz;	/* direction of view normal */

if (redraw_needed) {
	current_position(plv,&x,&y,&z);
	S3SetViewVrp(vw,x,y,z);
	S3MoveViewLight(vw,0,x,y,z);

	nx = cos(dtor(plv->ang))*cos(dtor(plv->tilt));
	ny = sin(dtor(plv->ang))*cos(dtor(plv->tilt));
	nz = sin(dtor(plv->tilt));
	S3SetViewN(vw,nx,ny,nz);
	S3SetViewV(vw,0,0,1);
	S3RenderView(display,vw,ot);
	redraw_needed = 0;
	return 1;
	}
return 0;
}


/* need_redraw
 * Indicate that a redraw should be one on the next call to render_3d()
 */
void need_redraw(void)
{
redraw_needed = 1;
}


/* check_3d
 * Poll for and handle events for the 3d view window 
 */
void check_3d(void)
{
while(XPending(display)) {
	XEvent event;

	XNextEvent(display,&event);
	if (event.type == Expose && !event.xexpose.count &&
	    event.xexpose.window == vw->win) {
		S3RefreshView(display,vw);
		}
	else if (event.type == ButtonPress &&
		 event.xbutton.window == vw->win) {
		last_click_button = event.xbutton.button;
		last_click_name   = onscreen_3d(event.xbutton.x,
						event.xbutton.y);
		}
	}
}


/* resize_3d
 * Change the size and position of the 3d view window
 */
void resize_3d(int x, int y, int w, int h)
{
XMoveResizeWindow(display,vw->win,x,y,w,h);
XRaiseWindow(display,vw->win);
S3SetViewSize(display,vw,w,h);
S3SetViewScale(vw,w*SCALE_FACTOR);
S3RenderView(display,vw,ot);
}


/* best_colour
 * Returns the pixel value of the colour which best matches the rgb values 
 * given.
 */
long best_colour(int red, int green, int blue)
{
return S3ClosestColourPixel(red*256,green*256,blue*256);
}


/* rotate_3d
 * Set the rotation of a shape to a given angle about some axis.
 * Rotation is about the centre of an object, or it's pos in the case of
 * a collection.
 */
void rotate_3d(char *name, double ang, double xa, double ya, double za)
{
S3object *obj = S3LookupObject(ot,name);
double px, py, pz;

S3GetObjectPos(obj,&px,&py,&pz);
S3SetRotateAbout(obj,ang,xa,ya,za,px,py,pz);
}


/* onscreen_3d
 * Returns the name of the shape at a given position in the 3d view 
 * window, or NULL
 */
char *onscreen_3d(int x, int y)
{
S3object *ob = S3ObjectAtPos(vw,ot,x,y);

if (ob) {
	/* Find top-level parent object
	 */
	S3object *par;
	char *nm;

	S3GetObjectParent(ob,&par);
	while(par) {
		ob = par;
		S3GetObjectParent(ob,&par);
		}
	S3GetObjectName(ob,&nm);
	return nm;
	}
return NULL;
}



/* highlight_3d
 * Highlight a named shape
 */
void highlight_3d(char *name)
{
unhighlight_3d();
if ((hs = S3LookupObject(ot,name)))
	S3HighLightObject(hs,127);
}


/* unhighlight_3d
 * Cancel highlighting on the currently highlighted shape
 */
void unhighlight_3d(void)
{
if (hs) {
	S3UnHighLightObject(hs);
	hs = NULL;
	}
}


/* colour_3d
 * Change the colour of a named 3d shape
 */
void colour_3d(char *name, int r, int g, int b)
{
S3object *obj = S3LookupObject(ot,name);

S3SetObjectColour(obj,S3ClosestColour(r*256,g*256,b*256));
}


/* shade_3d 
 * Change the surface type of a named 3d shape
 */
void shade_3d(char *name, int shd)
{
S3object *obj = S3LookupObject(ot,name);

if (shd == sft_flat)
	S3SetObjectPType(obj,0,pt_glass | pt_shiny);
else if (shd == sft_glass)
	S3SetObjectPType(obj,pt_glass,pt_glass);
else if (shd == sft_shiny)
	S3SetObjectPType(no,pt_shiny,pt_shiny);
}


/* get_3d_click
 * Returns 1 if the user has clicked in the 3d window since the last call
 */
int get_3d_click(char **n, int *b)
{
if (last_click_name) {
	*n = last_click_name;
	*b = last_click_button;
	last_click_name = NULL;
	return 1;
	}
return 0;
}


/* shutdown_3d
 * Called just before exit() to free up stuff
 */
void shutdown_3d(void)
{
S3FreeObjectTable(ot);
S3FreeView(display,vw);		/* free up shared mem */
}


/* boundingbox_3d
 * Compute the bounding box of the entire scene
 */
void boundingbox_3d(double *x1, double *y1, double *z1,
		    double *x2, double *y2, double *z2)
{
S3BoundingBoxTable(ot,x1,y1,z1,x2,y2,z2);
}


/* visible_3d
 * Make a shape visible or invisible
 */
void visible_3d(char *name, int vis)
{
S3object *obj = S3LookupObject(ot,name);

S3SetObjectInvisible(obj,!vis);
}

