#include "s3dt.h"
#include "s3dtint.h"
#include "s3view.h"

static XColor ReqTable[256];		/* ideal colours requested */
static unsigned long ColourTable[256];	/* pixel values of s3dt colours */

/* S3CreateDefaultColourmap
 * Allocate a colourmap containing the default s3dt colours
 */
Colormap S3CreateDefaultColourmap(Display *display)
{
XColor def[8];

def[0].red = 65535; def[0].green = 65535; def[0].blue = 65535;
def[1].red = 0    ; def[1].green = 65535; def[1].blue = 0;
def[2].red = 0    ; def[2].green = 0    ; def[2].blue = 65535;
def[3].red = 65535; def[3].green = 0    ; def[3].blue = 0;
def[4].red = 65535; def[4].green = 57343; def[4].blue = 49151;
def[5].red = 65535; def[5].green = 65535; def[5].blue = 0;
def[6].red = 65535; def[6].green = 0    ; def[6].blue = 65535;
def[7].red = 0    ; def[7].green = 65535; def[7].blue = 65535;
return S3CreateSimpleColourmap(display,def);
}


/* S3CreateSimpleColourmap
 * Create a colourmap containing 8 colours with 32 shades of each.
 */
Colormap S3CreateSimpleColourmap(Display *display, XColor def[])
{
XColor cmp[256];		/* complete colourmap */
int i;				/* colour number */

for(i=0; i<8; i++) {
	int j;			/* shade number */
	for(j=0; j<32; j++) {
		int k = i*32+j;	/* position in colour table */
		double scale = j / 31.0;

		cmp[k].red   = def[i].red * scale;
		cmp[k].green = def[i].green * scale;
		cmp[k].blue  = def[i].blue * scale;
		cmp[k].flags = DoRed | DoGreen | DoBlue;
		}
	}
return S3CreateColourmap(display,cmp);
}


/* S3CreateColourmap
 * Allocate a colourmap to the best of the abilities of the display
 */
Colormap S3CreateColourmap(Display *display, XColor cmp[])
{
XVisualInfo *visual, vinfo;	/* list of supported visuals */
int visualc;			/* count of visuals */
int screen_num;			/* screen number of display */
Window rootwin;			/* root window of display */
Colormap cm;			/* custom or default colourmap */
int i;

screen_num = DefaultScreen(display);
rootwin    = RootWindow(display,screen_num);

/* Copy the requested colours
 */
memcpy(ReqTable,cmp,256*sizeof(XColor));

/* Find the best visual
 */
visual = XGetVisualInfo(display,VisualNoMask,&vinfo,&visualc);

/* Try to allocate colours using the default colourmap
 */
cm = DefaultColormap(display,screen_num);
for(i=0; i<256; i++) {
	if (XAllocColor(display,cm,&cmp[i]) == 0) {
		printf("failed to alloc colour %d\n",i);
		break;
		}
	else
		ColourTable[i] = cmp[i].pixel;
	}
if (i == 256) {
	/* Got them all ok.
	 */
	printf("Using default colourmap\n");
	return cm;
	}
else {
	/* Free old colours
	 */
	XFreeColors(display,cm,ColourTable,i,0);
	}
	
/* Try to allocate a custom pseudocolour colourmap
 */
for(; visualc > 0; visualc--,visual++) {
	if (visual->class == PseudoColor && visual->depth >= 8) {
		/* Try to allocate a custom colourmap, and use that
		 */
		printf("Using pseudocolour colourmap\n");
		cm = XCreateColormap(display,rootwin,visual->visual,AllocAll);
		for(i=0; i<256; i++) {
			ColourTable[i] = cmp[i].pixel = i;
			XStoreColor(display,cm,&cmp[i]);
			}
		return cm;
		}
	}

fprintf(stderr,"S3CreateColourmap : None of the visuals of your display "
	"are suitable\n");
exit(10);
}


/* S3CreateWindowsColourmap
 * Create a colourmap suitable for 16-colours windows X servers
 */
Colormap S3CreateWindowsColourmap(Display *display)
{
static long colourtab[256] = {
0,0,0,0,0,0,0,0,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,15,15,15,15,0,0,0,0,0,0,
0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,10,10,10,10,10,10,10,10,0,0,0,0,0,0,0,0,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,12,12,12,12,12,12,12,12,0,0,0,0,0,0,0,0,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,9,9,9,9,9,9,9,9,0,0,0,0,0,0,0,0,7,7,7,7,7,7,7,7,7,7,7,7,8,8,
8,8,8,8,8,8,15,15,15,15,0,0,0,0,0,0,0,0,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,11,11,11
,11,11,11,11,11,0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,13,13,13,13,13,
13,13,13,0,0,0,0,0,0,0,0,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,14,14,14,14,14,14,14,14
};
int i;
int screen_num = DefaultScreen(display);

for(i=0; i<256; i++)
	ColourTable[i] = colourtab[i];
return DefaultColormap(display,screen_num);
}


/* S3SetColour
 * Set the drawing colour of a gc to the pixel value for the specified
 * 'ideal' colour.
 */
void S3SetColour(Display *display, GC gc, int colour)
{
XSetForeground(display,gc,ColourTable[colour]);
}


/* S3GetColour
 * Return the pixel value for the specified 'ideal' colour
 */
long S3GetColour(int colour)
{
if (colour < 0) colour = 0;
if (colour > 255) colour = 255;
return ColourTable[colour];
}


/* S3ClosestColour
 * Returns the index of the colour in the ideal colourmap closest to 
 * the given RBG value
 */
long S3ClosestColour(int red, int green, int blue)
{
int i;
double closest_rng = (65535.0*65535.0)*3.0;
long closest_index;

for(i=0; i<256; i++) {
	double rdiff = ReqTable[i].red - red;
	double gdiff = ReqTable[i].green - green;
	double bdiff = ReqTable[i].blue - blue;
	double rng = sqr(rdiff) + sqr(gdiff) + sqr(bdiff);

	if (rng < closest_rng) {
		closest_rng = rng;
		closest_index = i;
		}
	}
return closest_index;
}


/* S3ClosestColourPixel
 * Returns the pixel value of the colour closest to the rgb values
 * given.
 */
long S3ClosestColourPixel(int red, int green, int blue)
{
return S3GetColour(S3ClosestColour(red,green,blue));
}

