#include "s3dtint.h"
#include "s3dt.h"
#include "s3view.h"

static int S3IPutPix8(XImage *i, int x, int y, unsigned long v);
static int S3IPutPix24(XImage *i, int x, int y, unsigned long v);
static void S3AllocViewStuff(Display *display, S3view *nvw);
static void S3FreeViewStuff(Display *display, S3view *vw);

/* S3CreateView
 * Allocate a new view structure, and initialize it's GC and font
 */
S3view *S3CreateView(Display *display, Window win, S3rmodetype rm)
{
S3view *nvw;
XGCValues gcvals;			/* for setting GC options */
XWindowAttributes winattribs;		/* for querying window size */
char stipple_bits[] = { 0x01,0x02 };	/* for glass polygons */

/* Allocate new view and set attributes to sensible defaults
 */
nvw = S3Calloc(1,sizeof(S3view));
nvw->vrp.x   = DblToReal(10.0);
nvw->vrp.y   = DblToReal(10.0);
nvw->vrp.z   = DblToReal(10.0);
nvw->n.x     = DblToReal(-10.0);
nvw->n.y     = DblToReal(-10.0);
nvw->n.z     = DblToReal(-10.0);
nvw->v.x     = DblToReal(0.0);
nvw->v.y     = DblToReal(0.0);
nvw->v.z     = DblToReal(1.0);
nvw->d       = DblToReal(6.0);
nvw->clip    = DblToReal(0.001);
nvw->type    = vt_normal;
nvw->tar     = NULL;
nvw->backcol = 0;
nvw->win     = win;
nvw->lt      = S3Calloc(MAX_LIGHTS_PER_VIEW,sizeof(S3light *));
nvw->ltc     = 0;
nvw->ambient = DblToReal(0.3);
nvw->rmode   = rm;

/* Extract size and depth from window
 */
XGetWindowAttributes(display,nvw->win,&winattribs);
nvw->width   = winattribs.width;
nvw->height  = winattribs.height;
nvw->depth   = winattribs.depth;
nvw->scale   = DblToReal(VIEW_SCALE * nvw->width);
S3AllocViewStuff(display,nvw);

/* Create GC
 */
gcvals.graphics_exposures = 0;
gcvals.stipple = XCreateBitmapFromData(display,win,stipple_bits,2,2);
nvw->gc = XCreateGC(display,win,GCGraphicsExposures | GCStipple,&gcvals);

/* Load and selected font ("fixed" for now)
 */
if ((nvw->font = XLoadQueryFont(display,"fixed")) == NULL)
	fprintf(stderr,"S3CreateView : couldn't open font \"fixed\".  Using "
		"default font\n");
else
	XSetFont(display,nvw->gc,nvw->font->fid);
return nvw;
}


/* S3ChangeView
 * Attempt to change the view type to the one given
 */
void S3ChangeView(Display *display, S3view *vw, S3rmodetype rm)
{
S3FreeViewStuff(display,vw);
vw->rmode = rm;
S3AllocViewStuff(display,vw);
}


/* S3IPutPix8 and S3IPutPix24
 * Functions used for quick access to image data when using local server
 */
static int S3IPutPix8(XImage *i, int x, int y, unsigned long v)
{
i->data[y*i->bytes_per_line + x] = v;
return 0;
}

static int S3IPutPix24(XImage *i, int x, int y, unsigned long v)
{
register char *pp = (char *)&v;
register long idx = y*i->bytes_per_line + (x*(i->bits_per_pixel>>3));

i->data[idx]   = pp[0];
i->data[idx+1] = pp[1];
i->data[idx+2] = pp[2];
return 0;
}


/* S3AllocViewStuff
 * Allocates pixmaps, zbuffers and so on needed for a view.
 * May change the view type if something goes wrong.
 */
static void S3AllocViewStuff(Display *display, S3view *nvw)
{
if (nvw->rmode == rm_zbuf || nvw->rmode == rm_zbufshm) {
	Visual *visual = DefaultVisual(display,DefaultScreen(display));
	int depth      = DefaultDepth(display,DefaultScreen(display));

	nvw->zbuf    = calloc(nvw->width*nvw->height,sizeof(real));
	nvw->pbuf    = calloc(nvw->width*nvw->height,sizeof(S3polygon *));
	if (nvw->rmode == rm_zbufshm) {
		/* Attempt to use shared memory image
		 */
		if (!XShmQueryExtension(display))
			goto shmfailed;
		nvw->xshminfo = calloc(1,sizeof(XShmSegmentInfo));
		nvw->im	     = XShmCreateImage(display,visual,depth,ZPixmap,
					       NULL,nvw->xshminfo,nvw->width,
					       nvw->height);
		if (nvw->im == NULL) {
			free(nvw->xshminfo);
			goto shmfailed;
			}
		nvw->xshminfo->shmid = shmget(IPC_PRIVATE,
					nvw->im->bytes_per_line*nvw->im->height,
					IPC_CREAT | 0700);
		if (nvw->xshminfo->shmid == -1) {
			XDestroyImage(nvw->im);
			free(nvw->xshminfo);
			goto shmfailed;
			}
		nvw->dat = nvw->im->data = shmat(nvw->xshminfo->shmid, 0, 0);
		nvw->xshminfo->shmaddr = nvw->im->data;
		nvw->xshminfo->readOnly = False;
		S3TrapOn();
		XShmAttach(display, nvw->xshminfo);
		XSync(display,False);
		if (S3TrapOff() == 0) {
			/* Failed to connect shared mem to server!
			 */
			XDestroyImage(nvw->im);
			shmdt(nvw->xshminfo->shmaddr);
			shmctl(nvw->xshminfo->shmid,IPC_RMID,0);
			free(nvw->xshminfo);

			shmfailed:
			nvw->rmode = rm_zbuf;
			printf("Failed to use shared memory image ...\n");
			}
		else {
			if (nvw->im->depth == 8)
				nvw->im->f.put_pixel = S3IPutPix8;
			if (nvw->im->depth == 24)
				nvw->im->f.put_pixel = S3IPutPix24;
			printf("Using shared memory image ...\n");
			}
		}
	if (nvw->rmode == rm_zbuf) {
		/* Use an ordinary XImage
	 	 */
		nvw->im	     = XCreateImage(display,visual,depth,ZPixmap,0,
					    NULL,nvw->width,nvw->height,8,0);
		nvw->im->data = nvw->dat = malloc(nvw->im->bytes_per_line*
						  nvw->im->height);
		}
	}
else if (nvw->rmode == rm_plane || nvw->rmode == rm_wire) {
	nvw->dbuf    = XCreatePixmap(display,nvw->win,nvw->width,nvw->height,
				     nvw->depth);
	}
}


/* S3FreeViewStuff
 * De-allocates pixmap, z-buffers and so on allocated by S3AllocViewStuff()
 */
static void S3FreeViewStuff(Display *display, S3view *vw)
{
if (vw->rmode == rm_wire || vw->rmode == rm_plane) {
	/* Free up backing pixmap
	 */
	XFreePixmap(display,vw->dbuf);
	}
else if (vw->rmode == rm_zbuf || vw->rmode == rm_zbufshm) {
	/* Free z and polygon buffers
	 */
	S3Free(vw->zbuf);
	S3Free(vw->pbuf);
	if (vw->rmode == rm_zbuf) {
		/* Only need to free up image
		 */
		XDestroyImage(vw->im);
		}
	else {
		/* Kill off shared memory
		 */
		XShmDetach(display,vw->xshminfo);
		XDestroyImage(vw->im);
		shmdt(vw->xshminfo->shmaddr);
		shmctl(vw->xshminfo->shmid,IPC_RMID,0);
		free(vw->xshminfo);
		}
	}
}


/* S3SetViewSize
 * Change the size of an existing view
 */
void S3SetViewSize(Display *display, S3view *vw, int w, int h)
{
S3FreeViewStuff(display,vw);
vw->width  = w;
vw->height = h;
vw->scale  = DblToReal(VIEW_SCALE * vw->width);
S3AllocViewStuff(display,vw);
}


/* S3SetViewVrp
 * Change the view position of an existing view
 */
void S3SetViewVrp(S3view *vw, double x, double y, double z)
{
vw->vrp.x = DblToReal(x);
vw->vrp.y = DblToReal(y);
vw->vrp.z = DblToReal(z);
}


/* S3SetViewTarget
 * Set the object to be watched by this view
 */
void S3SetViewTarget(S3view *vw, S3object *obj)
{
vw->tar = obj;
}


/* S3SetViewN
 * Change the view direction vector
 */
void S3SetViewN(S3view *vw, double x, double y, double z)
{
vw->n.x = DblToReal(x);
vw->n.y = DblToReal(y);
vw->n.z = DblToReal(z);
}


/* S3SetViewV
 * Change the view up vector
 */
void S3SetViewV(S3view *vw, double x, double y, double z)
{
vw->v.x = DblToReal(x);
vw->v.y = DblToReal(y);
vw->v.z = DblToReal(z);
}


/* S3SetViewEyeDistance
 * Change the eye distance (the distance between the 'screen' and the eye)
 * of a view.
 */
void S3SetViewEyeDistance(S3view *vw, double d)
{
vw->d = DblToReal(d);
/* if (!RealEq(vw->clip,Real0))
	vw->clip = vw->d; */
}


/* S3SetViewBackColour
 * Change the background colour of a view
 */
void S3SetViewBackColour(S3view *vw, int col)
{
vw->backcol = col;
}


/* S3SetViewType
 * Set the view mode (default, circling, etc.. ) of a view
 */
void S3SetViewType(S3view *vw, S3viewtype type)
{
vw->type = type;
}


/* S3SetViewAmbient
 * Set the ambient light level of a view.
 * BUG - this could be better implemented as a 'ambient' light source type
 */
void S3SetViewAmbient(S3view *vw, double amb)
{
vw->ambient = DblToReal(amb);
}


/* S3SetViewScale
 * Change the drawing scale of a view
 */
void S3SetViewScale(S3view *vw, double sc)
{
vw->scale = DblToReal(sc * vw->width / 300.0);
}


/* S3FreeView
 * Totally free up a view struct
 */
void S3FreeView(Display *display, S3view *vw)
{
int i;

S3FreeViewStuff(display,vw);
for(i=vw->ltc-1; i>=0; i--)
	S3FreeViewLight(vw,i);
}
