#include "s3dt.h"
#include "s3dtint.h"

static int S3HashString(char *str);

/* S3CreateObjectTable
 * Create a new object hash table.
 */
S3objtab *S3CreateObjectTable(int type)
{
S3objtab *ntab = S3Calloc(1,sizeof(S3objtab));

ntab->type = type;
ntab->fa = calloc(10,sizeof(S3polygon *));
ntab->fas = 10;
return ntab;
}


/* S3FreeObjectTable
 * Free an object table and all objects in it
 */
void S3FreeObjectTable(S3objtab *tab)
{
S3ClearObjectTable(tab);
S3Free(tab->fa);
S3Free(tab);
}


/* S3AddObject
 * Insert an object into an object table
 */
void S3AddObject(S3objtab *tab, S3object *obj)
{
int hash;	/* position of object in table */
int i;

/* Link into view list
 */
obj->next = tab->head;
tab->head = obj;

if (tab->type == tt_id) {
	/* Hash to correct position in table by id.
	 */
	if (obj->id < 0) {
		fprintf(stderr,"S3AddObject : attempted to add object with "
			"invalid id : %d\n",obj->id);
		exit(10);
		}

	hash = obj->id % OBJTAB_SIZE;
	obj->hnext = tab->tab[hash];
	tab->tab[hash] = obj;
	}
else if (tab->type == tt_name && obj->name[0]) {
	/* Hash to correct position by name
	 */
	S3object **op;

	hash = S3HashString(obj->name);
	for(op=&tab->tab[hash]; *op; op=&((*op)->hnext))
		if (strequ(obj->name,(*op)->name)) {
			fprintf(stderr,"S3AddObjectByName : attempted to add "
				"object with duplicate name : %s\n",obj->name);
			exit(10);
			}
	(*op) = obj;
	}

/* Link polygons into view list
 */
for(i=0; i<obj->facec; i++)
	S3AddViewFace(tab,obj->face[i]);
}


/* S3DeleteObject
 * Delete an object from an object table and free it up.
 */
void S3DeleteObject(S3objtab *tab, S3object *obj)
{
S3object **prev;		/* pointer to pointer to it */
int hash;			/* position in hash table */
int i;

/* Unlink from display list
 */
for(prev=&tab->head; *prev != obj; prev = &((*prev)->next))
	;
*prev = obj->next;

if (tab->type == tt_id) {
	/* Remove from id hash table.
	 */
	hash = obj->id % OBJTAB_SIZE;
	for(prev=&tab->tab[hash]; *prev != obj; prev = &((*prev)->hnext))
		;
	*prev = obj->hnext;
	}
else if (tab->type == tt_name && obj->name[0]) {
	/* Remove from name hash table
	 */
	hash = S3HashString(obj->name);
	for(prev=&tab->tab[hash]; *prev != obj; prev = &((*prev)->hnext))
		;
	*prev = obj->hnext;
	}

/* Remove faces from view list
 */
for(i=0; i<obj->facec; i++)
	S3DelViewFace(tab,obj->face[i]);

S3FreeObject(obj);
}


/* S3LookupObject
 * Lookup an object in a table by it id or by name
 */
S3object *S3LookupObject(S3objtab *tab, ...)
{
va_list ap;
int hash;		/* position in hash table */
S3object *obj;

va_start(ap,tab);
if (tab->type == tt_id) {
	/* Lookup a numeric id in the hash table
	 */
	int id = va_arg(ap,int);

	if (id < 0)
		return NULL;	/* can't possibly be in table */
	hash = id%OBJTAB_SIZE;
	for(obj=tab->tab[hash]; obj && obj->id != id; obj=obj->next)
		;
	return obj;
	}
else if (tab->type == tt_name) {
	/* Lookup a name string
	 */
	char *name = va_arg(ap,char *);

	hash = S3HashString(name);
	for(obj=tab->tab[hash]; obj && !strequ(obj->name,name); obj=obj->hnext)
		;
	return obj;
	}
}


/* S3AddObjectAndChildren
 * Add an object and all it's children to an object table
 */
void S3AddObjectAndChildren(S3objtab *tab, S3object *obj)
{
int i;

S3AddObject(tab,obj);
for(i=0; i<obj->soc; i++)
	S3AddObjectAndChildren(tab,obj->so[i]);
}


/* S3DeleteObjectAndChildren
 * Remove and free an object and all it's children from a table
 */
void S3DeleteObjectAndChildren(S3objtab *tab, S3object *obj)
{
int i;

for(i=0; i<obj->soc; i++)
	S3DeleteObjectAndChildren(tab,obj->so[i]);
S3DeleteObject(tab,obj);
}


/* S3AddViewFace
 * Add a face to the polygon lists
 */
void S3AddViewFace(S3objtab *tab, S3polygon *p)
{
p->fnext = tab->fph;
tab->fph = p;
}


/* S3DelViewFace
 * Unlink a face from the polygon list
 */
void S3DelViewFace(S3objtab *tab, S3polygon *p)
{
S3polygon **pp;

for(pp=&tab->fph; *pp != p; pp=&((*pp)->fnext))
	;
*pp = p->fnext;
}



/* S3ClearObjectTable
 * Delete every object in an object table list and clear the hash table.
 */
void S3ClearObjectTable(S3objtab *tab)
{
int i;
S3object *o, *nxt;

for(o=tab->head; o; o=nxt) {
	nxt = o->next;
	S3FreeObject(o);
	}
for(i=0; i<OBJTAB_SIZE; i++)
	tab->tab[i] = NULL;

tab->head = NULL;
tab->fph  = NULL;
S3Free(tab->fa);
tab->fa = calloc(10,sizeof(S3polygon *));
tab->fas = 10;
}


/* S3HashString
 * Hash a string to a position in the hash table
 */
static int S3HashString(char *str)
{
int sum = 0;

while(*str)
	sum += *(str++);
return sum % OBJTAB_SIZE;
}

