#include "s3dt.h"
#include "s3dtint.h"

/* S3CreateCube
 * Create a unit cube object of the specified colour and face type
 */
S3object *S3CreateCube(int colour, int ftype)
{
S3object *nobj;

nobj = S3CreateObject("cube",0);

/* Add points at corners of cube
 */
S3AddPoint(nobj,-0.5,-0.5,-0.5);
S3AddPoint(nobj, 0.5,-0.5,-0.5);
S3AddPoint(nobj, 0.5, 0.5,-0.5);
S3AddPoint(nobj,-0.5, 0.5,-0.5);
S3AddPoint(nobj,-0.5,-0.5, 0.5);
S3AddPoint(nobj, 0.5,-0.5, 0.5);
S3AddPoint(nobj, 0.5, 0.5, 0.5);
S3AddPoint(nobj,-0.5, 0.5, 0.5);

/* Add 
 */
S3AddFace(nobj,colour,ftype,4,0,1,2,3);
S3AddFace(nobj,colour,ftype,4,7,6,5,4);
S3AddFace(nobj,colour,ftype,4,0,4,5,1);
S3AddFace(nobj,colour,ftype,4,1,5,6,2);
S3AddFace(nobj,colour,ftype,4,2,6,7,3);
S3AddFace(nobj,colour,ftype,4,3,7,4,0);

return nobj;
}


/* S3CreateCone
 * Create a cone object 1 unit wide and high, of the specified colour
 */
S3object *S3CreateCone(int colour, int ftype)
{
S3object *nobj;
int i;

nobj = S3CreateObject("cone",0);

/* Create 10 points around base, plus one on top
 */
for(i=0; i<360; i+=36) {
	double rad = dtor(i);
	double x,y;

	x = cos(rad)*0.5;
	y = sin(rad)*0.5;
	S3AddPoint(nobj,x,y,-0.5);
	}
S3AddPoint(nobj, 0.0, 0.0, 0.5);

/* Create 10 faces around the sides, plus one for the base
 */
for(i=0; i<10; i++)
	S3AddFace(nobj,colour,ftype,3,(i+1)%10,i,10);
S3AddFace(nobj,colour,ftype,10,0,1,2,3,4,5,6,7,8,9);

return nobj;
}


/* S3CreatePlane
 * Create a plane lying normal to the z axis, and as large as the current
 * real number representation allows
 */
S3object *S3CreatePlane(int colour, int ftype)
{
S3object *nobj;

nobj = S3CreateObject("plane",0);

/* Create corner points far in the distance
 */
S3AddPoint(nobj,-CoordMax,-CoordMax,0.0);
S3AddPoint(nobj, CoordMax,-CoordMax,0.0);
S3AddPoint(nobj, CoordMax, CoordMax,0.0);
S3AddPoint(nobj,-CoordMax, CoordMax,0.0);

/* Create plane polygon
 */
S3AddFace(nobj,colour,ftype,4,0,1,2,3);

return nobj;
}


/* S3CreateSphere
 * Create an (approximately) spherical object of radius 1 composed of polygons.
 */
S3object *S3CreateSphere(int colour, int ftype)
{
S3object *nobj;
int m, p;

nobj = S3CreateObject("sphere",0);

S3AddPoint(nobj,0,0,-0.5);
S3AddPoint(nobj,0,0,0.5);
for(m=-3; m<=3; m++) {
	double tilt = m*22.5;

	for(p=0; p<18; p++) {
		double ang = p*20;
		double x = cos(dtor(ang))*cos(dtor(tilt))*0.5;
		double y = sin(dtor(ang))*cos(dtor(tilt))*0.5;
		double z = sin(dtor(tilt))*0.5;

		S3AddPoint(nobj,x,y,z);
		}
	}
/* Add faces around sides
 */
for(m=-3; m<3; m++) {
	int ms = (m+3)*18 + 2;

	for(p=0; p<18; p++) {
		S3AddFace(nobj,colour,ftype,4, ms+p, ms+18+p,
			  ms+18+((p+1)%18), ms+((p+1)%18));
		}
	}
/* Add top and bottom faces
 */
for(p=0; p<18; p++) {
	S3AddFace(nobj,colour,ftype,3,0,p+2,((p+1)%18)+2);
	/*S3AddFace(nobj,colour,ftype,3,1,p+110,((p+1)%18)+110);*/
	S3AddFace(nobj,colour,ftype,3,1,((p+1)%18)+110,p+110);
	}
return nobj;
}


/* S3CreateTube
 * Create a cylindrical object with closed ends, of height and diameter 1
 */
S3object *S3CreateTube(int colour, int ftype)
{
S3object *nobj;
int i;

nobj = S3CreateObject("tube",0);

/* Create 10 points around base, and 10 points around top
 */
for(i=0; i<360; i+=36) {
	double rad = dtor(i);
	double x,y;

	x = cos(rad)*0.5;
	y = sin(rad)*0.5;
	S3AddPoint(nobj,x,y,0.0);
	}
for(i=0; i<360; i+=36) {
	double rad = dtor(i);
	double x,y;

	x = cos(rad)*0.5;
	y = sin(rad)*0.5;
	S3AddPoint(nobj,x,y,1.0);
	}

/* Create 10 faces around the sides.
 */
for(i=0; i<10; i++)
	S3AddFace(nobj,colour,ftype,4,10+i,10+((i+1)%10),(i+1)%10,i);

/* Create base and top faces
 */
S3AddFace(nobj,colour,ftype,10,0,1,2,3,4,5,6,7,8,9);
S3AddFace(nobj,colour,ftype,10,19,18,17,16,15,14,13,12,11,10);

return nobj;
}


/* S3CreateDisc
 * Create a disc shaped object aligned in the x-z plane, of unit diameter
 */
S3object *S3CreateDisc(int colour, int ftype)
{
S3object *nobj;
int i;

nobj = S3CreateObject("disc",0);

for(i=0; i<360; i+=36) {
	double rad = dtor(i);
	double x,z;

	x = cos(rad)*0.5;
	z = sin(rad)*0.5;
	S3AddPoint(nobj,x,0.0,z);
	}
S3AddFace(nobj,colour,ftype,10,0,1,2,3,4,5,6,7,8,9);

return nobj;
}

