#include "s3dt.h"
#include "s3dtint.h"
#include "s3view.h"

static void S3ComputeCamera(S3view *vw, S3objtab *oh);
static void S3ComputeViewMatrix(S3view *vw);
static void S3ComputeDepths(S3view *vw, S3objtab *oh);
static void S3DepthSort(S3view *vw, S3objtab *oh);
static void S3ComputeLights(S3view *vw);
static void S3PersDiv(S3view *vw, S3point *vp, S3point *sp);
static void S3ClipFace(S3view *vw, S3polygon *p, S3point fpt[], int *fptc);
static S3point S3Intercept(S3point *l, S3point *r, real z);
static int S3InFront(S3view *vw, S3polygon *b, S3polygon *a);
void S3ComputeDepths(S3view *vw, S3objtab *tab);
static void S3SortFarList(S3objtab *ot, int (*cf)(S3polygon *, S3polygon *));
static int S3FarFirst(S3polygon *a, S3polygon *b);
static int S3NearFirst(S3polygon *a, S3polygon *b);
static void S3ConvertToView(S3view *vw, S3objtab *oh);
static void S3Render3d(Display *display, S3view *vw, S3objtab *oh);
static S3polygon *S3MergeSort(S3polygon *hd, int len,
			      int (*cf)(S3polygon *, S3polygon *));
void S3ScanConvert(S3view *vw, S3point *spt, int sptc, S3polygon *poly,
		   real *ssh);
static int S3Overlap(S3view *vw, S3polygon *a, S3polygon *b);
static real S3ComputeBright(S3view *vw, S3polygon *poly, S3point *p,
			    S3point *nm);

/* S3vertex
 * Structure used by scan-converter only for storing vertex and interpolation
 * information.
 */
typedef struct {
	real x,y,z;	/* screen pos and depth */
	real c;		/* 'ideal' colour at vertex */
	real r,g,b;	/* rgb values at vertex, for 24bit */
	} S3vertex;
static void S3scanline(int y, S3vertex *l, S3vertex *r, S3view *vw,
		       S3polygon *poly);
static void S3incrementalize_y(S3vertex *p1, S3vertex *p2, S3vertex *p,
			       S3vertex *dp, int y, S3view *vw);
static void S3incrementalize_x(S3vertex *p1, S3vertex *p2, S3vertex *p,
			       S3vertex *dp, int x, S3view *vw);
static void S3increment(S3vertex *p, S3vertex *dp, S3view *vw, int dox);


/* S3RenderView
 * Perform all the steps necessary to draw a 3d scene into a view window
 */
void S3RenderView(Display *display, S3view *vw, S3objtab *oh)
{
if (vw->rmode == rm_wire || vw->rmode == rm_plane) {
	/* Clear out background of buffer
	 */
	S3SetColour(display,vw->gc,vw->backcol);
	XFillRectangle(display,vw->dbuf,vw->gc,0,0,vw->width,vw->height);
	}

S3CallBeforeFuncs(display,vw,oh);

/* Compute and draw 3d scene
 */
S3ComputeCamera(vw,oh);
S3ComputeViewMatrix(vw);
S3ComputeLights(vw);
S3ConvertToView(vw,oh);
S3ComputeDepths(vw,oh);
if (vw->rmode == rm_plane) {
	S3DepthSort(vw,oh);
	}
else {
	/* Build list of polygons for S3Render3d to use.
 	 */
	if (vw->rmode == rm_wire && oh->fph && oh->fph->fnext)
		S3SortFarList(oh,S3FarFirst);
	else if (vw->rmode == rm_zbuf || vw->rmode == rm_zbufshm) {
		/* BUG - use a faster sort.
		 */
		S3SortFarList(oh,S3NearFirst);
		}
	}
S3Render3d(display,vw,oh);
S3RefreshView(display,vw);
XFlush(display);
XSync(display,False);

S3CallAfterFuncs(display,vw,oh);
}


/* S3RefreshView
 * Copy the background pixmap or image to the view window
 */
void S3RefreshView(Display *display, S3view *vw)
{
if (vw->rmode == rm_plane || vw->rmode == rm_wire)
	XCopyArea(display,vw->dbuf,vw->win,vw->gc,0,0,vw->width,vw->height,0,0);
else if (vw->rmode == rm_zbuf) {
	XPutImage(display,vw->win,vw->gc,vw->im,0,0,0,0,vw->width,vw->height);
	}
else if (vw->rmode == rm_zbufshm) {
	XShmPutImage(display,vw->win,vw->gc,vw->im,0,0,0,0,vw->width,
		     vw->height,True);
	}
}


/* S3ComputeCamera
 * Computes values for the vrp, u and v for the current viewing mode.
 */
static void S3ComputeCamera(S3view *vw, S3objtab *oh)
{
S3object *obj = vw->tar;
double radius;					/* radius of target */
double tm = S3GetTime();

if (!obj && vw->type != vt_normal) {
	fprintf(stderr,"S3ComputeCamera : no view target object given\n");
	exit(10);
	}

if (obj) {
	/* Compute a sensible radius for the object being viewed
	 */
	radius = obj->radius;
	if (RealToDbl(radius) < 1.0)
		radius = DblToReal(1.0);
	}

switch(vw->type) {
case vt_normal:
	/* Accept programmer-specified values for vrp, u and v
	 */
	break;

case vt_close:
	/* Just outside object, looking in.
	 */
	vw->vrp.x = AddReal(obj->pos.x,ScaleReal(radius,3.0));
	vw->vrp.y = AddReal(obj->pos.y,ScaleReal(radius,3.0));
	vw->vrp.z = AddReal(obj->pos.z,ScaleReal(radius,3.0));
	vw->n.x = vw->n.y = vw->n.z = DblToReal(-100.0);
	vw->v.x = DblToReal(0.0);
	vw->v.y = DblToReal(0.0);
	vw->v.z = DblToReal(1.0);
	break;

case vt_far:
	/* Some distance from object, looking in
	 */
	vw->vrp.x = AddReal(obj->pos.x,ScaleReal(radius,8.0));
	vw->vrp.y = AddReal(obj->pos.y,ScaleReal(radius,8.0));
	vw->vrp.z = AddReal(obj->pos.z,ScaleReal(radius,8.0));
	vw->n.x = vw->n.y = vw->n.z = DblToReal(-100.0);
	vw->v.x = DblToReal(0.0);
	vw->v.y = DblToReal(0.0);
	vw->v.z = DblToReal(1.0);
	break;

case vt_circling:
	/* Outside object, circling it and looking in
	 */
	vw->n.x = MulReal3(DblToReal(cos(tm)),obj->radius,DblToReal(5.0));
	vw->n.y = MulReal3(DblToReal(sin(tm)),obj->radius,DblToReal(5.0));
	vw->n.z = MulReal3(DblToReal(cos(tm*1.3)),obj->radius,DblToReal(5.0));
	vw->vrp.x = SubReal(obj->pos.x,vw->n.x);
	vw->vrp.y = SubReal(obj->pos.y,vw->n.y);
	vw->vrp.z = SubReal(obj->pos.z,vw->n.z);
	vw->v.x = DblToReal(0.0);
	vw->v.y = DblToReal(0.0);
	vw->v.z = DblToReal(1.0);
	break;
	}
}


/* S3ComputeViewMatrix
 * Compute the matrix for world to view conversion from the view U, V and N
 * vectors.
 */
static void S3ComputeViewMatrix(S3view *vw)
{
real orth;

orth = DivReal( S3DotProduct(&vw->n,&vw->v), S3DotProduct(&vw->n,&vw->n) );
vw->v.x = SubReal(vw->v.x, MulReal(orth,vw->n.x) );
vw->v.y = SubReal(vw->v.y, MulReal(orth,vw->n.y) );
vw->v.z = SubReal(vw->v.z, MulReal(orth,vw->n.z) );

S3CrossProduct(&vw->v,&vw->n,&vw->u);

S3Normalise(&vw->u);
S3Normalise(&vw->v);
S3Normalise(&vw->n);

vw->vmat[0][0]=vw->u.x; vw->vmat[0][1]=vw->u.y; vw->vmat[0][2]=vw->u.z;
vw->vmat[1][0]=vw->v.x; vw->vmat[1][1]=vw->v.y; vw->vmat[1][2]=vw->v.z;
vw->vmat[2][0]=vw->n.x; vw->vmat[2][1]=vw->n.y; vw->vmat[2][2]=vw->n.z;
}


/* S3ComputeLights
 * Convert the positions of all light sources in this view to view coordinates
 */
static void S3ComputeLights(S3view *vw)
{
int i;

for(i=0; i<vw->ltc; i++) {
	S3light *light = vw->lt[i];
	S3point tmp;

	tmp.x = SubReal(light->pos.x, vw->vrp.x);
	tmp.y = SubReal(light->pos.y, vw->vrp.y);
	tmp.z = SubReal(light->pos.z, vw->vrp.z);
	S3MMult(&tmp,vw->vmat,&light->vpos);
	}
}


/* S3ComputeDepths
 * Compute the farpoint and closepoint distances for each polygon
 */
static void S3ComputeDepths(S3view *vw, S3objtab *tab)
{
S3polygon *p;

for(p=tab->fph; p; p=p->fnext) {
	int i;
	real fdist = Real0, cdist = CoordMax;

	for(i=0; i<p->vertc; i++) {
		S3object *ob = p->obj;
		real d = ob->vpt[p->vert[i]]->z;

		if (RealLt(d,Real0)) d = Real0;
		if (RealGt(d,fdist)) fdist = d;
		if (RealLt(d,cdist)) cdist = d;
		}
	p->dist = AddReal( fdist, MulReal(cdist,DblToReal(0.2)) );
	}
}


/* S3DepthSort
 * Create a list of polygons in the correct order for drawing
 */
static void S3DepthSort(S3view *vw, S3objtab *ot)
{
S3polygon *p, *q;
int pi, qi;
S3polygon *f;
int polyc;			/* number of polygons */

/* BUG - sort by farpoint distance only for now.
 */
if (ot->fph && ot->fph->fnext) {
	S3SortFarList(ot,S3FarFirst);
	}

#if 0
/* Build array of polygons in farpoint order
 */
if (ot->fph && ot->fph->fnext) {
	S3SortFarList(ot,S3FarFirst);
	}
for(polyc=0,f=ot->fph; f; f=f->fnext,polyc++) {
	if (polyc >= ot->fas) {
		ot->fas *= 2;
		ot->fa = realloc(ot->fa,ot->fas*sizeof(S3polygon *));
		}
	ot->fa[polyc] = f;
	f->mvd = -1;
	}

pi = 0;
while(pi < polyc) {
retry:	qi = pi+1;
	p = ot->fa[pi]; q = ot->fa[qi];
	while(qi < polyc && q->fdist >= p->cdist) {
		if (S3Overlap(vw,p,q)) {
			if (!S3Overlap(vw,q,p)) {
				/* q should really be behind. Swap them.
				 */
				ot->fa[pi] = q;
				ot->fa[qi] = p;
				goto retry;
				}
			else {
				/* No definate order can be found! 
				 */
				printf("Indefinate order\n");
				}
			}
		qi++; q = ot->fa[qi];
		}
	/* p is in the right spot. Add to display list and carry on
	 */
	*sp = p;
	sp = &p->next;
	pi++;
	}
*sp = NULL;
#endif
}


/* S3Overlap
 * Returns non-zero if polygon a is in front of polygon b on screen
 */
static int S3Overlap(S3view *vw, S3polygon *a, S3polygon *b)
{
if (a->xmax < b->xmin || a->xmin > b->xmax)
	return 0;
if (a->ymax < b->ymin || a->ymin > b->ymax)
	return 0;
if (S3InFront(vw,a,b) <= 0)
	return 0;
if (S3InFront(vw,b,a) >= 0)
	return 0;
return 1;
}


/* S3SortFarList
 * Sort a list of polygons by their farpoint distance. Uses bubble sort if
 * there isn't much 'disorder', or merge sort if there is.
 */
static void S3SortFarList(S3objtab *ot, int (*cf)(S3polygon *, S3polygon *))
{
int disorder;
S3polygon *p;
int len = 0;		/* polygons in list */

if (!ot->fph || !ot->fph->fnext)
	return;	/* only 0 or 1 polys */

/* Find how many polygons are out of order
 */
for(p=ot->fph; p->fnext; p=p->fnext, len++)
	if (cf(p,p->fnext)) disorder++;

if (disorder < 3) {
	/* Bubble sort is likely to be the best option..
	 */
	int sorting;

	do {
		S3polygon **pp;

		sorting = 0;
		for(pp=&ot->fph; (*pp)->fnext; pp=&((*pp)->fnext)) {
			if (cf((*pp),(*pp)->fnext)) {
				/* swap 'em! */
				S3polygon *temp = (*pp)->fnext;
	
				(*pp)->fnext = (*pp)->fnext->fnext;
				temp->fnext = *pp; *pp = temp;
				sorting = 1;
				}
			}
		} while(sorting);
	}
else {
	/* Need to use merge sort..
	 */
	ot->fph = S3MergeSort(ot->fph,len,cf);
	}
}


/* S3MergeSort
 * Perform recursive merge sorting on a list of polygons
 */
static S3polygon *S3MergeSort(S3polygon *hd, int len,
			      int (*cf)(S3polygon *, S3polygon *))
{
if (len <= 1)
	return hd;
else {
	int halfway = len/2;
	S3polygon **hwp;		/* pointer to pointer to 1/2 way node */
	S3polygon *hw;			/* pointer to 1/2 way node */
	S3polygon *l1, *l2;		/* two sorted halves */
	S3polygon *slh;			/* head of sorted list */
	S3polygon **sl;
	int i;

	/* Split and sort halves
	 */
	for(hwp=&hd,i=0; i<halfway; hwp=&((*hwp)->fnext),i++)
		;
	hw = *hwp;
	*hwp = NULL;
	l1 = S3MergeSort(hd,halfway,cf);
	l2 = S3MergeSort(hw,len-halfway,cf);

	/* Rejoin two halves
	 */
	sl = &slh;
	while(l1 || l2) {
		if (l1 && (!l2 || !cf(l1,l2))) {
			/* The polygon at the head of l1 is first
			 */
			*sl = l1;
			sl = &(l1->fnext);
			l1 = l1->fnext;
			}
		else {
			/* The head of l2 is first
			 */
			*sl = l2;
			sl = &(l2->fnext);
			l2 = l2->fnext;
			}
		}
	*sl = NULL;
	return slh;
	}
}


/* S3FarFirst, S3NearFirst
 * Comparison functions for sorting a list of polygons in normal and reverse
 * order, respectively.
 */
static int S3FarFirst(S3polygon *a, S3polygon *b)
{
return RealLt(a->dist,b->dist);
}

static int S3NearFirst(S3polygon *a, S3polygon *b)
{
return RealGt(a->dist,b->dist);
}


/* S3ConvertToView
 * Convert all points in all objects to view coordinates
 */
static void S3ConvertToView(S3view *vw, S3objtab *oh)
{
S3object *o;

for(o=oh->head; o; o=o->next) {
	int i;

	for(i=0; i<o->ptc; i++) {
		S3point tmp;

		tmp.x = SubReal(AddReal(o->pt[i]->x,o->pos.x), vw->vrp.x);
		tmp.y = SubReal(AddReal(o->pt[i]->y,o->pos.y), vw->vrp.y);
		tmp.z = SubReal(AddReal(o->pt[i]->z,o->pos.z), vw->vrp.z);
		S3MMult(&tmp,vw->vmat,o->vpt[i]);
		}
	}
}


/* S3Render3d
 * Draw the list of polygons into the view pixmap or image
 */
static void S3Render3d(Display *display, S3view *vw, S3objtab *oh)
{
S3object *o;
S3polygon *poly;
int p;

if (vw->rmode == rm_zbuf || vw->rmode == rm_zbufshm) {
	int pval = S3GetColour(vw->backcol);

	/* Clear out z buffer and image
	 */
	for(p=0; p<vw->width*vw->height; p++) {
		vw->zbuf[p] = DblToReal(CoordMax);
		}
	if (pval) {
		register int x,y;
		for(x=0; x<vw->width; x++)
			for(y=0; y<vw->height; y++)
				XPutPixel(vw->im,x,y,pval);
		}
	else {
		memset(vw->dat,0,vw->im->bytes_per_line*vw->im->height);
		}
	}

/* Reset on-screen bounding boxes for objects
 */
for(o=oh->head; o; o=o->next) {
	o->onscreen = 0;
	o->bminx =  32767; o->bminy =  32767;
	o->bmaxx = -32767; o->bmaxy = -32767;
	}

/* Draw each face in the view window
 */
for(poly=oh->fph; poly; poly=poly->fnext) {
	S3point fpt[MAX_VERTS_PER_FACE]; /* points in view coords */
	int fptc;			 /* no. of points */
	S3point spt[MAX_VERTS_PER_FACE]; /* points after perspective division */
	int sptc;			 /* no. of screen points */
	int infront = 1, behind = 1;
	int j;
	S3point fn;			 /* face normal */
	real ssh[MAX_VERTS_PER_FACE];	 /* shade of each vertex */
	S3object *obj = poly->obj;

	poly->onscreen = 0;

	if (obj->invis)
		continue;

	/* Check if polygon is in front or behind viewer
	 */
	for(j=0; j<poly->vertc; j++) {
		real pz = obj->vpt[poly->vert[j]]->z;

		if (RealLt(pz,vw->clip))
			infront = 0;
		else
			behind = 0;
		}

	if (behind)
		continue;	/* totally behind view */


	if (!infront) {
		/* Build an array of view-coordinate vertex
		 * positions.
		 */
		for(fptc=0; fptc<poly->vertc; fptc++)
			fpt[fptc] = *obj->vpt[poly->vert[fptc]];

		/* Clip face at z=0
		 */
		S3ClipFace(vw,poly,fpt,&fptc);

		for(sptc=0; sptc<fptc; sptc++) {
			S3PersDiv(vw,&fpt[sptc],&spt[sptc]);
			}
		spt[sptc] = spt[0];
		}
	else {
		/* Just convert to screen coords
		 */
		for(sptc=0,fptc=0; sptc<poly->vertc; sptc++,fptc++) {
			fpt[fptc] = *obj->vpt[poly->vert[fptc]];
			S3PersDiv(vw,&fpt[sptc],&spt[sptc]);
			}
		spt[sptc] = spt[0];
		}

	if (poly->type & pt_oneside) {
		/* Check if face should be backface removed
		 */
		real ax,ay,bx,by;

		ax = SubReal( spt[0].x , spt[1].x );
		ay = SubReal( spt[0].y , spt[1].y );
		bx = SubReal( spt[2].x , spt[1].x );
		by = SubReal( spt[2].y , spt[1].y );
		if (!RealGt( MulReal(ax,by) , MulReal(ay,bx) ))
			continue;
		}

	/* Find on-screen extents of this polygon
	 */
	poly->xmin = poly->ymin = 32767;
	poly->xmax = poly->ymax = -32767;
	for(j=0; j<sptc; j++) {
		poly->xmin = min(RealToInt(spt[j].x),poly->xmin);
		poly->ymin = min(RealToInt(spt[j].y),poly->ymin);
		poly->xmax = max(RealToInt(spt[j].x),poly->xmax);
		poly->ymax = max(RealToInt(spt[j].y),poly->ymax);
		}
	if (poly->xmax < 0 || poly->ymax < 0 ||
	    poly->xmin > vw->width || poly->ymin > vw->height)
		continue;	/* totally off screen */
	else
		poly->onscreen = 1;

	/* Compute face normal in view coords
	 */
	S3MMult(&poly->norm,vw->vmat,&fn);

	/* Compute face shading
	 */
	if ((poly->type & pt_lighted) && (poly->type & pt_shaded) &&
					 (vw->rmode >= rm_zbuf)) {
		/* Compute shade at every vertex
		 */
		for(j=0; j<fptc; j++)
			ssh[j] = S3ComputeBright(vw,poly,&fpt[j],&fn);
		}
	else if (poly->type & pt_lighted) {
		/* Flat-shade polygon with shade at centre
		 */
		S3point tmp, cen;
		real bri;

		tmp.x = SubReal( AddReal(poly->cen.x,obj->pos.x) , vw->vrp.x );
		tmp.y = SubReal( AddReal(poly->cen.y,obj->pos.y) , vw->vrp.y );
		tmp.z = SubReal( AddReal(poly->cen.z,obj->pos.z) , vw->vrp.z );
		S3MMult(&tmp,vw->vmat,&cen);
		bri = S3ComputeBright(vw,poly,&cen,&fn);
		for(j=0; j<fptc; j++)
			ssh[j] = bri;
		}
	else {
		/* No light source effect at all.
		 */
		for(j=0; j<fptc; j++)
			ssh[j] = Real1;
		}

	/* Add the extents of this polygon to bounding box of object
	 */
	obj->bmaxx = max(poly->xmax,obj->bmaxx);
	obj->bmaxy = max(poly->ymax,obj->bmaxy);
	obj->bminx = min(poly->xmin,obj->bminx);
	obj->bminy = min(poly->ymin,obj->bminy);
	obj->onscreen = 1;

	if (vw->rmode == rm_zbuf || vw->rmode == rm_zbufshm) {
		S3ScanConvert(vw,spt,sptc,poly,ssh);
		}
	else if (vw->rmode == rm_plane || vw->rmode == rm_wire) {
		int colbase = poly->colour / 32;
		int colshade = poly->colour % 32;
		XPoint xpt[MAX_VERTS_PER_FACE];
		int xp;

		/* Force coords to be on-screen, and convert to Xpoints
		 */
		for(xp=0; xp<sptc+1; xp++) {
			xpt[xp].x = min(max(-32000,RealToDbl(spt[xp].x)),32000);
			xpt[xp].y = min(max(-32000,RealToDbl(spt[xp].y)),32000);
			}
		S3SetColour(display,vw->gc,colbase*32 +
					  colshade*RealToDbl(ssh[0]));
		if (poly->type & pt_glass) {
			/* Change to transparent mode
			 */
			XSetFillStyle(display,vw->gc,FillStippled);
			}

		if (vw->rmode == rm_plane) {
			/* Draw polygon as a filled shape
			 */
			XFillPolygon(display,vw->dbuf,vw->gc,xpt,sptc,
			     Convex,CoordModeOrigin);
			}
		else {
			/* Only draw outline of polygon
			 */
			XDrawLines(display,vw->dbuf,vw->gc,xpt,sptc+1,
				   CoordModeOrigin);
			}

		if (poly->type & pt_glass) {
			/* Change back to normal opaque mode
			 */
			XSetFillStyle(display,vw->gc,FillSolid);
			}

		if (poly->obj->highlight != -1 && (poly->type & pt_plane)) {
			/* Draw highlight around face
			 */
			S3SetColour(display,vw->gc,poly->obj->highlight);
			XDrawLines(display,vw->dbuf,vw->gc,xpt,sptc+1,
				   CoordModeOrigin);
			}
		}

	} /* end for */

}


/* S3PersDiv
 * Convert one point from view to screen coordinates
 */
static void S3PersDiv(S3view *vw, S3point *vp, S3point *sp)
{
real pers = AddReal( Real1, DivReal(vp->z,vw->d) );
S3point pp;

sp->x = AddReal( MulReal(vw->scale, DivReal(vp->x,pers) ),
		 IntToReal(vw->width/2) );
sp->y = AddReal( MulReal3(vw->scale, DivReal(vp->y,pers),
		 DblToReal(-1.0) ), IntToReal(vw->height/2) );
sp->z = DivReal( vp->z , pers );
/*
if (RealGt(pp.x,IntToReal(32767))) {
	pp.x = IntToReal(32767);
	}
else if (RealLt(pp.x,IntToReal(-32767))) {
	pp.x = IntToReal(-32767);
	}
if (RealGt(pp.y,IntToReal(32767))) {
	pp.y = IntToReal(32767);
	}
else if (RealLt(pp.y,IntToReal(-32767))) {
	pp.y = IntToReal(-32767);
	}
sp->x = RealToInt(pp.x);
sp->y = RealToInt(pp.y);
*/
}


/* S3ClipFace
 * Change an array of points representing a face to cut it off at z=0.
 */
static void S3ClipFace(S3view *vw, S3polygon *p, S3point fpt[], int *fptc)
{
S3point old[MAX_VERTS_PER_FACE];	/* points before clipping */
int nptc = 0;				/* no. of points after clipping */
int i;

memcpy(old,fpt,(*fptc)*sizeof(S3point));
for(i=0; i<*fptc; i++) {
	if (!RealLt(old[i].z,vw->clip))
		fpt[nptc++] = old[i];
	else {
		int pn,nn;	/* indexes of previous and next points */
		int prev;	/* is previous point at z<0 ? */
		int next;	/* is next point at z<0 ? */

		pn = (i == 0) ? (*fptc)-1 : i-1;
		nn = (i == (*fptc)-1) ? 0 : i+1;
		prev = RealLt(old[pn].z,vw->clip);
		next = RealLt(old[nn].z,vw->clip);
		if (prev && next) {
			/* Both neighbours are at z < 0
			 */
			/*
			fpt[nptc] = old[i];
			fpt[nptc].z = vw->clip;
			nptc++;
			*/
			/* chuck away! */
			}
		else if (!prev && !next) {
			/* Both neighbours are at z > 0
			 */
			fpt[nptc] = S3Intercept(&old[pn],&old[i],vw->clip);
			fpt[nptc+1] = S3Intercept(&old[nn],&old[i],vw->clip);
			nptc += 2;
			}
		else if (!prev) {
			/* Only previous is at z > 0
			 */
			fpt[nptc] = S3Intercept(&old[pn],&old[i],vw->clip);
			nptc++;
			}
		else {
			/* Only next is at z > 0
			 */
			fpt[nptc] = S3Intercept(&old[nn],&old[i],vw->clip);
			nptc++;
			}
		}
	}
*fptc = nptc;
}


/* S3Intercept
 * Find the position of the point where the line from l to r intercepts
 * the plane at z = 0
 * l = point at z>0, r = point at z<0
 */
static S3point S3Intercept(S3point *l, S3point *r, real z)
{
S3point i;
real xzslope, yzslope;

xzslope = DivReal( SubReal(r->x,l->x) , SubReal(r->z,l->z) );
yzslope = DivReal( SubReal(r->y,l->y) , SubReal(r->z,l->z) );
i.x = SubReal(l->x, MulReal(xzslope,SubReal(l->z,z)));
i.y = SubReal(l->y, MulReal(yzslope,SubReal(l->z,z)));
i.z = z;
return i;
}


/* S3ObjectAtPos
 * Find the object closest to the viewer, at position x,y in the view window
 */
S3object *S3ObjectAtPos(S3view *vw, S3objtab *ot, int x, int y)
{
if (vw->rmode == rm_zbuf || vw->rmode == rm_zbufshm) {
	/* Lookup polyon in polygon-pixel buffer
	 */
	int pix = y*vw->width + x;

	if (vw->zbuf[pix] == CoordMax)
		return NULL;
	else
		return vw->pbuf[pix]->obj;
	}
else if (vw->rmode == rm_wire || vw->rmode == rm_plane) {
	/* Find closest polygon in display list that covers point
	 */
	S3object *tar = NULL;		/* object  at position */
	S3polygon *p;

	for(p=ot->fph; p; p=p->fnext) {
		if (!p->obj->invis && p->onscreen && x >= p->xmin &&
		    x <= p->xmax && y >= p->ymin && y <= p->ymax) {
			tar = p->obj;
			}
		}
	return tar;
	}
}


/* S3InFront
 * Returns 1 if polygon b is entirely on the same side of a as the vrp,
 * -1 if b is entirely on the opposite side and 0 if unsure.
 */
static int S3InFront(S3view *vw, S3polygon *b, S3polygon *a)
{
S3point ai, acen;
int aidotn;
int i;
int infront = 1, behind = 1;

ai.x = SubReal(vw->vrp.x, AddReal(a->cen.x,a->obj->pos.x) );
ai.y = SubReal(vw->vrp.y, AddReal(a->cen.y,a->obj->pos.y) );
ai.z = SubReal(vw->vrp.z, AddReal(a->cen.z,a->obj->pos.z) );
aidotn = RealToInt(SgnReal(S3DotProduct(&a->norm,&ai)));

acen.x = AddReal(a->obj->pos.x,a->cen.x);
acen.y = AddReal(a->obj->pos.y,a->cen.y);
acen.z = AddReal(a->obj->pos.z,a->cen.z);
for(i=0; i<b->vertc; i++) {
	S3point p, ap;
	S3object *ob = b->obj;
	int apdotn;

	p.x = AddReal(ob->pos.x,ob->pt[b->vert[i]]->x);
	p.y = AddReal(ob->pos.y,ob->pt[b->vert[i]]->y);
	p.z = AddReal(ob->pos.z,ob->pt[b->vert[i]]->z);
	ap.x = SubReal(p.x,acen.x);
	ap.y = SubReal(p.y,acen.y);
	ap.z = SubReal(p.z,acen.z);
	apdotn = RealToInt(SgnReal(S3DotProduct(&a->norm,&ap)));
	if (apdotn == aidotn)
		behind = 0;
	else if (apdotn == -aidotn)
		infront = 0;
	}
if (infront)		return 1;
else if (behind)	return -1;
else			return 0;
}


/* S3ComputeBright
 * Computes the brightness from all lights at a given point in view coords
 * with a given normal.
 */
static real S3ComputeBright(S3view *vw, S3polygon *poly, S3point *p,
			    S3point *nm)
{
real brite = Real0;			/* total brightness */
int i;
S3point n = *nm;			/* copy of normal */

S3Normalise(&n);
for(i=0; i<vw->ltc; i++) {
	S3light *light = vw->lt[i];	/* light under consideration */
	S3point l;			/* vector from point to light */
	register real ndotl, ndote;

	l.x = SubReal(light->vpos.x,p->x);
	l.y = SubReal(light->vpos.y,p->y);
	l.z = SubReal(light->vpos.z,p->z);
	S3Normalise(&l);

	ndotl = S3DotProduct(&n,&l);
	ndote = S3DotProduct(&n,p);
	if (RealEq( SgnReal(ndotl), SgnReal(ndote) ))
		continue;	/* on the dark side */

	if (poly->type & pt_shiny) {
		/* Use specular lighting model
		 */
		S3point r;		/* reflection vector */
		real rdotp;
		S3point pp;		/* vector from eye to point */

		r = n;
		pp = *p;
		S3Scale(&r,MulReal(DblToReal(2.0),ndotl));
		S3Subtract(&r,&l);
		S3Normalise(&r);
		S3Normalise(&pp);
		rdotp = S3DotProduct(&r,&pp);
		if (!RealGt(rdotp,Real0)) {
			brite = AddReal( brite, MulReal( SqrReal(rdotp) ,
						 light->intensity ) );
			}
		}

	/* Add diffuse component to brightness
	 */
	ndotl = AbsReal(ndotl);
	brite = AddReal( brite , MulReal( AbsReal(ndotl) ,
					  light->intensity ) );
	}
brite = AddReal(brite,vw->ambient);
if (RealGt(brite,Real1))
	brite = Real1;
return brite;
}


/* S3ScanConvert
 * Faster z-buffer scan convert, from Graphic Gems I
 */
void S3ScanConvert(S3view *vw, S3point *spt, int sptc, S3polygon *poly,
		   real *ssh)
{
register int i,li,ri,y,ly,ry,top,rem;
real ymin;
S3vertex l,r,dl,dr;
S3vertex vert[MAX_VERTS_PER_FACE];	/* vertex structs for scan converter */
real red,grn,blu;			/* for 24-bit colour */
real basecol, shadecol;			/* for non-24 bit colour */

/* Initialize vertex structures with colour and depth info
 */
if (vw->im->depth == 24) {
	long col24 = S3GetColour(poly->colour);

	red = IntToReal((col24&0xff0000)>>16);
	grn = IntToReal((col24&0x00ff00)>>8);
	blu = IntToReal(col24&0x0000ff);
	}
else {
	basecol  = IntToReal((poly->colour / 32)*32);
	shadecol = IntToReal(poly->colour % 32);
	}
for(i=0; i<sptc; i++) {
	/*real fz = fpt[i].z;*/

	vert[i].x = spt[i].x;
	vert[i].y = spt[i].y;
	vert[i].z = spt[i].z;
	/*
	vert[i].x = IntToReal(spt[i].x);
	vert[i].y = IntToReal(spt[i].y);
	vert[i].z = DivReal( fz , AddReal(Real1,DivReal(fz,vw->d)) );
	*/
	if (vw->im->depth == 24) {
		vert[i].r = MulReal(red,ssh[i]);
		vert[i].g = MulReal(grn,ssh[i]);
		vert[i].b = MulReal(blu,ssh[i]);
		}
	else {
		vert[i].c = AddReal(basecol,MulReal(shadecol,ssh[i]));
		}
	}

/* Find lowest y valued vertex
 */
ymin = IntToReal(32767);
for(i=0; i<sptc; i++)
	if (RealLt(vert[i].y,ymin)) {
		ymin = vert[i].y;
		top = i;
		}

li = ri = top;
rem = sptc;
y = ceil(RealToDbl(ymin)-0.5);
ly = ry = y-1;

while(rem > 0) {
	while(ly<=y && rem>0) {
		rem--;
		i = li-1;
		if (i<0) i = sptc-1;
		S3incrementalize_y(&vert[li],&vert[i],&l,&dl,y,vw);
		ly = floor(RealToDbl(vert[i].y) + 0.5);
		li = i;
		}
	while(ry<=y && rem>0) {
		rem--;
		i = ri+1;
		if (i>=sptc) i = 0;
		S3incrementalize_y(&vert[ri],&vert[i],&r,&dr,y,vw);
		ry = floor(RealToDbl(vert[i].y) + 0.5);
		ri = i;
		}
	while(y<ly && y<ry) {
		if (y>=0 && y<vw->height) {
			if (RealLt(l.x,r.x))
				S3scanline(y,&l,&r,vw,poly);
			else
				S3scanline(y,&r,&l,vw,poly);
			}
		y++;
		S3increment(&l,&dl,vw,1);
		S3increment(&r,&dr,vw,1);
		}
	}
}


static void S3scanline(int y, S3vertex *l, S3vertex *r, S3view *vw,
		       S3polygon *poly)
{
register int x,lx,rx;
S3vertex p,dp;
int hicolour = poly->obj->highlight;

lx = ceil(RealToDbl(l->x) + 0.5);
if (lx < 0) lx = 0;
rx = floor(RealToDbl(r->x) + 0.5);
if (rx > vw->width-1) rx = vw->width-1;
if (lx > rx) return;
S3incrementalize_x(l,r,&p,&dp,lx,vw);
for(x=lx; x<=rx; x++) {
	int czp = y*vw->width + x;

	/*
	if (vw->zbuf[czp] != DblToReal(CoordMax))
		printf("old = %f, new = %f\n",vw->zbuf[czp],p.z);
	*/
	if (RealLt(p.z,vw->zbuf[czp])) {
		vw->pbuf[czp] = poly;
		if (!((poly->type & pt_glass) && (x+y)%2)) {
			/* Plot pixel, possibly highlighted
			 */
			if (hicolour != -1 && (x+y)%2) {
				XPutPixel(vw->im,x,y,S3GetColour(hicolour));
				}
			else if (vw->im->depth == 24) {
				XPutPixel(vw->im,x,y,(RealToInt(p.r)<<16) +
						     (RealToInt(p.g)<<8) +
						     (RealToInt(p.b)));
				}
			else {
				XPutPixel(vw->im,x,y,
					  S3GetColour(RealToInt(p.c)));
				}
			vw->zbuf[czp] = p.z;
			}
		}
	S3increment(&p,&dp,vw,0);
	}
}


static void S3incrementalize_y(S3vertex *p1, S3vertex *p2, S3vertex *p,
			       S3vertex *dp, int y, S3view *vw)
{
real dy,frac;

dy = SubReal(p2->y, p1->y);
if (RealEq(dy,Real0)) dy = Real1;
frac = SubReal( DblToReal(y+0.5) , p1->y );

dp->x = DivReal( SubReal(p2->x,p1->x) , dy );
p->x  = AddReal( p1->x , MulReal(dp->x,frac) );
dp->z = DivReal( SubReal(p2->z,p1->z) , dy );
p->z  = AddReal( p1->z , MulReal(dp->z,frac) );
if (vw->im->depth == 24) {
	dp->r = DivReal( SubReal(p2->r,p1->r) , dy );
	p->r  = AddReal( p1->r , MulReal(dp->r,frac) );
	dp->g = DivReal( SubReal(p2->g,p1->g) , dy );
	p->g  = AddReal( p1->g , MulReal(dp->g,frac) );
	dp->b = DivReal( SubReal(p2->b,p1->b) , dy );
	p->b  = AddReal( p1->b , MulReal(dp->b,frac) );
	}
else {
	dp->c = DivReal( SubReal(p2->c,p1->c) , dy );
	p->c  = AddReal( p1->c , MulReal(dp->c,frac) );
	}
}


static void S3incrementalize_x(S3vertex *p1, S3vertex *p2, S3vertex *p,
			       S3vertex *dp, int x, S3view *vw)
{
real dx,frac;

dx = SubReal(p2->x,p1->x);
if (RealEq(dx,Real0)) dx = Real1;
frac = SubReal( DblToReal(x+0.5) , p1->x );

dp->z = DivReal( SubReal(p2->z,p1->z) , dx );
p->z  = AddReal( p1->z , MulReal(dp->z,frac) );
if (vw->im->depth == 24) {
	dp->r = DivReal( SubReal(p2->r,p1->r) , dx );
	p->r  = AddReal( p1->r , MulReal(dp->r,frac) );
	dp->g = DivReal( SubReal(p2->g,p1->g) , dx );
	p->g  = AddReal( p1->g , MulReal(dp->g,frac) );
	dp->b = DivReal( SubReal(p2->b,p1->b) , dx );
	p->b  = AddReal( p1->b , MulReal(dp->b,frac) );
	}
else {
	dp->c = DivReal( SubReal(p2->c,p1->c) , dx );
	p->c  = AddReal( p1->c , MulReal(dp->c,frac) );
	}
}


static void S3increment(S3vertex *p, S3vertex *dp, S3view *vw, int dox)
{
if (dox)
	p->x = AddReal(p->x,dp->x);
p->z = AddReal(p->z,dp->z);
if (vw->im->depth == 24) {
	p->r = AddReal(p->r,dp->r);
	p->g = AddReal(p->g,dp->g);
	p->b = AddReal(p->b,dp->b);
	}
else {
	p->c = AddReal(p->c,dp->c);
	}
}

