#include "s3dt.h"
#include "s3dtint.h"

/* S3MultPoints
 * Multiply all points in an object (but NOT it's children) by a matrix
 */
void S3MultPoints(S3object *obj, real mat[3][3])
{
int i;

for(i=0; i<obj->ptc; i++) {
	S3point tmp;

	tmp = *obj->pt[i];
	S3MMult(&tmp,mat,obj->pt[i]);
	}
for(i=0; i<obj->facec; i++) {
	S3ComputeFace(obj,obj->face[i]);
	}
}


/* S3ApplyRotation
 * Common code from rotation functions for applying a rotation matrix
 * to the world-coord points of an object
 */
static void S3ApplyRotation(S3object *ob, real mat[3][3], double cx,
			    double cy, double cz)
{
int i;

S3MovePoints(ob,RealToDbl(ob->pos.x),RealToDbl(ob->pos.y),RealToDbl(ob->pos.z));
S3MovePoints(ob,-cx,-cy,-cz);
S3MultPoints(ob,mat);
S3MovePoints(ob,cx,cy,cz);
S3MovePoints(ob,-RealToDbl(ob->pos.x),-RealToDbl(ob->pos.y),
	     -RealToDbl(ob->pos.z));
for(i=0; i<ob->soc; i++)
	S3ApplyRotation(ob->so[i],mat,cx,cy,cz);
}


/* S3RotateXPos
 * Rotate an object and it's children about the x axis through a point
 * in world coordinates.
 */
void S3RotateXPos(S3object *ob, double ang, double cx, double cy, double cz)
{
real mat[3][3] = {	{Real1, Real0, Real0},
			{Real0, Real1, Real0},
			{Real0, Real0, Real1} };
double rad = dtor(ang);

mat[1][1] = DblToReal(cos(rad)); mat[2][1] = DblToReal(sin(rad));
mat[1][2] = -mat[2][1];		 mat[2][2] = mat[1][1];
S3ApplyRotation(ob,mat,cx,cy,cz);
}


/* S3RotateYPos
 * Rotate an object and it's children about the y axis through a point
 * in world coordinates.
 */
void S3RotateYPos(S3object *ob, double ang, double cx, double cy, double cz)
{
real mat[3][3] = {	{Real1, Real0, Real0},
			{Real0, Real1, Real0},
			{Real0, Real0, Real1} };
double rad = dtor(ang);

mat[0][0] = DblToReal(cos(rad)); mat[2][0] = DblToReal(sin(rad));
mat[0][2] = -mat[2][0];		 mat[2][2] = mat[0][0];
S3ApplyRotation(ob,mat,cx,cy,cz);
}


/* S3RotateZPos
 * Rotate an object and it's children about the z axis through a point
 * in world coordinates.
 */
void S3RotateZPos(S3object *ob, double ang, double cx, double cy, double cz)
{
real mat[3][3] = {	{Real1, Real0, Real0},
			{Real0, Real1, Real0},
			{Real0, Real0, Real1} };
double rad = dtor(ang);

mat[0][0] = DblToReal(cos(rad)); mat[1][0] = DblToReal(sin(rad));
mat[0][1] = -mat[1][0];		 mat[1][1] = mat[0][0];
S3ApplyRotation(ob,mat,cx,cy,cz);
}


/* S3RotateX
 * Like S3RotateXPos, but rotates about the object's position
 */
void S3RotateX(S3object *ob, double ang)
{
S3RotateXPos(ob,ang,RealToDbl(ob->pos.x),RealToDbl(ob->pos.y),
	     RealToDbl(ob->pos.z));
}


/* S3RotateY
 * Like S3RotateYPos, but rotates about the object's position
 */
void S3RotateY(S3object *ob, double ang)
{
S3RotateYPos(ob,ang,RealToDbl(ob->pos.x),RealToDbl(ob->pos.y),
	     RealToDbl(ob->pos.z));
}


/* S3RotateZ
 * Like S3RotateZPos, but rotates about the object's position
 */
void S3RotateZ(S3object *ob, double ang)
{
S3RotateZPos(ob,ang,RealToDbl(ob->pos.x),RealToDbl(ob->pos.y),
	     RealToDbl(ob->pos.z));
}


/* S3RotateAbout
 * Rotate an object by a given angle about some axis, about some position.
 */
void S3RotateAbout(S3object *ob, double ang, double xa, double ya, double za,
		   double x, double y, double z)
{
real mat[3][3];

double rad = dtor(ang);

mat[0][0] = DblToReal(xa*xa + cos(rad)*(1-xa*xa));
mat[0][1] = DblToReal(xa*ya*(1-cos(rad)) - za*sin(rad));
mat[0][2] = DblToReal(za*xa*(1-cos(rad)) + ya*sin(rad));

mat[1][0] = DblToReal(xa*ya*(1-cos(rad)) + za*sin(rad));
mat[1][1] = DblToReal(ya*ya + cos(rad)*(1-ya*ya));
mat[1][2] = DblToReal(ya*za*(1-cos(rad)) - xa*sin(rad));

mat[2][0] = DblToReal(za*xa*(1-cos(rad)) - ya*sin(rad));
mat[2][1] = DblToReal(ya*za*(1-cos(rad)) + xa*sin(rad));
mat[2][2] = DblToReal(za*za + cos(rad)*(1-za*za));

S3ApplyRotation(ob,mat,x,y,z);
}


/* S3SetRotateAbout
 * Set the rotation of an object to a given angle about the given axis and
 * position. This function will undo the previous rotate from a call to
 * S3SetRotateAbout() before applying the new one.
 */
void S3SetRotateAbout(S3object *ob, double ang, double xa, double ya, double za,
		      double x, double y, double z)
{
S3RotateAbout(ob,-(ob->rotang),ob->rax,ob->ray,ob->raz,ob->rx,ob->ry,ob->rz);
S3RotateAbout(ob,ang,xa,ya,za,x,y,z);
ob->rotang = ang;
ob->rax = xa;
ob->ray = ya;
ob->raz = za;
ob->rx = x;
ob->ry = y;
ob->rz = z;
}
