#include "s3dt.h"
#include "s3dtint.h"

/* S3SetObjectAttribs
 * Set attributes of an object from a zero-terminated list of arguments
 */
void S3SetObjectAttribs(S3object *obj, ...)
{
va_list ap;

va_start(ap,obj);
S3VSetObjectAttribs(obj,ap);
va_end(ap);
}


/* S3VSetObjectAttribs
 * Set attribs of an object from a varargs list.
 */
void S3VSetObjectAttribs(S3object *obj, va_list ap)
{
int attrib;

while((attrib = va_arg(ap,int)) != at_end) {
	switch(attrib) {
	case at_id: {
		/* Set the object id
		 */
		S3SetObjectId(obj,va_arg(ap,int));
		break;
		}
	case at_pos: {
		/* Set the object position
		 */
		double x = va_arg(ap,double);
		double y = va_arg(ap,double);
		double z = va_arg(ap,double);

		S3SetObjectPos(obj,x,y,z);
		break;
		}
	case at_addpoint: {
		/* Add a point to the object
		 */
		double x = va_arg(ap,double);
		double y = va_arg(ap,double);
		double z = va_arg(ap,double);

		S3AddPoint(obj,x,y,z);
		break;
		}
	case at_addface: {
		/* Add a face to the object
		 */
		int *vert = va_arg(ap,int *);
		int vertc = va_arg(ap,int);
		int colour = va_arg(ap,int);
		int type = va_arg(ap,int);

		S3AddPolygon(obj,vert,vertc,colour,type);
		break;
		}
	case at_name: {
		/* Set the object's name
		 */
		S3SetObjectName(obj,va_arg(ap,char *));
		break;
		}

		}	/* end switch */
	}	/* end while */
}


/* S3SetObjectId
 * Change the id of an object.
 */
void S3SetObjectId(S3object *obj, int id)
{
obj->id = id;
}


/* S3SetObjectPos
 * Change the position of an object and all it's sub-objects.
 */
void S3SetObjectPos(S3object *obj, double x, double y, double z)
{
double dx = x - RealToDbl(obj->pos.x);
double dy = y - RealToDbl(obj->pos.y);
double dz = z - RealToDbl(obj->pos.z);
int i;

S3DeltaObjectPos(obj,dx,dy,dz);
}


/* S3DeltaObjectPos
 * Change the position of an object and all it's children by x,y,z
 */
void S3DeltaObjectPos(S3object *obj, double x, double y, double z)
{
int i;

obj->pos.x = AddReal( obj->pos.x, DblToReal(x) );
obj->pos.y = AddReal( obj->pos.y, DblToReal(y) );
obj->pos.z = AddReal( obj->pos.z, DblToReal(z) );
obj->rx +=x; obj->ry +=y; obj->rz +=z;
for(i=0; i<obj->soc; i++)
	S3DeltaObjectPos(obj->so[i],x,y,z);
}


/* S3SetObjectName 
 * Change the name of an object
 */
void S3SetObjectName(S3object *obj, char *name)
{
strcpy(obj->name,name);
}


/* S3AddChildObject
 * Add a child object to a parent
 */
void S3AddChildObject(S3object *child, S3object *parent)
{
child->par = parent;

if (!parent->so) {
	parent->somax = 10;
	parent->so = malloc(parent->somax*sizeof(S3object *));
	}
else if (parent->soc >= parent->somax) {
	parent->somax = parent->somax*2 + 1;
	parent->so = realloc(parent->so,parent->somax*sizeof(S3object *));
	}
parent->so[parent->soc++] = child;
}


/* S3SetObjectColour
 * Set the colour of every face in an object
 */
void S3SetObjectColour(S3object *obj, int col)
{
int i;

for(i=0; i<obj->facec; i++)
	S3SetFaceColour(obj,i,col);
}


/* S3SetObjectInvisible
 * Make an object (and all it's children) visible or invisible
 */
void S3SetObjectInvisible(S3object *obj, int inv)
{
int i;

obj->invis = inv;
for(i=0; i<obj->soc; i++)
	S3SetObjectInvisible(obj->so[i],inv);
}


/* S3SetObjectPType
 * Change the surface type of all faces in an object, and all it's children
 */
void S3SetObjectPType(S3object *obj, int ptype, int mask)
{
int i;

for(i=0; i<obj->facec; i++)
	S3SetFacePType(obj,i,ptype,mask);
for(i=0; i<obj->soc; i++)
	S3SetObjectPType(obj->so[i],ptype,mask);
}


/* S3SetFaceColour
 * Change the colour of a particular face of an object 
 */
void S3SetFaceColour(S3object *obj, int fnum, int colour)
{
obj->face[fnum]->colour = colour;
}


/* S3SetFacePType
 * Change some of the type flags of a particular face of an object
 */
void S3SetFacePType(S3object *obj, int fnum, int ptype, int mask)
{
obj->face[fnum]->type = (obj->face[fnum]->type & ~mask) | (ptype & mask);
}

