/* s3dt.h
 *
 * Function prototypes and data structures for programmers making use
 * of the object modelling part of s3dt.
 */
#if !defined(S3DT_H)
#define S3DT_H

#include <stdio.h>

/* Global constants for use inside and outside the library
 */
#define OBJTAB_SIZE 100

/* A fixed or floating point value
 */
#if defined(S3FIXED)
typedef int real;		/* use fixed-point reals */
#define CoordMax 32767/2
#else
typedef float real;		/* use floating-point reals */
#include <float.h>
#define CoordMax FLT_MAX/2
#endif	/* defined(S3FIXED) */

/* S3point
 * One point in 3d space
 */
typedef struct S3point {
	real x,y,z;
	} S3point;

struct S3object;		/* forward def */

/* S3polygon
 * One polygon, line or face of an object. The surface of a polygon is
 * determined by ORing the flags below together.
 */
#define pt_plane	0x01	/* filled polygon */
#define pt_glass	0x02	/* 50% transparent */
#define pt_lighted	0x04	/* coloured by angle from light */
#define pt_oneside	0x08	/* backface-culled */
#define pt_shaded	0x10	/* smoothly shaded */
#define pt_shiny	0x20	/* use specular lighting */

typedef struct S3polygon {
	int *vert;		/* list of points making up this face */
	int vertc;		/* number of points in vert */
	int colour;		/* colour of this face */
	int type;		/* logical or of flags above */
	real dist;		/* distance to poly for sorting */
	S3point norm;		/* face normal */
	S3point cen;		/* face centroid */
	struct S3object *obj;	/* object this face is part of */
	struct S3polygon *fnext;/* next polygon farpoint view list */
	int mvd;		/* has poly been moved to end of list? */
	int xmin,xmax,ymin,ymax;/* extent of polygon on screen */
	int onscreen;		/* nonzero if this polygon got drawn */
	} S3polygon;


/* S3object
 * One vehicle / planet / whatever in the 3d world.
 * An object can have a number of sub-objects that are moved with it
 * (but not freed or copied. This is (for now) the programmer's responsibility)
 */
typedef struct S3object {
	int id;			/* unique identifier of this object */
	char name[256];		/* unique object name */
	S3point **pt;		/* array of points used in object */
	S3point **vpt;		/* points converted to view coords */
	int ptc;		/* number of points in pt */
	S3polygon **face;	/* array of faces making up object */
	int facec;		/* number of faces in face */
	real dist;		/* distance from viewpoint */
	real radius;		/* radius of bounding sphere */
	S3point pos;		/* position of object in 3d space */
	void *extra_data;	/* user data for this object */
	struct S3object *next;	/* next object in view list */
	struct S3object *hnext;	/* next object in hash table */
	struct S3object **so;	/* array of sub-objects */
	int soc;		/* number of sub-objects */
	int somax;		/* maximum sub-objects before re-allocation */
	struct S3object *par;	/* parent object */
	int invis;		/* is this object invisible? */
	int highlight;		/* colour to highlight in, or -1 for none */
	int bminx,bminy;	/* top-left of on-screen bounding box */
	int bmaxx,bmaxy;	/* bottom-right of on-screen bounding box */
	int onscreen;		/* non-zero is and part of object drawn */
	double rotang;		/* current angle of rotation */
	double rax, ray, raz;	/* current axis of rotation */
	double rx, ry, rz;	/* current point of rotation */
	} S3object;


/* S3objtab
 * A table of objects hashed by id or by name
 */
#define tt_id		1	/* objects hashed by id */
#define tt_name		2	/* objects hashed by name */

typedef struct S3objtab {
	int type;			/* one of the definitions above */
	S3object *tab[OBJTAB_SIZE];	/* hash table */
	S3object *head;			/* head of object list */
	S3polygon *fph;			/* polygon list sorted by far point */
	S3polygon **fa;			/* array of polygons in far list */
	int fas;			/* max size of fa */
	} S3objtab;


/* Function type for creating a new extra data buffer
 */
typedef void *(*S3ecreate)(void);
/* Function type for freeing the extra_data buffer
 */
typedef void (*S3efree)(void *);
/* Function type for calling when an extra_data item is parsed
 */
typedef void (*S3eadd)(S3object *, char *, char **, int);
/* Function type for calling when extra_data is to be written
 */
typedef void (*S3ewrite)(int, void *, void (*)(int, char *, char **, int));


/* S3attribtype
 * An indicator of which attribute to set for S3CreateObject()
 */
typedef enum {	at_end, at_id, at_pos, at_addpoint, at_addface, at_name }
	S3attribtype;


/* Function prototypes
 */
S3object *S3CreateObject(char *name, ...);

void S3SetObjectAttribs(S3object *obj, ...);
void S3SetObjectId(S3object *obj, int id);
void S3SetObjectPos(S3object *obj, double x, double y, double z);
void S3DeltaObjectPos(S3object *obj, double x, double y, double z);
void S3SetObjectName(S3object *obj, char *name);
void S3AddChildObject(S3object *child, S3object *parent);
void S3SetObjectColour(S3object *obj, int col);
void S3SetObjectInvisible(S3object *obj, int inv);
void S3SetObjectPType(S3object *obj, int ptype, int mask);
void S3SetFaceColour(S3object *obj, int fnum, int colour);
void S3SetFacePType(S3object *obj, int fnum, int ptype, int mask);

void S3AddPoint(S3object *obj, double x, double y, double z);
void S3AddPoints(S3object *obj, int ptc, S3point pt[]);

void S3AddPolygon(S3object *obj, int *vert, int vertc, int colour, int type);
void S3AddFace(S3object *obj, int colour, int type, int vertc, ...);

S3objtab *S3CreateObjectTable(int type);
void S3FreeObjectTable(S3objtab *tab);
void S3AddObject(S3objtab *tab, S3object *obj);
void S3DeleteObject(S3objtab *tab, S3object *ob);
S3object *S3LookupObject(S3objtab *tab, ...);
void S3AddObjectAndChildren(S3objtab *tab, S3object *obj);
void S3DeleteObjectAndChildren(S3objtab *tab, S3object *obj);
void S3AddViewFace(S3objtab *tab, S3polygon *p);
void S3DelViewFace(S3objtab *tab, S3polygon *p);
void S3ClearObjectTable(S3objtab *tab);

void S3FreeObject(S3object *);

S3object *S3CreateCube(int colour, int ftype);
S3object *S3CreateCone(int colour, int ftype);
S3object *S3CreatePlane(int colour, int ftype);
S3object *S3CreateSphere(int colour, int ftype);
S3object *S3CreateTube(int colour, int ftype);
S3object *S3CreateDisc(int colour, int ftype);

void S3SetEFreeFunc(S3efree eff);
S3efree S3GetEFreeFunc(void);
void S3SetECreateFunc(S3ecreate ecf);
S3ecreate S3GetECreateFunc(void);
void S3SetEAddFunc(S3eadd eaf);
S3eadd S3GetEAddFunc(void);
void S3SetEWriteFunc(S3ewrite ewf);
S3ewrite S3GetEWriteFunc(void);

double S3GetTime(void);

void S3CountFps(void);

int S3ParseFd(int filed, S3objtab *ohead);
int S3ParseFile(char *fname, S3objtab *ohead);

char *S3Itoa(int i);

void S3MovePoints(S3object *ob, double x, double y, double z);
void S3MovePoint(S3object *ob, int p, double x, double y, double z);
void S3PositionPoint(S3object *ob, int p, double x, double y, double z);

void S3ScaleObject(S3object *ob, double sc);
void S3ScaleObject3(S3object *ob, double x, double y, double z);
void S3ScaleObjectAbout(S3object *ob, double cx, double cy, double cz,
				      double sc);
void S3ScaleObjectAbout3(S3object *ob, double cx, double cy, double cz,
				       double x, double y, double z);

void S3GetObjectId(S3object *ob, int *id);
void S3GetObjectPos(S3object *ob, double *x, double *y, double *z);
void S3GetObjectParent(S3object *ob, S3object **par);
void S3GetObjectChildren(S3object *ob, S3object ***children, int *cc);
void S3GetObjectName(S3object *ob, char **name);
void S3GetObjectPointC(S3object *ob, int *ptc);
void S3GetObjectFaceC(S3object *ob, int *facec);

void S3ObjectCentroid(S3object *obj, double *cx, double *cy, double *cz);

void S3RotateXPos(S3object *ob, double ang, double cx, double cy, double cz);
void S3RotateYPos(S3object *ob, double ang, double cx, double cy, double cz);
void S3RotateZPos(S3object *ob, double ang, double cx, double cy, double cz);
void S3RotateX(S3object *ob, double ang);
void S3RotateY(S3object *ob, double ang);
void S3RotateZ(S3object *ob, double ang);
void S3RotateAbout(S3object *ob, double ang, double xa, double ya, double za,
		   double x, double y, double z);
void S3SetRotateAbout(S3object *ob, double ang, double xa, double ya, double za,
		      double x, double y, double z);

void S3WriteShowPoly(S3objtab *ot, FILE *fp);

void S3BoundingBox(S3object *obj, double *x1, double *y1, double *z1,
				  double *x2, double *y2, double *z2);
void S3BoundingBoxTable(S3objtab *ot, double *x1, double *y1, double *z1,
				      double *x2, double *y2, double *z2);

#endif /* !defined(S3DT_H) */

