/* s3dtint.h
 *
 * Internal includes, prototypes and definitions for use within the
 * s3dt library only
 */
#if !defined(S3DTINT_H)
#define S3DTINT_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "s3dt.h"
#include "s3view.h"

/* Real number macros and functions for use only within the library
 */
#if defined(S3FIXED)
/* Real numbers are 32-bit integers
 */
#define Real1		65536
#define Real0		0
#define DblToReal(x)	((int)((x)*Real1))
#define RealToDbl(x)	((x)/(double)Real1)
#define IntToReal(x)	((x)*Real1)
#define RealToInt(x)	((x)/Real1)
#define ScaleReal(x,s)	((x)*(s))
#define AddReal(x,y)	((x)+(y))
#define SubReal(x,y)	((x)-(y))
#if 0 && LONG_MAX == 9223372036854775807	/* 64 bit longs to play with */
#define MulReal(x,y)	((int)((long)(x)*(long)(y)/((long)Real1*(long)Real1)))
#define DivReal(x,y)	((int)((long)(x)*(long)Real1 / (long)(y)))
#else
#define Hword(x)	(((x)&0xffff0000)>>16)
#define Lword(x)	((x)&0x0000ffff)
/*
#define MulReal(x,y)	((Hword(x)*Hword(y) << 16) + (Hword(x)*Lword(y)) +\
			 (Lword(x)*Hword(y))       + (Lword(x)*Lword(y) >> 16))
*/
#define MulReal(x,y)	(((double)(x)*(double)(y))/65536.0)
#define DivReal(x,y)	(DblToReal((double)(x)/(double)(y)))
#endif
#define MulReal3(x,y,z)	(MulReal(x,MulReal(y,z)))
#define AddReal3(x,y,z)	(AddReal(x,AddReal(y,z)))
#define SqrReal(x)	(MulReal(x,x))
#define SqrtReal(x)	(DblToReal(sqrt(RealToDbl(x))))
#define SgnReal(x)	((x) == 0 ? Real0 : ((x) > 0 ? Real1 : -Real1))
#define AbsReal(x)	((x) < 0 ? -(x) : (x))

#define RealLt(x,y)	((x)<(y))
#define RealGt(x,y)	((x)>(y))
#define RealEq(x,y)	((x)==(y))

#else
/* Real numbers are floating-point
 */
#define Real1		1.0
#define Real0		0.0
#define DblToReal(x)	((real)(x))
#define RealToDbl(x)	((double)(x))
#define IntToReal(x)	((real)(x))
#define RealToInt(x)	((int)(x))
#define ScaleReal(x,s)	((x)*(s))
#define AddReal(x,y)	((x)+(y))
#define SubReal(x,y)	((x)-(y))
#define MulReal(x,y)	((x)*(y))
#define DivReal(x,y)	((x)/(y))
#define MulReal3(x,y,z)	((x)*(y)*(z))
#define AddReal3(x,y,z)	((x)+(y)+(z))
#define SqrReal(x)	((x)*(x))
#define SqrtReal(x)	(sqrt(x))
#define SgnReal(x)	(RealEq(x,Real0) ? Real0 : (RealGt(x,Real0) ? Real1 : ScaleReal(Real1,-1)))
#define AbsReal(x)	(RealLt(x,Real0) ? ScaleReal(x,-1) : (x))

#define RealLt(x,y)	((x)<(y))
#define RealGt(x,y)	((x)>(y))
#define RealEq(x,y)	((x)==(y))

#endif	/* defined(S3FIXED) */

/* Global constants for use within the library
 */
#define MAX_POINTS_PER_OBJECT 1000
#define MAX_FACES_PER_OBJECT  1000
#define MAX_VERTS_PER_FACE    50
#define VIEW_SCALE            0.333
#ifndef PI
#define PI 3.1415927
#endif

/* Macros for use within the library
 */
#define dtor(x)		((x)/180.0 * PI)
#define abs(x)		((x) < 0 ? -(x) : (x))
#define min(x,y)	((x) < (y) ? (x) : (y))
#define max(x,y)	((x) > (y) ? (x) : (y))
#define sqr(x)		((x) * (x))
#define strequ(a,b)	(!strcmp(a,b))

/* Function prototypes
 */
S3object *S3VCreateObject(char *name, va_list ap);

void *S3Malloc(int size);
void *S3Calloc(int num, int size);
void S3Free(void *ptr);

void S3VSetObjectAttribs(S3object *obj, va_list ap);

void *S3MemDup(void *src, int size);

real S3DotProduct(S3point *p1, S3point *p2);
void S3Normalise(S3point *p);
void S3MMult(S3point *p, real m[3][3], S3point *r);
real S3Magnitude(S3point *p);
void S3CrossProduct(S3point *p1, S3point *p2, S3point *r);
void S3Scale(S3point *p, real s);
void S3Subtract(S3point *a, S3point *b);
void S3Add(S3point *a, S3point *b);

void S3CallObjectFuncs(Display *display, S3view *vw, S3object *ob,S3objtab *oh);
void S3CallBeforeFuncs(Display *display, S3view *vw, S3objtab *oh);
void S3CallAfterFuncs(Display *display, S3view *vw, S3objtab *oh);
S3efree S3GetEFreeFunc(void);
S3ecreate S3GetECreateFunc(void);

void S3BoundingSphere(S3object *obj);
void S3ComputeFace(S3object *ob, S3polygon *p);

void S3MultPoints(S3object *obj, real mat[3][3]);

void S3TrapOn(void);
int S3TrapOff(void);

#endif /* !defined(S3DTINT_H) */

