/* S3view.h
 *
 * Function prototypes and data structures for programmers making use of
 * the graphical part of s3dt.
 */
#if !defined(S3VIEW_H)
#define S3VIEW_H

/* Include files needed for drawing graphics
 */
#undef NeedFunctionPrototypes
#define NeedFunctionPrototypes 1
#include "s3dt.h"
#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#define MAX_LIGHTS_PER_VIEW	30

/* S3light
 * A point in 3d space that emits light
 */
typedef struct S3light {
	int id;			/* unique identifier for this light */
	S3point pos;		/* position of light in 3d space */
	S3point vpos;		/* position in view coordinates */
	real intensity;		/* power of light */
	} S3light;


/* S3view
 * Describes the 3d viewpoint, viewing mode and window to draw it into
 */
typedef enum { vt_normal, vt_close, vt_far, vt_circling } S3viewtype;

typedef enum { rm_wire, rm_plane, rm_zbuf, rm_zbufshm } S3rmodetype;

typedef struct S3view {
	S3point vrp;		/* point looking from */
	S3point u,v,n;		/* viewing direction */
	real d;			/* eye distance */
	real clip;		/* plane to clip polygons to */
	real scale;		/* scale factor for objects in view window */
	real vmat[3][3];	/* matrix for world -> view conversion */
	S3viewtype type;	/* close / far / circling / ... */
	S3rmodetype rmode;	/* wireframe, zbuf, etc.. */
	S3object *tar;		/* object being watched */
	int width,height;	/* size of view window */
	int depth;		/* bitplanes in view window */
	int backcol;		/* background colour of view window */
	Window win;		/* X11 window to blit scene into */
	Pixmap dbuf;		/* buffer for background drawing */
	GC gc;			/* gc for drawing in this view */
	XFontStruct *font;	/* font for in this view */
	S3light **lt;		/* array of lights in this view */
	int ltc;		/* number of lights */
	real ambient;		/* ambient light level */
	real *zbuf;		/* z buffer for rendering */
	XImage *im;		/* image for rasterising into */
	char *dat;		/* data in image */
	XShmSegmentInfo *xshminfo;	/* for shared memory */
	S3polygon **pbuf;	/* polygons at particular pixels */
	} S3view;

/* Prototypes 
 */
Colormap S3CreateDefaultColourmap(Display *display);
Colormap S3CreateSimpleColourmap(Display *display, XColor def[]);
Colormap S3CreateColourmap(Display *display, XColor cmp[]);
Colormap S3CreateWindowsColourmap(Display *display);
void S3SetColour(Display *display, GC gc, int colour);
long S3GetColour(int colour);
long S3ClosestColour(int red, int green, int blue);
long S3ClosestColourPixel(int red, int green, int blue);

S3view *S3CreateView(Display *display, Window win, S3rmodetype rm);
void S3ChangeView(Display *display, S3view *vw, S3rmodetype rm);
void S3SetViewSize(Display *display, S3view *vw, int w, int h);
void S3SetViewVrp(S3view *vw, double x, double y, double z);
void S3SetViewTarget(S3view *vw, S3object *obj);
void S3SetViewN(S3view *vw, double x, double y, double z);
void S3SetViewV(S3view *vw, double x, double y, double z);
void S3SetViewEyeDistance(S3view *vw, double d);
void S3SetViewBackColour(S3view *vw, int col);
void S3SetViewType(S3view *vw, S3viewtype mode);
void S3SetViewAmbient(S3view *vw, double amb);
void S3SetViewScale(S3view *vw, double sc);
void S3FreeView(Display *display, S3view *vw);

void S3RenderView(Display *display, S3view *vw, S3objtab *oh);
void S3RefreshView(Display *display, S3view *vw);
S3object *S3ObjectAtPos(S3view *vw, S3objtab *ot, int x, int y);

S3view *S3CreateSimpleWindow(Display *display, int w, int h);
S3view *S3CreateWindow(Display *display, Window par, int x,int y, int w, int h,
		       S3rmodetype rmode);

int S3AddViewLight(S3view *vw, double x, double y, double z, double intensity);
void S3MoveViewLight(S3view *vw, int num, double x, double y, double z);
void S3FreeViewLight(S3view *vw, int num);

void S3HighLightObject(S3object *obj, int col);
void S3UnHighLightObject(S3object *obj);
void S3UnHighLightAll(S3objtab *ot);

#endif /* !defined(S3VIEW_H) */

