/* sim.h
 *
 * Function prototypes, macros and data structures use by both the sim
 * clients and server.
 */
#if !defined(SIM_H)
#define SIM_H

/* Includes
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <setjmp.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <math.h>
#include <fcntl.h>
#include <dirent.h>
#include <pwd.h>

/* Macro functions
 */
#define strequ(x,y)     (!strcmp(x,y))
#define dtor(x) ((x)/180.0*PI)
#define rtod(x) ((x)/PI*180.0)
#define fixang(x)       (((x) < 0) ? ((x)+360.0) : \
                         ((x) > 360.0) ?  ((x)-360.0) : (x))
#define fixang180(x)	((x) < -180 ? ((x)+360) : \
			 ((x) > 180) ? ((x)-360) : (x))
#define abs(x)  ((x) < 0 ? -(x) : (x))
#define sqr(x)	((x) * (x))
#define max(x,y)	((x)>(y) ? (x) : (y))
#define min(x,y)	((x)<(y) ? (x) : (y))
#if __STDC__ && !defined(UnixCpp)
#define cat2(x,y) x##y
#else
#define cat2(x,y) x/**/y
#endif

/* Constant definitions
 */
#define SIM_PORT 13333          /* default tcp port for servers */
#define SIM_DATA_PORT 14444	/* base port for sending data */
#define SIM_VERSION "2.0"       /* protocol version */
#define MAX_PLAYER_DATA 500000	/* max data a player can carry */
#define FADEOUT_TIME 60*60*24*5	/* inactivity time to make a player hidden */
#define METASERVER_HOST "indy14.cs.monash.edu.au"
#define METASERVER_PORT 8000
#ifndef PI
#define PI 3.1415927
#endif
#define MIME_TYPES ".mime.types"	/* name of mime types file */
#define MAILCAP ".mailcap"		/* name of mailcap file */

/* node
 * The map is made up of nodes, each one at a fixed location and with
 * a unique identifying name. A node can either be a normal 'room' or
 * location on a map, or a link to another named node on another server
 */
struct path;		/* forward definition */
#define MAX_LINKS 30
typedef struct node {
	char name[256];			/* unique name */
	struct node *next;		/* next node is hash table */
	struct node *ne;		/* next node in list */
	double x,y,z;			/* position in 3d space */
	int special;			/* 0=normal, 1=remote link */
	char *desc;			/* long description of node */
	struct {
		char *addr;		/* name of remote server */
		struct in_addr ip;	/* ip address of remote server */
		int port;		/* port to connect to on server */
		} serv;
	struct path *link[MAX_LINKS];	/* connections out from this node */
	int linkc;			/* number of paths */
	} node;

/* path
 * Paths are the links between nodes along which players move. A link
 * can be closed (for example by a door) and thus untraversable.
 */
typedef struct path {
	node *na, *nb;			/* nodes connected */
	int closed;			/* 0=passable, 1=unpassable */
	struct path *next;		/* next path in list */
	} path;

/* person
 * A person is a player or creature of some kind in the simulation.
 * Each person is identified by a unique name, and is represented by a 3d
 * shape of the same name.
 */
#define INV_SIZE 30
typedef struct person {
	char name[256];		/* unique name */
	struct person *next;	/* next person in hash table */
	struct person *ne;	/* next person in list */
	node *pos;		/* position in map */
	char *inv[INV_SIZE];	/* names of objects held by this person */
	int   invc;		/* number of objects in inventory */
	double ang;		/* facing direction */
	char *email;		/* email address (for players only) */
	char *fullname;		/* full name */
	} person;

/* object
 * An object is an item, door or similiar inanimate thing. Each object
 * is located either at a node, or in a player's inventory. Each object
 * is represented by a 3d shape with the same name in the server and
 * graphical clients. 
 */
typedef struct object {
	char name[256];		/* unique identifying name */
	struct object *next;	/* next object in hash table */
	struct object *ne;	/* next object is list */
	char *desc;		/* long description */
	node *pos;		/* node object is at, or NULL */
	int gettable;		/* can this object be picked up? */
	double weight;		/* weight in kg */
	person *inv;		/* person carrying this object, or NULL */
	} object;

/* player
 * A player is a human-controlled person in the simulation. Each player has
 * a unique name, and is represented by a person of that same name.
 * Each player can have data files (gifs, postscript, whatever) associated
 * with him/her, which are stored in a subdirectory of the data directory
 * with the player's name.
 */
typedef struct player {
	char name[256];		/* unique identifying name */
	struct player *next;	/* next player in hash table */
	struct player *ne;	/* next player in list */
	int fd;			/* connection to client */
	struct in_addr ip;	/* ip address of client's machine */
	int ilevel;		/* 3d info level requested */
	int hp;			/* health left */
	int money;		/* cash left */
	char *password;		/* password needed to login */
	int wizard;		/* nonzero if this player has wizard status */
	time_t inout;		/* login/out time (server only) */
	} player;

/* hashtable
 * A generic hash table. Functions from hash.c are used to insert, remove
 * and find things in a table.
 */
#define HASH_TABLE_SIZE 256
typedef struct hashtable {
	void *tab[HASH_TABLE_SIZE];	/* lists of entrys with same hash val */
	void *head;			/* list of entries in table */
	void **tailp;			/* where to add next entry */
	void *np;			/* where position() funcs are up to */
	} hashtable;

/* Function type called on entries by clear_entries function in hash.c
 */
typedef void (*freefunc)(void *);

/* reserved
 * A reserved word and token in a parser's reserved word table
 */
typedef struct {
	char *word;
	int tok;
	} reserved;

/* mesgargs
 * A list of the arguments for a command / request
 */
#define arg_str		1
#define arg_dbl		2
#define arg_int		3
#define arg_strs	4
typedef struct {
	int mtype;		/* req_move, com_ls, etc.. */
	int margs[10];		/* from defines above */
	} mesgargs;

/* From fdio.c
 */
char *fdgets(int fd, char *buf, int len);
int fdprintf(int fd, char *fmt, ...);
int fdgetc(int fd);
void fdputc(int fd, char c);
void fdputs(int fd, char *s);
char *fdgetw(int fd, char *buf, int len);
int fdread(int fd, void *buf, int size, int n);
int fdwrite(int fd, void *buf, int size, int n);
int pipethrough(int fd, int *pid, char *cmd);
int pipeclose(int fd);
int pipeinto(int fd, int *pid, char *cmd);
int copy_data(int fd1, int fd2, int len);
int copy_to_eof(int fd1, int fd2);
void reset_fds(void);

/* From net.c
 */
char *findmyhostname(void);
struct in_addr findmyip();
int open_tcp_dest(int port);
int new_connection(int sfd, char *hostname, struct in_addr *addr);
int open_tcp_send(char *host, int port);
int open_tcp_send_ip(struct in_addr ip, int port);
int send_version(int fd);
int check_version(int fd);
int send_data(int port, int fd, int size);
int recv_data(struct in_addr addr, int port, int fd, int size);
int alloc_port(void);
char *dotted_notation(struct in_addr ip);
int send_password(int fd, char *pass);
int check_password(int fd, char *pass);

/* From misc.c
 */
char *strdupe(char *s);
int open_log_file(char *filename);
void logerror(char *fmt, ...);
void logmessage(char *fmt, ...);
void *smalloc(size_t n);
void *scalloc(size_t c, size_t n);
void *srealloc(void *m, size_t n);
pid_t sfork(void);
void *memdupe(void *mem, int size);
double timenow(void);
char *error_string(int ern);
void parse_string(char *cmd, char ***cmdv, int *cmdc);
char *itoa(int i);
char *ftoa(double f);
int count_chars(char *txt, char c);
int isinteger(char *s);
char *basenam(char *fn);
char *dirnam(char *fn);
void swapdbl(double *a, double *b);
double current_time(void);
void delete_dir(char *dir);
int open_data_file(char *file);
char *expandstr(char *s);
int need_quotes(char *s);
char *get_home_dir(void);
char *tabs(int t);
void strip_return(char *s);
double nrand(void);
void set_signals(void (*sfunc)(int s));
char *concat_vec(char **vecv, int vecc);
void send_message(int fd, int type, va_list ap, mesgargs ma[], int mac);
long directory_size(char *d);
#ifdef NO_STRERROR
char *strerror(int en);
#endif

/* From hash.c
 */
void add_entry(hashtable *ht, void *e);
void *lookup_entry(hashtable *ht, char *name);
void remove_entry(hashtable *ht, void *e);
void reset_position(hashtable *ht);
void *next_position(hashtable *ht);
void clear_entries(hashtable *ht, freefunc ff);

/* From node.c
 */
node *create_node(void);
void add_node(node *n);
node *lookup_node(char *name);
void reset_node(void);
node *next_node(void);
void remove_node(node *n);
void clear_nodes(void);
void free_node(node *n);
int close_node(node *n1, node *n2);
path *create_path(void);
void add_path(path *p);
node *other_end(path *pt, node *n);
void reset_path(void);
path *next_path(void);
void free_path(path *p);
void clear_paths(void);

/* From object.c
 */
object *create_object(void);
void add_object(object *o);
object *lookup_object(char *name);
void reset_object(void);
object *next_object(void);
void remove_object(object *o);
void clear_objects(void);
void free_object(object *o);

/* From person.c
 */
person *create_person(void);
void add_person(person *p);
person *lookup_person(char *name);
void reset_person(void);
person *next_person(void);
void remove_person(person *p);
void clear_persons(void);
void free_person(person *p);
void pickup_item(person *p, object *o);
void putdown_item(person *p, object *o);

/* From player.c
 */
player *create_player(void);
void add_player(player *p);
player *lookup_player(char *name);
void reset_player(void);
player *next_player(void);
void remove_player(player *p);
void clear_players(void);
void free_player(player *p);
player *next_online_player(void);

/* From nodeparser.y
 */
int parse_node(int filed, node **nptr);
int parse_nodes(int filed);
void write_node(int fd, int tl, node *n);
void write_nodes(int fd);

/* From objectparser.y
 */
int parse_object(int filed, object **nptr);
int parse_objects(int filed);
void write_object(int fd, int tl, object *o);
void write_objects(int fd);

/* From pathparser.y
 */
int parse_path(int filed, path **pptr);
int parse_paths(int filed);
void write_path(int fd, int tl, path *p);
void write_paths(int fd);

/* From personparser.y
 */
int parse_person(int filed, person **pptr);
int parse_persons(int filed);
void write_person(int fd, int tl, person *p);
void write_persons(int fd);

/* From playerparser.y
 */
int parse_player(int filed, player **pptr);
int parse_players(int filed);
void write_player(int fd, int tl, player *p);
void write_players(int fd);

/* From texture.c
 */
unsigned char *read_ppm_texture(int fd, int *w, int *h);
void encode_rgb(unsigned char *in, char *out, int size);
void decode_rgb(char *in, unsigned char *out, int size);
void write_encoded_texture(int fd, unsigned char *data, int w, int h);


/* Connection protocol definitions (see protocol.html for description)
 */
#define con_newplayer		1
#define con_oldplayer		2
#define con_locateplayer	3
#define con_transfer_player	4
#define con_transfer_person	5
#define con_transfer_shape	6
#define con_transfer_object	7
#define con_transfer_action	8
#define con_transfer_obdata	9
#define con_transfer_pldata	10
#define con_shutdown		11
#define con_noop		12
#define con_status		13
#define con_getshapes		14

/* 3d info level definitions for old player request
 */
#define lvl_none		1
#define lvl_flat		2
#define lvl_textured		3

/* Client-server request numbers (see protocol.html)
 */
#define req_move		1
#define req_say			2
#define req_yell		3
#define req_get			4
#define req_do			5
#define req_drop		6
#define req_turn		7
#define req_give		8
#define req_upload		9
#define req_send		10
#define req_actlist		11
#define req_ls			12
#define req_rm			13
#define req_suicide		14
#define req_avatar		15
#define req_setpass		16
#define req_email		17
#define req_fullname		18

/* Server-client request numbers
 */
#define com_move		1
#define com_msg			2
#define com_err			3
#define com_addobject		4
#define com_addperson		5
#define com_addshape		6
#define com_moveobject		7
#define com_moveperson		8
#define com_moveshape		9
#define com_delobject		10
#define com_delperson		11
#define com_delshape		12
#define com_newserv		13
#define com_getobject		14
#define com_dropobject		15
#define com_rotshape		16
#define com_giveobject		17
#define com_displaymime		18
#define com_text		19
#define com_actlist		20
#define com_colshape		21
#define com_shdshape		22
#define com_ls			23
#define com_chpath		24
#define com_shutdown		25
#define com_email		26
#define com_fullname		27
#define com_teleport		28

/* Server-client error numbers
 */
#define err_nopath		0	/* path to move along doesn't exist! */
#define err_blocked		1	/* path is blocked */
#define err_notfound		2	/* object not found */
#define err_down		3	/* remote server is down */
#define err_invfull		4	/* no space in inventory */
#define err_donthave		5	/* object not in inventory */
#define err_nothere		6	/* person not nearby */
#define err_nofile		7	/* player data file not found */
#define err_noaction		8	/* cannot do action to object */
#define err_execfailed		9	/* execution of do action failed */
#define err_notonline		10	/* player not logged in */
#define err_cantget		11	/* can't get object */
#define err_parse		12	/* error parsing input from player */
#define err_maxdata		13	/* data quota exceeded */
#define err_avatar		14	/* avatar is not valid */

/* Shape surface types, used by com_shdshape and in shape descriptions
 */
#define sft_flat		0
#define sft_glass		1
#define sft_shiny		2

#endif /* !defined(SIM_H) */
