/* sim3d.c
 *
 * main() function for s3dt sim client
 */
#include "sim.h"
#include "client.h"
#include "3d.h"

static void wait_for_activity(int fd1, int fd2, int fd3);
static void check_server_command(int *fdptr, player **pptr, person **prptr,
				 plview *plv);
static void move_frame(plview *plv);
static void client_killed(int s);
static void reaper(int sig);

int parse_debug		= 0;
int fd_debug		= 0;
int is_sim_server       = 0;	/* this is a client */

int main(int argc, char **argv)
{
char *name = NULL;	/* name of this player */
int fd;			/* connection to server */
player *me;		/* info on this player */
person *pr;		/* person representing player */
plview plv;		/* where this player is looking */
int i;

/* Read prefs and command line args
 */
if (!parse_client_args(argc, argv, &name)) {
	exit(1);
	}

/* Find a name for this player
 */
if (!name) {
	if ((name = getenv("USER")) == NULL) {
		if ((name = getenv("LOGNAME")) == NULL) {
			logerror("No name given or found");
			exit(1);
			}
		}
	}

if (!init_ui()) {
	logerror("Error initialising user interface");
	exit(1);
	}

/* setup signal handlers
 */
set_signals(client_killed);
signal(SIGCHLD,reaper);

/* Retreive player info and nodes from server
 */
waiting_cursor();
logmessage("Connecting to sim server. Please wait..");
if ((fd = connect_server(name,&me,supported_3d(),parse_3ds)) == -1) {
	logerror("Error connecting to server");
	exit(3);
	}
logmessage("Connected to %s port %d",get_sprefs(NULL)->server,
				     get_sprefs(NULL)->port);
save_sprefs();	/* changed by connect_server() */
pr = lookup_person(me->name);
visible_3d(me->name,0);

create_base_ui_map();	/* add nodes and paths to map */

/* Init player's view
 */
plv.x2   = pr->pos->x;
plv.y2   = pr->pos->y;
plv.z2   = pr->pos->z + PLAYER_HEIGHT;
plv.dst  = plv.range = 1.0;
plv.ang  = pr->ang;
plv.tilt = 0.0;
plv.move = 0;
need_redraw();
draw_frame(me,&plv);
position_ui_view(&plv);

/* Init item list in ui
 */
fill_hereinv_list(pr);

normal_cursor();
/* Main loop
 */
while(1) {
	int ac;
	int moving = 0;

	if (plv.dst < plv.range) {
		/* Move viewpoint smoothly forward
		 */
		move_frame(&plv);
		need_redraw();
		moving = 1;
		}
	if (!moving)
		wait_for_activity(interface_fd(),fd,threed_fd());
	check_3d_ui(fd,me,pr,&plv);
	check_ui(fd,me,pr,&plv);
	check_server_command(&fd,&me,&pr,&plv);
	draw_frame(me,&plv);
	}
}


/* wait_for_activity
 * Wait on user interface, network and 3d fd's, returning only when there is
 * some activity.
 */
static void wait_for_activity(int fd1, int fd2, int fd3)
{
fd_set readfds;

FD_ZERO(&readfds);
FD_SET(fd1,&readfds);
FD_SET(fd2,&readfds);
FD_SET(fd3,&readfds);
select(max(fd1,max(fd2,fd3))+1,&readfds,NULL,NULL,NULL);
}


/* move_frame
 * Move the viewpoint forward by a distance determined by the time the last
 * frame took to draw. Not totally realistic, but the only solution I could
 * think of.
 */
static void move_frame(plview *plv)
{
userprefs *cp = get_prefs(NULL);

plv->dst += last_frame_len()*cp->move_speed;
if (plv->dst > plv->range)
	plv->dst = plv->range;
}


/* check_server_command
 * Check for and handle a command from the server
 */
static void check_server_command(int *fdptr, player **pptr, person **prptr,
				 plview *plv)
{
fd_set readfds;
struct timeval tv = {0,0};
int fd = *fdptr;		/* connection to server */
player *p = *pptr;
person *pr = *prptr;
char comline[256];	/* command line from server */
char **comv;		/* comline broken into tokens */
int comc;		/* tokens in comv */

while(1) {
	/* Poll for activity
	 */
	FD_ZERO(&readfds);
	FD_SET(fd,&readfds);
	select(fd+1,&readfds,NULL,NULL,&tv);
	if (!FD_ISSET(fd,&readfds))
		return;

	fdgets(fd,comline,256);
	parse_string(comline,&comv,&comc);
	if (comc == 0)
		return;		/* empty command line */

	switch(atoi(comv[0])) {
	case com_move: {
		pr->pos = lookup_node(comv[1]);
		plv->x1 = plv->x2; plv->y1 = plv->y2; plv->z1 = plv->z2;
		plv->x2 = pr->pos->x;
		plv->y2 = pr->pos->y;
		plv->z2 = pr->pos->z + PLAYER_HEIGHT;
		plv->dst = 0.0;
		plv->range = sqrt(sqr(plv->x1 - plv->x2) +
				  sqr(plv->y1 - plv->y2) +
				  sqr(plv->z1 - plv->z2));
		fill_here_list(pr);
		break;
		}

	case com_msg: {
		char buf[256] = "";
		int i;

		for(i=1; i<comc; i++) {
			strcat(buf,comv[i]);
			strcat(buf," ");
			}
		display_message("%s",buf);
		break;
		}

	case com_err: {
		display_message("error : %s",error_string(atoi(comv[1])));
		break;
		}
	
	case com_addobject: {
		object *nob;
	
		parse_object(fd,&nob);
		add_object(nob);
		break;
		}
	
	case com_addperson: {
		person *npr;
	
		parse_person(fd,&npr);
		add_person(npr);
		break;
		}
	
	case com_addshape: {
		parse_3d(fd);
		if (strequ(p->name,comv[1]))
			visible_3d(p->name,0);
		else
			need_redraw();
		break;
		}

	case com_moveobject: {
		object *mob = lookup_object(comv[1]);
	
		mob->pos = lookup_node(comv[2]);
		mob->inv = NULL;
		break;
		}

	case com_moveperson: {
		person *mpr = lookup_person(comv[1]);
	
		mpr->pos = lookup_node(comv[2]);
		break;
		}

	case com_moveshape: {
		double dx = atof(comv[2]);
		double dy = atof(comv[3]);
		double dz = atof(comv[4]);
		move_3d(comv[1],dx,dy,dz);
		if (!strequ(comv[1],p->name)) {
			need_redraw();
			}
		break;
		}

	case com_delobject: {
		object *dob = lookup_object(comv[1]);
	
		remove_object(dob);
		free_object(dob);
		break;
		}
	
	case com_delperson: {
		person *dpr = lookup_person(comv[1]);
	
		remove_person(dpr);
		free_person(dpr);
		break;
		}

	case com_delshape: {
		del_3d(comv[1]);
		if (!strequ(comv[1],p->name)) {
			need_redraw();
			}
		break;
		}

	case com_newserv: {
		player *oldme = *pptr;
		userservprefs csp;
	
		waiting_cursor();
		logmessage("Connecting to sim server %s",comv[1]);

		/* Clear out all data in preparation for loading new stuff
	 	 * from new server.
		 */
		clear_nodes();
		clear_paths();
		clear_persons();
		clear_objects();
		unhighlight_3d();
		clear_3d();
	
		/* Re-connect to the new server, with a new fd
		 */
		get_sprefs(&csp);
		csp.port = atoi(comv[2]);
		strcpy(csp.server,comv[1]);
		put_sprefs(&csp);
		save_sprefs();
		*fdptr = reconnect_server((*pptr)->name,pptr,supported_3d(),
					  parse_3ds);
	
		/* Get rid of the old player struct, and find the new one
		 * from the new server */
		free_player(oldme);
		p = *pptr;
		*prptr = lookup_person(p->name);
		pr = *prptr;
	
		/* set up smooth movement to new node */
		plv->x1 = plv->x2; plv->y1 = plv->y2; plv->z1 = plv->z2;
		plv->x2 = pr->pos->x;
		plv->y2 = pr->pos->y;
		plv->z2 = pr->pos->z + PLAYER_HEIGHT;
		plv->x1 = (plv->x1 + plv->x2) / 2;	/* position 1/2 way */
		plv->y1 = (plv->y1 + plv->y2) / 2;	/* to next node. */
		plv->z1 = (plv->z1 + plv->z2) / 2;
		plv->dst = 0.0;
		plv->range = sqrt(sqr(plv->x1 - plv->x2) +
				  sqr(plv->y1 - plv->y2) +
				  sqr(plv->z1 - plv->z2));
	
		/* Delete this player's shape */
		visible_3d(p->name,0);
	
		fill_here_list(pr);
		create_base_ui_map();
		need_redraw();
		normal_cursor();
		break;
		}
	
	case com_getobject: {
		object *ob = lookup_object(comv[1]);
		person *gpr = lookup_person(comv[2]);

		pickup_item(gpr,ob);
		if (pr == gpr)
			fill_inventory_list(pr);
		fill_here_list(pr);
		break;
		}

	case com_dropobject: {
		object *ob = lookup_object(comv[1]);
		person *gpr = lookup_person(comv[2]);
	
		putdown_item(gpr,ob);
		if (pr == gpr)
			fill_inventory_list(pr);
		fill_here_list(pr);
		break;
		}

	case com_rotshape: {
		double ang = atof(comv[2]);
		double xa = atof(comv[3]);
		double ya = atof(comv[4]);
		double za = atof(comv[5]);
		rotate_3d(comv[1],ang,xa,ya,za);
		if (!strequ(comv[1],p->name)) {
			need_redraw();
			}
		break;
		}

	case com_giveobject: {
		object *ob = lookup_object(comv[1]);
		person *gpr = lookup_person(comv[2]);
		person *rpr = lookup_person(comv[3]);
	
		putdown_item(gpr,ob);
		pickup_item(rpr,ob);
		if (pr == gpr || pr == rpr)
			fill_inventory_list(pr);
		if (pr == rpr)
			display_messagebox("Received %s from %s",ob->name,
					   gpr->name);
		break;
		}

	case com_displaymime: {
		/*
		userprefs *cp      = get_prefs(NULL);
		userservprefs *csp = get_sprefs(NULL);
		char dfn[256];
		int recved_data;

		recved_data = cp->recved_data;
		open_received_requestor(comv[1],dfn,&recved_data);
		if (recved_data != 3) {
			display_message("Receiving data...");
			recv_display_data(csp->server,comv[1],atoi(comv[2]),
					  atoi(comv[3]),recved_data,dfn);
			}
		*/
		display_message("Receiving data...");
		download_data(comv[1],atoi(comv[3]),atoi(comv[2]));
		break;
		}

	case com_text: {
		int lines = atoi(comv[1]);
	
		for(; lines>0; lines--) {
			char line[256];
			fdgets(fd,line,256);
			strip_return(line);
			display_message("%s",line);
			}
		break;
		}

	case com_actlist: {
		fill_do_requestor(comv[1],&comv[2],comc-2);
		break;
		}

	case com_colshape: {
		int r = atoi(comv[2]);
		int g = atoi(comv[2]);
		int b = atoi(comv[2]);
		colour_3d(comv[1],r,g,b);
		if (!strequ(comv[1],p->name)) {
			need_redraw();
			}
		break;
		}

	case com_shdshape: {
		int sd = atoi(comv[2]);
		shade_3d(comv[1],sd);
		if (!strequ(comv[1],p->name)) {
			need_redraw();
			}
		break;
		}

	case com_ls: {
		fill_send_requestor(&comv[1],comc-1);
		break;
		}

	case com_chpath: {
		int i;
		node *n1 = lookup_node(comv[1]);
		node *n2 = lookup_node(comv[2]);

		for(i=0; i<n1->linkc; i++)
			if (other_end(n1->link[i],n1) == n2)
				n1->link[1]->closed = atoi(comv[3]);
		break;
		}

	case com_shutdown: {
		close(fd);
		shutdown_3d();
		exit(0);
		break;
		}

	case com_email: {
		person *epr = lookup_person(comv[1]);

		if (epr->email) free(epr->email);
		epr->email = strdupe(comv[2]);
		break;
		}

	case com_fullname: {
		person *fpr = lookup_person(comv[1]);

		if (fpr->fullname) free(fpr->fullname);
		fpr->fullname = strdupe(comv[2]);
		break;
		}

	case com_teleport: {
		pr->pos = lookup_node(comv[1]);
		plv->x1 = plv->x2 = pr->pos->x;
		plv->y1 = plv->y2 = pr->pos->y;
		plv->z1 = plv->z2 = pr->pos->z + PLAYER_HEIGHT;
		plv->dst = plv->range = 0.0;
		need_redraw();
		fill_here_list(pr);
		break;
		}

		} /* end switch */
	} /* end while */
}


/* client_killed
 * Called when client program is killed or ctrl-c'd
 */
static void client_killed(int s)
{
shutdown_3d();
exit(5);
}


/* reaper
 * Clean up after a child process dies
 */
static void reaper(int sig)
{
int status;

wait(&status);
signal(SIGCHLD,reaper);
}

