/* simtext.c
 *
 * main() function for a text-only sim client
 */
#include "sim.h"
#include "client.h"

static void print_location(person *p, node *loc);
static char *get_player_command(void);
static void execute_player_command(char *cmd, int fd, player *p, person *pr);
static void get_server_command(int *fdptr, player **me, person **prptr);
static void prompt();
static void arrive_person(person *me, person *other);
static void arrive_object(person *me, object *ob);
static void depart_person(person *me, person *other);
static void depart_object(person *me, object *ob);
static void move_and_turn(int fd, int linknum, person *pr);
static void reaper(int sig);

int parse_debug		= 0;	/* print each token parsed */
int fd_debug		= 0;	/* print everything sent to/from server */
int is_sim_server       = 0;	/* this is a client */

int main(int argc, char **argv)
{
struct in_addr servaddr;	/* ip address of server */
char *name = NULL;		/* name of this player */
int fd;				/* connection to server */
player *me;			/* info on this player */
person *pr;			/* person representing this player */
int i;

if (!parse_client_args(argc, argv, &name)) {
	exit(1);
	}

/* Find a name for this player
 */
if (!name) {
	if ((name = getenv("USER")) == NULL) {
		if ((name = getenv("LOGNAME")) == NULL) {
			logerror("No name given or found");
			exit(1);
			}
		}
	}

/* Setup signal handlers
 */
signal(SIGCHLD,reaper);

/* Retreive player info and nodes from server
 */
logmessage("Connecting to sim server. Please wait..");
if ((fd = connect_server(name,&me,lvl_none,NULL))==-1){
	logerror("Error connecting to server");
	exit(3);
	}
logmessage("Connected to %s port %d",get_sprefs(NULL)->server,
				     get_sprefs(NULL)->port);
save_sprefs();	/* changed by connect_server() */
pr = lookup_person(me->name);

fdprintf(1,"\n****** Starting game *******\n");
print_location(pr,pr->pos);
prompt();
while(1) {
	fd_set readfds;
	
	/* Wait for either a player or server command
	 */
	FD_ZERO(&readfds);
	FD_SET(0,&readfds);
	FD_SET(fd,&readfds);
	select(fd+1,&readfds,NULL,NULL,NULL);
	if (FD_ISSET(0,&readfds)) {
		char *cmd;

		if ((cmd = get_player_command()) != NULL) {
			execute_player_command(cmd,fd,me,pr);
			prompt();
			}
		}
	if (FD_ISSET(fd,&readfds)) {
		get_server_command(&fd,&me,&pr);
		}
	}
}


/* prompt
 * Print user input prompt
 */
static void prompt()
{
fdprintf(1,"> ");
}


/* print_location
 * Print info on the location this player is at
 */
static void print_location(person *p, node *loc)
{
person *pr;			/* person at this location */
object *ob;			/* object at this location */
int person_header = 0;
int object_header = 0;
int i;

fdputc(1,'\n');
fdprintf(1,"Location    : %s\n",loc->name);
if (loc->desc)
	fdprintf(1,"Description : %s\n",loc->desc);
if (p->pos == loc) {
	static char *headn[8] = {"n","ne","e","se","s","sw","w","nw"};

	fdprintf(1,"Exits :\n");
	for(i=0; i<loc->linkc; i++) {
		path *pt = loc->link[i];
		node *ex = other_end(pt,loc);
		double ang = atan2(ex->y - loc->y, ex->x - loc->x);
		int hn = fixang(rtod(ang)+22.5+90) / 45;

		fdprintf(1,"\t%d - %s (%s) %s\n",i,ex->name,headn[hn],
			 pt->closed ? "(closed)" : "");
		}
	}

/* List other persons here
 */
reset_person();
while((pr = next_person()))
	if (pr->pos == loc && pr != p) {
		if (!person_header) {
			fdprintf(1,"People here:\n");
			person_header = 1;
			}
		fdprintf(1," %s\n",pr->name);
		}

/* List objects here
 */
reset_object();
while((ob = next_object())) 
	if (ob->pos == loc) {
		if (!object_header) {
			fdprintf(1,"Objects here:\n");
			object_header = 1;
			}
		fdprintf(1," %s\n",ob->name);
		}
}


/* get_player_command
 * Read one character from standard input, and add it to the command buffer.
 */
static char *get_player_command(void)
{
static char combuf[256];
static int compos = 0;

combuf[compos] = fdgetc(0);
if (combuf[compos] == '\n') {
	combuf[compos] = '\0';
	compos = 0;
	return combuf;
	}
else {
	compos++;
	return NULL;
	}
}


/* execute_player_command
 * Parse a command from the player and execute it. Most commands simply
 * send a request to the server, which sends back a message to actually
 * carry it out.
 * Commands:
 *   go <node name>		- move to a new node
 *   go <number>		- move along the given path number from current
 *   say <word> <word> ...	- message to players at same node
 *   yell <word> <word> ...	- message to players on same server
 *   get <thing>		- pick something up
 *   look <thing>		- more info on thing
 *   do <action> <thing>	- invoke action of thing
 *   help			- list all commands
 *   quit			- diconnect from server
 *   drop <object>		- drop an object in inventory
 *   inv			- list items carried
 *   face <person>		- turns player's 3d shape
 *   give <object> <person>	- give an object to nearby person
 *   upload <file>		- send named file to server
 *   send <file> <player>	- send to a player a file already on server
 *   what <object>		- request list of methods for an object
 *   set <pref> <value>		- set a preferences value
 */
static void execute_player_command(char *cmd, int fd, player *p, person *pr)
{
char **cmdv;			/* array of words in command */
int    cmdc;			/* number of words in cmdv */
int opcode;

/* Parse and check command
 */
parse_string(cmd,&cmdv,&cmdc);
if (!cmdc) {
	/* empty line
	 */
	return;
	}
opcode = identify_command(cmdv[0]);
if (opcode != cmd_unknown && validate_command(opcode,cmdv,cmdc) == 0) {
	logerror("%s : Incorrect arguments",cmdv[0]);
	return;
	}

/* Carry out command
 */
switch(opcode) {
case cmd_go: {
	int linknum;		/* link number chosen */
	int status;

	if (isdigit(cmdv[1][0])) {
		/* numbered link given
		 */
		linknum = atoi(cmdv[1]);
		}
	else {
		/* Named node given
		 */
		node *cp = pr->pos;	/* node currently at */
		int j;

		for(j=0; j<cp->linkc; j++)
			if (strequ(cmdv[1],other_end(cp->link[j],cp)->name))
				break;
		if (j != cp->linkc)
			linknum = j;
		else {
			logerror("There is no position %s nearby",cmdv[1]);
			return;
			}
		}
	move_and_turn(fd,linknum,pr);
	break;
	}

case cmd_say: {
	client_request(fd,req_say,&cmdv[1],cmdc-1);
	break;
	}

case cmd_yell: {
	client_request(fd,req_yell,&cmdv[1],cmdc-1);
	break;
	}

case cmd_get: {
	client_request(fd,req_get,cmdv[1]);
	break;
	}

case cmd_look: {
	if (cmdc == 1)
		print_location(pr,pr->pos);
	else {
		node *loc;

		if (isinteger(cmdv[1])) {
			/* Numbered path given
			 */
			int ln = atoi(cmdv[1]);

			if (ln >= 0 && ln < pr->pos->linkc)
				loc = other_end(pr->pos->link[ln],pr->pos);
			else
				loc = NULL;
			}
		else {
			/* Named node given
			 */
			loc = lookup_node(cmdv[1]);
			}
		if (!loc || !close_node(loc,pr->pos))
			fdprintf(1,"There is no location %s nearby\n",cmdv[1]);
		else {
			double xd = loc->x - pr->pos->x;
			double yd = loc->y - pr->pos->y;

			print_location(pr,loc);
			client_request(fd,req_turn,rtod(atan2(yd,xd)));
			}
		}
	break;
	}

case cmd_do: {
	client_request(fd,req_do,cmdv[1],cmdv[2],&cmdv[3],cmdc-3);
	break;
	}

case cmd_exam: {
	object *ob = lookup_object(cmdv[1]);

	if (!ob || (ob->pos != pr->pos && ob->inv != pr))
		fdprintf(1,"There is no %s here\n",cmdv[1]);
	else if (ob->desc == NULL)
		fdprintf(1,"No description given\n");
	else
		fdprintf(1,"%s\n",ob->desc);
	break;
	}

case cmd_help: {
	if (cmdc == 1) {
		fdputs(1,"Available commands :");
		list_commands();
		fdputs(1,"Type  help <command>  for more details on a command");
		}
	else {
		command_help(cmdv[1]);
		}
	break;
	}

case cmd_quit: {
	close(fd);
	exit(0);
	break;
	}

case cmd_drop: {
	client_request(fd,req_drop,cmdv[1]);
	break;
	}

case cmd_inv: {
	if (pr->invc == 0)
		fdprintf(1,"Inventory empty\n");
	else {
		int i;

		fdprintf(1,"Inventory:\n");
		for(i=0; i<pr->invc; i++)
			fdprintf(1," %s\n",pr->inv[i]);
		}
	break;
	}

case cmd_face: {
	person *fpr = lookup_person(cmdv[1]);

	if (!fpr)
		fdprintf(1,"Person %s not found\n",cmdv[1]);
	else if (fpr == pr)
		fdprintf(1,"Cannot face self!\n");
	else if (fpr->pos == pr->pos)
		fdprintf(1,"Cannot face a person at the same location\n");
	else {
		double xd = fpr->pos->x - pr->pos->x;
		double yd = fpr->pos->y - pr->pos->y;

		client_request(fd,req_turn,rtod(atan2(yd,xd)));
		}
	break;
	}

case cmd_give: {
	client_request(fd,req_give,cmdv[1],cmdv[2]);
	break;
	}

case cmd_upload: {
	if (upload_data(fd,cmdv[1]) == 0)
		fdprintf(1,"Cannot find file %s\n",cmdv[1]);
	break;
	}

case cmd_send: {
	client_request(fd,req_send,cmdv[1],cmdv[2]);
	break;
	}

case cmd_ls: {
	client_request(fd,req_ls);
	break;
	}

case cmd_rm: {
	client_request(fd,req_rm,cmdv[1]);
	break;
	}

case cmd_what: {
	client_request(fd,req_actlist,cmdv[1]);
	break;
	}

case cmd_set: {
	userprefs cp;

	get_prefs(&cp);
	if (strequ(cmdv[1],"default_3d")) {
		int lvl = atoi(cmdv[2]);

		if (lvl < 0 || lvl > 2)
			logerror("3d level must be 0,1 or 2");
		else
			cp.default_3d = lvl;
		}
	else if (strequ(cmdv[1],"recved_data")) {
		int rcv = atoi(cmdv[2]);

		if (rcv < 0 || rcv > 3)
			logerror("Received data mode must be 0,1,2 or 3");
		else
			cp.recved_data = rcv;
		}
	else if (strequ(cmdv[1],"move_speed")) {
		double msp = atof(cmdv[2]);

		if (msp < 0.0 || msp > 5.0)
			logerror("Move speed must be between 0 and 5 m/s");
		else
			cp.move_speed = msp;
		}
	else if (strequ(cmdv[1],"password")) {
		strcpy(cp.password,cmdv[2]);
		client_request(fd,req_setpass,cp.password);
		}
	else if (strequ(cmdv[1],"display_transfer")) {
		int mode = atoi(cmdv[2]);

		if (mode != 0 && mode != 1)
			logerror("Display transfer must be 0 or 1");
		else
			cp.display_transfer = mode;
		}
	else if (strequ(cmdv[1],"transfer_block")) {
		int size = atoi(cmdv[2]);

		if (size <= 0)
			logerror("Transfer block size must be >= 0");
		else
			cp.transfer_block = size;
		}
	else
		logerror("Unknown preference %s",cmdv[1]);
	put_prefs(&cp);
	if (save_prefs() == 0)
		logerror("Couldn't save preferences");
	break;
	}

case cmd_prefs: {
	userprefs *cp = get_prefs(NULL);

	fdprintf(1,"Current preferences\n");
	fdprintf(1,"\tdefault_3d       = %d\n",cp->default_3d);
	fdprintf(1,"\trecved_data      = %d\n",cp->recved_data);
	fdprintf(1,"\tmove_speed       = %f\n",cp->move_speed);
	fdprintf(1,"\tpassword         = %s\n",cp->password);
	fdprintf(1,"\tdisplay_transfer = %d\n",cp->display_transfer);
	fdprintf(1,"\ttransfer_block   = %d\n",cp->transfer_block);
	break;
	}

case cmd_suicide: {
	char resp[20];
	fdprintf(1,"Do you REALLY want to do this? : ");
	fdgetw(0,resp,20);
	if (toupper(resp[0]) == 'Y') {
		client_request(fd,req_suicide);
		}
	break;
	}

case cmd_avatar: {
	int afd;
	if ((afd = open(cmdv[1],O_RDONLY)) == -1)
		logerror("File %s not found",cmdv[1]);
	else {
		int cppafd = pipethrough(afd,NULL,CPPCMD);

		client_request(fd,req_avatar);
		copy_to_eof(cppafd,fd);
		fdprintf(fd,"\nend\n");
		pipeclose(cppafd);
		close(afd);
		}
	break;
	}

case cmd_email: {
	client_request(fd,req_email,cmdv[1]);
	break;
	}

case cmd_fullname: {
	client_request(fd,req_fullname,cmdv[1]);
	break;
	}

case cmd_about: {
	person *apr;

	if ((apr = lookup_person(cmdv[1])) == NULL ||
	    !close_node(apr->pos,pr->pos))
		logerror("There is not person %s nearby",cmdv[1]);
	else {
		fdprintf(1,"Person details\n");
		fdprintf(1,"Name      : %s\n",apr->name);
		fdprintf(1,"Pos       : %s\n",apr->pos->name);
		fdprintf(1,"Email     : %s\n",apr->email ? apr->email
							 : "unknown");
		fdprintf(1,"Full name : %s\n",apr->fullname ? apr->fullname
							    : "unknown");
		}
	break;
	}

case cmd_unknown: {
	if (isinteger(cmdv[0])) {
		/* Assume this is a 'go' command.
		 */
		int linknum = atoi(cmdv[0]);

		move_and_turn(fd,linknum,pr);
		}
	else if (cmdc >= 2) {
		/* Assume this is a 'do' command.
		 */
		client_request(fd,req_do,cmdv[0],cmdv[1],&cmdv[2],cmdc-2);
		}
	else {
		/* This may be a directional command (ie n, w, ne, ...)
		 */
		static char *headn[8] = {"n","ne","e","se","s","sw","w","nw"};
		int i;

		for(i=0; i<8; i++)
			if (strequ(headn[i],cmdv[0])) break;
		if (i != 8) {
			/* Valid direction name given. 
			 */
			double dirang = fixang(i*45.0 - 90.0);
			double *ang = path_angles(pr->pos);
			double mindiff = 360;
			int    minpos = -1;
			int i;

			for(i=0; i<pr->pos->linkc; i++) {
				double diff;

				diff = abs(dirang - ang[i]);
				if (diff > 180.0) diff = 360.0 - diff;
				if (diff < mindiff) {
					mindiff = diff;
					minpos = i;
					}
				}
			if (mindiff < 22.5)
				move_and_turn(fd,minpos,pr);
			else
				logerror("No path in direction \"%s\"",cmdv[0]);
			}
		else {
			logerror("Unknown command : %s\n",cmdv[0]);
			}
		}
	break;
	}

	} /* end switch */
}


/* get_server_command
 * Read and execute a command from the connected server.
 */
static void get_server_command(int *fdptr, player **me, person **prptr)
{
int fd = *fdptr;	/* connection to server */
person *pr = *prptr;	/* person representing this player */
char comline[256];	/* command line read from server */
char **comv;		/* comline broken into tokens */
int    comc;		/* tokens in conv */

fdgets(fd,comline,256);
parse_string(comline,&comv,&comc);
if (comc == 0)
	return;
switch(atoi(comv[0])) {
case com_move: {
	pr->pos = lookup_node(comv[1]);
	print_location(pr,pr->pos);
	prompt();
	break;
	}

case com_msg: {
	int i;

	for(i=1; i<comc; i++)
		fdprintf(1,"%s ",comv[i]);
	fdprintf(1,"\n");
	prompt();
	break;
	}

case com_err: {
	fdprintf(1,"error : %s\n",error_string(atoi(comv[1])));
	prompt();
	break;
	}

case com_addobject: {
	object *nob;

	parse_object(fd,&nob);
	add_object(nob);
	arrive_object(pr,nob);
	break;
	}

case com_addperson: {
	person *npr;

	parse_person(fd,&npr);
	add_person(npr);
	arrive_person(pr,npr);
	break;
	}

case com_moveobject: {
	object *mob = lookup_object(comv[1]);

	depart_object(pr,mob);
	mob->pos = lookup_node(comv[2]);
	arrive_object(pr,mob);
	break;
	}

case com_moveperson: {
	person *mpr = lookup_person(comv[1]);

	depart_person(pr,mpr);
	mpr->pos = lookup_node(comv[2]);
	arrive_person(pr,mpr);
	break;
	}

case com_delobject: {
	object *dob = lookup_object(comv[1]);

	depart_object(pr,dob);
	remove_object(dob);
	free_object(dob);
	break;
	}

case com_delperson: {
	person *dpr = lookup_person(comv[1]);

	depart_person(pr,dpr);
	remove_person(dpr);
	free_person(dpr);
	break;
	}

case com_newserv: {
	player *oldme = *me;
	userservprefs csp;

	clear_nodes();
	clear_paths();
	clear_persons();
	clear_objects();
	get_sprefs(&csp);
	csp.port = atoi(comv[2]);
	strcpy(csp.server,comv[1]);
	put_sprefs(&csp);
	save_sprefs();
	*fdptr = reconnect_server((*me)->name,me,lvl_none,NULL);
	free_player(oldme);
	*prptr = lookup_person((*me)->name);
	print_location(*prptr,(*prptr)->pos);
	prompt();
	break;
	}

case com_getobject: {
	object *ob = lookup_object(comv[1]);
	person *gpr = lookup_person(comv[2]);

	pickup_item(gpr,ob);
	if (pr == gpr) {
		fdprintf(1,"Picked up %s\n",ob->name);
		prompt();
		}
	break;
	}

case com_dropobject: {
	object *ob = lookup_object(comv[1]);
	person *gpr = lookup_person(comv[2]);

	if (pr == gpr) {
		fdprintf(1,"Put down %s\n",ob->name);
		prompt();
		}
	putdown_item(gpr,ob);
	arrive_object(pr,ob);
	break;
	}

case com_giveobject: {
	object *ob = lookup_object(comv[1]);
	person *gpr = lookup_person(comv[2]);
	person *rpr = lookup_person(comv[3]);

	if (pr == gpr) {
		fdprintf(1,"Given %s to %s\n",ob->name,rpr->name);
		prompt();
		}
	if (pr == rpr) {
		fdprintf(1,"Received %s from %s\n",ob->name,gpr->name);
		prompt();
		}
	putdown_item(gpr,ob);
	pickup_item(rpr,ob);
	break;
	}

case com_displaymime: {
	userprefs *cp      = get_prefs(NULL);
	userservprefs *csp = get_sprefs(NULL);
	int recved_data;
	char dfn[256];			/* filename to save data as */

	recved_data = cp->recved_data;
	if (recved_data == 0 || recved_data == 1)
		fdprintf(1,"Data of type %s has arrived.\n",comv[1]);
	if (recved_data == 1) {
		char resp[20];

		fdprintf(1,"Do you wish to [v]iew or [s]ave it? : ");
		fdgets(0,resp,20);
		if (toupper(resp[0]) == 'V') {
			recved_data = 2;
			prompt();
			}
		else
			recved_data = 0;
		}
	if (recved_data == 0) {

		fdprintf(1,"Enter a filename to save this data as : ");
		fdgets(0,dfn,256);
		strip_return(dfn);
		prompt();
		}
	recv_display_data(csp->server,comv[1],atoi(comv[2]),atoi(comv[3]),
			  recved_data,dfn);
	break;
	}

case com_text: {
	int lines = atoi(comv[1]);
	int i;

	fdputc(1,'\n');
	for(i=0; i<lines; i++) {
		char line[256];
		fdgets(fd,line,256);
		fdprintf(1,"%s",line);
		}
	prompt();
	break;
	}

case com_actlist: {
	int i;

	fdprintf(1,"\n%d action%s for %s\n",comc-2,comc==3 ? "" : "s",comv[1]);
	for(i=2; i<comc; i++)
		fdprintf(1,"%s\n",comv[i]);
	prompt();
	break;
	}

case com_ls: {
	int i;

	fdprintf(1,"\n%d data file%s :\n",comc-1,comc-1 == 1 ? "" : "s");
	for(i=1; i<comc; i++)
		fdprintf(1,"%s\t",comv[i]);
	fdprintf(1,"\n");
	prompt();
	break;
	}

case com_chpath: {
	int i;
	node *n1 = lookup_node(comv[1]);
	node *n2 = lookup_node(comv[2]);

	for(i=0; i<n1->linkc; i++)
		if (other_end(n1->link[i],n1) == n2) {
			n1->link[i]->closed = atoi(comv[3]);
			break;
			}
	if (n1 == pr->pos || n2 == pr->pos) {
		node *other;

		if (n1 == pr->pos) other = n2;
		else               other = n1;
		fdprintf(1,"The path to %s is now %s\n",other->name,
			 n1->link[i]->closed ? "closed" : "open");
		prompt();
		}
	break;
	}

case com_shutdown: {
	close(fd);
	fdprintf(1,"\n");
	exit(0);
	break;
	}

case com_email: {
	person *epr = lookup_person(comv[1]);

	if (epr->email) free(epr->email);
	epr->email = strdupe(comv[2]);
	break;
	}

case com_fullname: {
	person *fpr = lookup_person(comv[1]);

	if (fpr->fullname) free(fpr->fullname);
	fpr->fullname = strdupe(comv[2]);
	break;
	}

case com_teleport: {
	pr->pos = lookup_node(comv[1]);
	fdprintf(1,"You have been moved to a new location!\n");
	print_location(pr,pr->pos);
	prompt();
	break;
	}

	} /* end switch */
}


/* arrive_person
 * Called when a person changes location or enters.
 */
static void arrive_person(person *me, person *other)
{
if (me->pos == other->pos && me != other) {
	fdprintf(1,"%s has arrived\n",other->name);
	prompt();
	}
else if (close_node(me->pos,other->pos) && me != other) {
	fdprintf(1,"%s is nearby\n",other->name);
	prompt();
	}
}


/* arrive_object
 * Called when an object is moved or dropped
 */
static void arrive_object(person *me, object *ob)
{
if (me->pos == ob->pos) {
	fdprintf(1,"Object %s is now here\n",ob->name);
	prompt();
	}
}


/* depart_person
 * Called before a person moves or leaves the simulation
 */
static void depart_person(person *me, person *other)
{
if (me->pos == other->pos && me != other) {
	fdprintf(1,"%s has left\n",other->name);
	prompt();
	}
}


/* depart_object
 * Called before an object is moved or taken
 */
static void depart_object(person *me, object *ob)
{
if (me->pos == ob->pos) {
	fdprintf(1,"Object %s is no longer here\n",ob->name);
	prompt();
	}
}


/* move_and_turn
 * Move along a numbered path, and turn to face that direction
 */
static void move_and_turn(int fd, int linknum, person *pr)
{
node *nn;
double xd,yd;

if (linknum < 0 || linknum >= pr->pos->linkc) {
	logerror("There is no path %d from this position",linknum);
	return;
	}
client_request(fd,req_move,linknum);
nn = other_end(pr->pos->link[linknum],pr->pos);
xd = nn->x - pr->pos->x;
yd = nn->y - pr->pos->y;
client_request(fd,req_turn,rtod(atan2(yd,xd)));
}


/* reaper
 * Clean up after a child process dies
 */
static void reaper(int sig)
{
int status;

wait(&status);
signal(SIGCHLD,reaper);
}

