/* simui.c
 *
 * Functions for creating and using the JX user interface for a graphical
 * sim client.
 */
#include "sim.h"
#include "client.h"
#include "3d.h"
#include "graphics.h"
#include "jx/jx.h"
#include <X11/cursorfont.h>
#include "help_text.h"

static void handle_ui_event(XEvent *event, int fd, player *pl, person *pr,
			    plview *plv);
static void get_pointer_pos(Display *d, int *px, int *py);
static void open_do_requestor(void);
static void close_command_requestor(void);
static void resize_gadgets(void);
static void handle_key_press(XEvent *event, int fd, player *pl, person *pr,
			     plview *plv);
static void fill_command_requestor(int fd, person *pr);
static void choose_download_mode(Display *tdisplay, char *mtype, int *rd);
static void choose_download_filename(Display *tdisplay, int *rd, char *fn);
static int copy_with_status(char *message, Display *tdisplay, int fd1,
			    int fd2, int len);

static Display *display;	/* Connection to X server */
static Window   root_win;	/* Root window of this display */
static int	 screen_num;	/* default screen of display */
static Screen	*screen_ptr;

static Window win;		/* top level window, containing gadgets */

static long black,white;	/* pixel colours to use with gadgets */
static long red,grey;

static JXgadget *vert,*horiz;	/* scrollbars for view */
static JXgadget *left,*right;	/* movement gadgets */
static JXgadget *forward;
static JXgadget *compass;	/* rough current facing */
static JXgadget *look;		/* location info */
static JXgadget *mesg;		/* text message entry box */
static JXgadget *saymsg, *yellmsg;
static JXgadget *mlist;		/* messages list gadget */
static JXgadget *ilist;		/* list of items in inventory / here */
static JXgadget *iname;		/* item of interest */
static JXgadget *hereinv;	/* toggle item list */
static JXgadget *dropget;	/* drop or get an item */
static JXgadget *examine;	/* examine an item */
static JXgadget *give;		/* give item to someone */
static JXgadget *doo;		/* do something to an item */
static JXgadget *command;	/* open commands menu */
static JXgadget *wizard;	/* open wizard command menu */
static JXgadget *prefs;		/* open preferences window */
static JXgadget *help;		/* open help window */
static JXgadget *mapp;		/* open map window */
static JXgadget *quit;		/* exit simulation */

static Cursor arrow_cur;	/* normal cursor for interface and 3d window */
static Cursor clock_cur;	/* waiting cursor */

static JXgadget *givewin = NULL;/* give requestor */
static JXgadget *giveok;	/* ok gadget in give requestor */
static JXgadget *givecan;	/* cancel gadget in give requestor */
static JXgadget *giveper;	/* person name in give requestor */
static JXgadget *giveobj;	/* item name in give requestor */

static JXgadget *dowin = NULL;	/* do requestor */
static JXgadget *dook;		/* ok button in do requestor */
static JXgadget *docan;		/* cancel button in do requestor */
static JXgadget *doaction;	/* action combo box in do requestor */
static JXgadget *doobject;	/* item text box in do requestor */
static JXgadget *doparams;	/* parameters text box in do requestor */

static JXgadget *helpwin = NULL;/* help requestor */
static JXgadget *helpok;	/* ok button in help window */

static JXgadget *cmdwin = NULL;	/* commands requestor */
static JXgadget *cmdcombo;	/* combo box for choosing command mode */
static JXgadget *cmddata;	/* data file to send/remove combo box */
static JXgadget *cmdperson;	/* person to send to combo box */
static JXgadget *cmdsend;	/* send button */
static JXgadget *cmdfile;	/* file to upload / for avatar text box */
static JXgadget *cmddir;	/* file browser toggle button */
static JXgadget *cmdupload;	/* upload button */
static JXgadget *cmdfilelist=NULL;	/* file browser for uploading */
static JXgadget *cmdsuicide;	/* destroy player button */
static JXgadget *cmdavset;	/* set avatar button */
static JXgadget *cmdrm;		/* remove data button */
static JXgadget *cmdpdset;	/* set player details button */
static JXgadget *cmdemail;	/* email address text box */
static JXgadget *cmdfullname;	/* full name text box */
static JXgadget *cmdclose;	/* close button in command req. */
static int cmdmode = 0;		/* 0=send, 1=upload, 2=rm, 3=avatar,
				   4=player detail, 5=suicide */

static JXgadget *prefswin;	/* preferences requestor */
static JXgadget *prefsd3d;	/* default 3d level combo box */
static JXgadget *prefsrecv;	/* received data combo box */
static JXgadget *prefsms;	/* move speed slider */
static JXgadget *prefspass;	/* password text box */
static JXgadget *prefssave;	/* save button */
static JXgadget *prefsuse;	/* use button */
static JXgadget *prefscan;	/* cancel button */
static JXgadget *prefsdtrans;	/* display transfer toggle */
static JXgadget *prefstblock;	/* transfer block size */

static JXgadget *mapwin;	/* map requestor */
static JXgadget *mappane;	/* scrolling window with map */
static JXgadget *mapclose;	/* close button in map window */

static int invent = 1;		/* 0 = here, !0 =  inventory */

/* init_ui
 * Opens connection to the X server, creates the top level and 3d view
 * windows, creates gadgets and initializes s3dt in the 3d view window
 */
int init_ui(void)
{
XWindowAttributes attribs;
Font cfont;			/* cursor font */
XColor curfore, curback;	/* foreground and background cursor cols */
XSizeHints hints;
userprefs *cp=get_prefs(NULL);	/* current prefs, for size and pos */
int x = cp->x, y = cp->y, w = cp->w, h = cp->h, lw = (w-100)/2;

/* Initialise X
 */
if ((display = XOpenDisplay(NULL)) == NULL) {
	logerror("Couldn't open display %s",XDisplayName(NULL));
	return 0;
	}
screen_num = DefaultScreen(display);
screen_ptr = DefaultScreenOfDisplay(display);
root_win   = RootWindow(display,screen_num);

/* Create top level windows
 */
win = XCreateSimpleWindow(display,root_win,x,y,w,h,0,0,0);
hints.flags = PMinSize;
hints.min_width  = 400;
hints.min_height = 400;
XSetStandardProperties(display,win,"sim3d","sim3d",None,NULL,0,&hints);
XSelectInput(display,win,StructureNotifyMask | KeyPressMask);
XMapWindow(display,win);
XFlush(display);

/* Create and init 3d view window
 */
init_3d(NULL,win,0,0,w-20,h-280);

/* Create and init map pixmap
 */
init_map(display,500,500);

/* Get gadget colours
 */
black = best_colour(0,0,0);
white = best_colour(255,255,255);
grey  = best_colour(100,100,100);
red   = best_colour(255,0,0);

/* Create gadgets main window gadgets 
 */
JXStartup(display,screen_num);
horiz	  = JXNewGadget(display,win,white,black,grey,scroller,0,h-280,w,20,0.0);
JXSetOption(display,horiz,WrapSlider,1,ScrollStep,0.05,SmoothMove,1,0);
vert	  = JXNewGadget(display,win,white,black,grey,scroller,w-20,0,20,h-280,
			0.5);
JXSetOption(display,vert,ScrollStep,0.05,SmoothMove,1,0);
left	  = JXNewGadget(display,win,white,black,grey,boolean,0,h-260,w/5,20,
			"Left");
forward	  = JXNewGadget(display,win,white,black,grey,boolean,w/5,h-260,w/5,20,
			"Forward");
compass   = JXNewGadget(display,win,white,black,grey,string,(w/5)*2,h-260,w/5,
			20,"",100);
JXSetOption(display,compass,StringAlign,a_centre,StringEditable,False,0);
look	  = JXNewGadget(display,win,white,black,grey,boolean,(w/5)*3,h-260,w/5,
			20,"Look");
right	  = JXNewGadget(display,win,white,black,grey,boolean,(w/5)*4,h-260,w/5,
			20,"Right");

mlist     = JXNewGadget(display,win,white,black,grey,slist,0,h-240,lw,220,
			"Starting sim...",NULL);

ilist     = JXNewGadget(display,win,white,black,grey,slist,lw,h-240,lw,200,
			NULL);
JXSetOption(display,ilist,WrapList,False,0);
iname     = JXNewGadget(display,win,white,black,grey,string,lw,h-40,lw,20,
			"",256);
JXSetOption(display,iname,StringAlign,a_right,0);
hereinv   = JXNewGadget(display,win,white,black,grey,toggle,w-100,h-240,100,20,
			0,"Inventory");
dropget   = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-220,100,20,
			"Drop");
examine   = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-200,100,20,
			"Examine");
give      = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-180,100,20,
			"Give..");
doo       = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-160,100,20,
			"Do...");
command   = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-140,100,20,
			"Command..");
wizard    = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-120,100,20,
			"Wizard..");
prefs     = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-100,100,20,
			"Prefs..");
help      = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-80,100,20,
			"Help..");
mapp	  = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-60,100,20,
			"Map..");
quit	  = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-40,100,20,
			"Quit");

mesg	  = JXNewGadget(display,win,white,black,grey,string,0,h-20,w-100,20,
			"",256);
saymsg    = JXNewGadget(display,win,white,black,grey,boolean,w-100,h-20,50,20,
			"Say");
yellmsg   = JXNewGadget(display,win,white,black,grey,boolean,w-50,h-20,50,20,
			"Yell");

JXSetOptionAll(display,HighlightCol,red,0);

/* Create a green cursor for interface and 3d view window, to replace the
 * bogus black cursor on some systems
 */
cfont = XLoadFont(display,"cursor");
curfore.red = 0; curfore.green = 65535; curfore.blue = 0;
curback.red = 0; curback.green = 0;     curback.blue = 0;
arrow_cur = XCreateGlyphCursor(display,cfont,cfont,XC_top_left_arrow,
			       XC_top_left_arrow+1,&curfore,&curback);
clock_cur = XCreateGlyphCursor(display,cfont,cfont,XC_watch,
			       XC_watch+1,&curfore,&curback);
XUnloadFont(display,cfont);
normal_cursor();

return 1;
}


/* create_base_ui_map
 * Stub to call create_base_map()
 */
void create_base_ui_map(void)
{
create_base_map(display);
}


/* position_ui_view
 * Adjust position of sliders for current facing
 */
void position_ui_view(plview *plv)
{
JXSetOption(display,horiz,SliderNPos,plv->ang/360.0,0);
}


/* interface_fd
 * Returns the display's file descriptor
 */
int interface_fd(void)
{
return ConnectionNumber(display);
}


/* check_ui
 * Check for and handle any input from the user interface.
 */
void check_ui(int fd, player *pl, person *pr, plview *plv)
{
while(XPending(display)) {
	XEvent event;

	XNextEvent(display,&event);
	if (!JXGadgetEvent(display,&event)) {
		if (event.type == ClientMessage) {
			/* Message from user interface gadgets
			 */
			handle_ui_event(&event,fd,pl,pr,plv);
			}
		else if (event.type == ConfigureNotify &&
			 event.xconfigure.window == win) {
			/* Window re-sized or moved
			 */
			userprefs cp;

			get_prefs(&cp);
			if (event.xconfigure.width == cp.w &&
			    event.xconfigure.height == cp.h)
				continue;	/* only moved */
			logmessage("re-sizing");
			cp.w = event.xconfigure.width;
			cp.h = event.xconfigure.height;
			put_prefs(&cp);
			resize_3d(0,0,cp.w-20,cp.h-280);
			resize_gadgets();
			}
		else if (event.type == KeyPress) {
			/* Movement key pressed
			 */
			handle_key_press(&event,fd,pl,pr,plv);
			}
		}
	} /* end while */
}


/* check_3d_iu
 * Lets the 3d window handle any events for it.
 * Checks if the mouse was clicked in the 3d window
 */
void check_3d_ui(int fd, player *pl, person *pr, plview *plv)
{
char *sh;	/* name of shape clicked */
int cb;		/* number of button clicked */

check_3d();	/* ask the 3d window to check for events */

if (get_3d_click(&sh,&cb)) {
	/* Mouse clicked in the 3d view window at cx,cy
	 */
	object *ob;	/* object clicked */
	person *cpr;	/* person clicked */

	unhighlight_3d();
	JXSetOption(display,iname,CurrentString,"",0);
	if (sh && (ob = lookup_object(sh)) && ob->pos == pr->pos) {
		/* Clicked on an object
		 */
		JXSetOption(display,iname,CurrentString,ob->name,0);
		switch(cb) {
		case 1:
			/* Just select object
			 */
			highlight_3d(sh);
			break;
		case 2:
			/* Get object
			 */
			client_request(fd,req_get,ob->name);
			break;
		case 3:
			/* Do something to object
			 */
			if (!dowin) {
				highlight_3d(sh);
				open_do_requestor();
				client_request(fd,req_actlist,ob->name);
				}
			break;
		} /* end switch */
		}
	else if (sh && (cpr = lookup_person(sh))) {
		/* Clicked on a person
		 */
		display_messagebox("Person details\n \n"
				   "Name      : %s\n"
				   "Pos       : %s\n"
				   "Email     : %s\n"
				   "Full name : %s\n",
		 		   cpr->name, cpr->pos->name,
				   cpr->email ? cpr->email : "unknown",
				   cpr->fullname ? cpr->fullname : "unknown");
		}
	need_redraw();
	}
}


/* handle_ui_event
 */
static void handle_ui_event(XEvent *event, int fd, player *pl, person *pr,
			    plview *plv)
{
/* View control gadgets
 */
if (JXGMatch(horiz,event)) {
	/* Horizontal slider moved
	 */
	plv->ang = NPosFromScroller(horiz)*360;
	if (DataFromEvent(event) == NewSliderPos) {
		client_request(fd,req_turn,plv->ang);
		}
	need_redraw();
	}
else if (JXGMatch(vert,event)) {
	/* Vertical slider moved
	 */
	plv->tilt = (0.5-NPosFromScroller(vert))*179;
	need_redraw();
	}
else if (JXGMatch(left,event)) {
	/* Turn to next path on left, or 45 deg whichever is less
	 */
	rotate_left(pr,plv);
	JXSetOption(display,horiz,SliderNPos,plv->ang/360.0,0);
	JXRedrawGadget(display,horiz);
	client_request(fd,req_turn,plv->ang);
	need_redraw();
	}
else if (JXGMatch(right,event)) {
	/* Turn to next path on right
	 */
	rotate_right(pr,plv);
	JXSetOption(display,horiz,SliderNPos,plv->ang/360.0,0);
	JXRedrawGadget(display,horiz);
	client_request(fd,req_turn,plv->ang);
	need_redraw();
	}
else if (JXGMatch(forward,event)) {
	move_along(fd,pr,plv,1);
	}
else if (JXGMatch(look,event)) {
	/* Display description of location
	 */
	if (pr->pos->desc)
		display_message("%s : %s",pr->pos->name,pr->pos->desc);
	else
		display_message("%s : No description");
	}
/* Message gadgets
 */
else if (JXGMatch(mesg,event) || JXGMatch(saymsg,event) ||
	 JXGMatch(yellmsg,event)) {
	/* Message to be spoken entered and return hit
	 */
	char **words;
	int wordc;

	parse_string(StringFromString(mesg),&words,&wordc);
	if (JXGMatch(yellmsg,event))
		client_request(fd,req_yell,words,wordc);
	else
		client_request(fd,req_say,words,wordc);
	JXSetOption(display,mesg,CurrentString,"",0);
	}
/* Item list gadgets
 */
else if (JXGMatch(ilist,event)) {
	/* A name chosen from the item list
	 */
	char *str;	/* string chosen */

	str = JXListEntry(ilist,DataFromEvent(event));
	if (str) {
		object *ob = lookup_object(str);

		unhighlight_3d();
		JXSetOption(display,iname,CurrentString,str,0);
		if (ob && !ob->inv)
			highlight_3d(ob->name);
		need_redraw();
		}
		
	}
else if (JXGMatch(iname,event)) {
	/* Object name entered in item name box
	 */
	char *str = StringFromString(iname);
	object *ob = lookup_object(str);

	unhighlight_3d();
	if (ob && !ob->inv)
		highlight_3d(ob->name);
	}
/* Gadgets along right side
 */
else if (JXGMatch(hereinv,event)) {
	/* Toggle display of items at this location or in inventory
	 */
	invent = !invent;
	if (invent) {
		JXSetOption(display,hereinv,ButtonLabel,"Inventory",0);
		JXSetOption(display,dropget,ButtonLabel,"Drop",0);
		}
	else {
		JXSetOption(display,hereinv,ButtonLabel,"Here",0);
		JXSetOption(display,dropget,ButtonLabel,"Get",0);
		}
	fill_hereinv_list(pr);
	}
else if (JXGMatch(dropget,event) && invent) {
	/* Drop an item from inventory
	 */
	char *name = StringFromString(iname);
	object *obj = lookup_object(name);

	if (!obj || obj->inv != pr)
		display_message("You have no %s",name);
	else
		client_request(fd,req_drop,obj->name);
	}
else if (JXGMatch(dropget,event) && !invent) {
	/* Get an item from the node player is at
	 */
	char *name = StringFromString(iname);
	object *obj = lookup_object(name);

	if (!obj || obj->pos != pr->pos)
		display_message("There is no %s here",name);
	else
		client_request(fd,req_get,obj->name);
	}
else if (JXGMatch(examine,event)) {
	/* Examine object named in item name gadget
	 */
	char *name = StringFromString(iname);
	object *obj = lookup_object(name);

	if (strlen(name) == 0)
		display_message("No object selected");
	else if (!obj || (obj->pos != pr->pos && obj->inv != pr))
		display_message("There is no %s here",name);
	else if (!obj->desc)
		display_message("%s : No description given");
	else
		display_message("%s : %s",obj->name,obj->desc);
	}
else if (JXGMatch(give,event) && !givewin) {
	/* Pop up window for giving an item to a person
	 */
	person *pp;
	int px,py;		/* pointer position */
	int i;

	get_pointer_pos(display,&px,&py);
	givewin = JXNewGadget(display,win,white,black,grey,window,px,py,200,
			      130,"Give item",True,True);

	/* Create gadgets in give window
	 */
	JXWindowAdd(display,givewin,white,black,grey,stext,5,5,190,20,a_left,
		    "Person to give to");
	giveper = JXWindowAdd(display,givewin,white,black,grey,combo,5,30,190,
			      20,-1,NULL);
	JXSetOption(display,giveper,LimitToList,True,0);
	JXWindowAdd(display,givewin,white,black,grey,stext,5,55,190,20,a_left,
		    "Item to give");
	giveobj = JXWindowAdd(display,givewin,white,black,grey,combo,5,80,190,
			      20,-1,NULL);
	giveok = JXWindowAdd(display,givewin,white,black,grey,boolean,5,105,90,
			     20,"Ok");
	givecan = JXWindowAdd(display,givewin,white,black,grey,boolean,105,105,
			      90,20,"Cancel");
	JXSetOption(display,givewin,HighlightCol,red,0);

	/* Fill persons list with nearby people
	 */
	reset_person();
	while((pp = next_person())) {
		if (close_node(pp->pos,pr->pos) && pp != pr) {
			JXAddMenuItem(display,giveper,pp->name);
			}
		}

	/* Fill objects list with stuff from inventory
	 */
	for(i=0; i<pr->invc; i++) {
		JXAddMenuItem(display,giveobj,pr->inv[i]);
		if (strequ(pr->inv[i],StringFromString(iname)))
			JXSetOption(display,giveobj,CurrentString,pr->inv[i],0);
		}
	}
else if (JXGMatch(doo,event) && !dowin) {
	/* Open do requestor
	 */
	char *oname = StringFromString(iname);

	open_do_requestor();
	if (lookup_object(oname))
		client_request(fd,req_actlist,oname);
	}
else if (JXGMatch(command,event) && !cmdwin) {
	/* Open command requestor
	 */
	int px,py;
	char cdirbuf[256];	/* current dir */
	person *pp;

	get_pointer_pos(display,&px,&py);
	cmdwin = JXNewGadget(display,win,white,black,grey,window,px,py,
			     200,250,"Command",True,True);
	fill_command_requestor(fd,pr);
	}
else if (JXGMatch(wizard,event)) {
	display_message("Not implemented yet");
	}
else if (JXGMatch(prefs,event) && !prefswin) {
	/* Open prefs dialog box
	 */
	userprefs cp;		/* current preferences */
	int px,py;

	get_prefs(&cp);
	if (cp.move_speed < 0) cp.move_speed = 0;
	if (cp.move_speed > 5) cp.move_speed = 5;

	get_pointer_pos(display,&px,&py);
	prefswin = JXNewGadget(display,win,white,black,grey,window,px,py,200,
			       400,"Sim prefs",True,True);

	JXWindowAdd(display,prefswin,white,black,grey,stext,5,5,190,20,a_left,
		    "Default 3d");
	prefsd3d = JXWindowAdd(display,prefswin,white,black,grey,combo,5,30,
			       190,20,cp.default_3d,"Worst","Intermediate",
			       "Best",NULL);
	JXSetOption(display,prefsd3d,LimitToList,True,0);

	JXWindowAdd(display,prefswin,white,black,grey,stext,5,60,190,20,a_left,
		    "Received data");
	prefsrecv = JXWindowAdd(display,prefswin,white,black,grey,combo,5,85,
				190,20,cp.recved_data,"Always save",
				"Save or display","Always display","Ignore",
				NULL);
	JXSetOption(display,prefsrecv,LimitToList,True,0);

	JXWindowAdd(display,prefswin,white,black,grey,stext,5,115,190,20,
		    a_left,"Movement speed");
	prefsms = JXWindowAdd(display,prefswin,white,black,grey,scroller,5,
			      140,190,20,cp.move_speed/5);
	JXSetOption(display,prefsms,SmoothMove,1,0);

	JXWindowAdd(display,prefswin,white,black,grey,stext,5,170,190,20,
		    a_left,"Server password");
	prefspass = JXWindowAdd(display,prefswin,white,black,grey,string,5,
				195,190,20,cp.password,256);

	JXWindowAdd(display,prefswin,white,black,grey,stext,25,230,190,20,
		    a_left,"Display transfer?");
	prefsdtrans = JXWindowAdd(display,prefswin,white,black,grey,toggle,
				  5,235,15,15,cp.display_transfer,"");

	JXWindowAdd(display,prefswin,white,black,grey,stext,5,265,190,20,
		    a_left,"Block size");
	prefstblock = JXWindowAdd(display,prefswin,white,black,grey,string,
				  5,290,190,20,itoa(cp.transfer_block),50);

	prefssave = JXWindowAdd(display,prefswin,white,black,grey,boolean,5,
				375,60,20,"Save");
	prefsuse  = JXWindowAdd(display,prefswin,white,black,grey,boolean,70,
				375,60,20,"Use");
	prefscan =  JXWindowAdd(display,prefswin,white,black,grey,boolean,135,
				375,60,20,"Cancel");
	JXSetOption(display,prefswin,HighlightCol,red,0);
	}
else if (JXGMatch(help,event) && !helpwin) {
	/* Open a help window
	 */
	int i;
	JXgadget *helplist;
	int px,py;

	get_pointer_pos(display,&px,&py);
	helpwin = JXNewGadget(display,win,white,black,grey,window,px,py,300,
			      500,"Sim help",True,True);
	helplist = JXWindowAdd(display,helpwin,white,black,grey,slist,5,5,290,
			       465,NULL);
	helpok  = JXWindowAdd(display,helpwin,white,black,grey,boolean,5,475,
			      90,20,"Ok");
	for(i=0; help_text[i]; i++)
		JXAddListText(display,helplist,help_text[i]);
	JXRedrawGadget(display,helplist);
	JXSetOption(display,helpwin,HighlightCol,red,0);
	}
else if (JXGMatch(mapp,event) && !mapwin) {
	/* Open map window
	 */
	int px,py;
	XColor curfore;

	get_pointer_pos(display,&px,&py);
	mapwin  = JXNewGadget(display,win,white,black,grey,window,px,py,300,
			      325,"Sim map",True,True);
	mappane = JXWindowAdd(display,mapwin,white,black,grey,scrollpane,5,5,
			      290,290,get_map_pixmap());
	mapclose= JXWindowAdd(display,mapwin,white,black,grey,boolean,5,300,
			      100,20,"Close");

	curfore.red = 0; curfore.green = 65535; curfore.blue = 0;
	JXSetOption(display,mappane,CursorColour,curfore,ReportMove,False,0);
	JXSetOption(display,mapwin,HighlightCol,red,0);
	}
else if (JXGMatch(quit,event)) {
	/* Exit simulation
	 */
	shutdown_3d();
	exit(0);
	}

/* Gadgets in give requestor
 */
else if (givewin && JXGMatch(giveok,event)) {
	/* Give command ok'd
	 */
	char *obname = StringFromCombo(giveobj);
	char *rprname = StringFromCombo(giveper);
	object *ob = lookup_object(obname);
	person *rpr = lookup_person(rprname);

	if (!ob || ob->inv != pr)
		display_message("Item %s not in inventory",obname);
	else if (!rpr || !close_node(rpr->pos,pr->pos))
		display_message("Person %s is not nearby",rprname);
	else
		client_request(fd,req_give,ob->name,rpr->name);
	JXFreeGadget(display,givewin);
	givewin = NULL;
	}
else if (givewin && JXGMatch(givecan,event)) {
	/* Give command cancelled
	 */
	JXFreeGadget(display,givewin);
	givewin = NULL;
	}

/* Gadgets in do requestor
 */
else if (dowin && JXGMatch(dook,event)) {
	/* Do action ok'd
	 */
	char *action = StringFromCombo(doaction);
	char *obname = StringFromString(doobject);
	char *params = StringFromString(doparams);
	object *ob = lookup_object(obname);

	if (!ob || (ob->inv != pr && ob->pos != pr->pos))
		display_message("Object %s not found",obname);
	else {
		char **parv;
		int parc;

		parse_string(params,&parv,&parc);
		client_request(fd,req_do,action,ob->name,parv,parc);
		}
	JXFreeGadget(display,dowin);
	dowin = NULL;
	}
else if (dowin && JXGMatch(docan,event)) {
	/* Do requestor cancelled
	 */
	JXFreeGadget(display,dowin);
	dowin = NULL;
	}

/* Gadgets in help requestor
 */
else if (helpwin && JXGMatch(helpok,event)) {
	/* Help requestor closed
	 */
	JXFreeGadget(display,helpwin);
	helpwin = NULL;
	}

/* Gadgets in command requestor
 */
else if (cmdwin && JXGMatch(cmdcombo,event)) {
	/* Command mode changed
	 */
	cmdmode = ChoiceFromCombo(cmdcombo);
	fill_command_requestor(fd,pr);
	}
else if (cmdwin && JXGMatch(cmdclose,event)) {
	/* Command requestor closed
	 */
	close_command_requestor();
	}
else if (cmdwin && cmdmode == 0 && JXGMatch(cmdsend,event)) {
	/* Send data
	 */
	client_request(fd,req_send,StringFromCombo(cmddata),
		       StringFromCombo(cmdperson));
	close_command_requestor();
	}
else if (cmdwin && cmdmode == 1 && JXGMatch(cmdupload,event)) {
	/* Upload a file
	 */
	char *ufile = StringFromString(cmdfile);
	struct stat stbuf;
	int port = get_sprefs(NULL)->port + alloc_port();

	if (stat(ufile,&stbuf) == -1 || (stbuf.st_mode & S_IFMT) == S_IFDIR)
		display_message("Couldn't open file %s",ufile);
	else {
		/* Upload with status.
		 */
		int fsize = stbuf.st_size;

		client_request(fd,req_upload,basenam(ufile),fsize,port);
		if (!sfork()) {
			Display *tdisplay;
			int f, basefd, connfd;
			struct in_addr addr;

			set_signals(SIG_DFL);
			reset_fds();
			if ((f = open(ufile,O_RDONLY)) == -1) {
				logerror("Couldn't open %s for upload",ufile);
				exit(1);
				}
			if ((tdisplay = XOpenDisplay(NULL)) == NULL) {
				logerror("Failed to open display for upload "
					 "status");
				exit(1);
				}
			JXStartup(tdisplay,screen_num);
			basefd = open_tcp_dest(port);
			connfd = new_connection(basefd,NULL,&addr);
			copy_with_status("Uploading data to server",tdisplay,
					 f,connfd,fsize);
			close(f); close(connfd); close(basefd);
			exit(0);
			}
		close_command_requestor();
		}
	}
else if (cmdwin && (cmdmode == 1 || cmdmode == 3) && JXGMatch(cmddir,event) &&
	 !cmdfilelist) {
	/* Open browser for file to upload / use as avatar
	 */
	char *filen = StringFromString(cmdfile);
	char *base, *dir;

	if (JXIsDirectory(filen)) {
		dir = filen;
		base = "";
		}
	else {
		dir = dirnam(filen);
		base = basenam(filen);
		}
	cmdfilelist = JXNewGadget(display,win,white,black,grey,filewin,100,100,
				  300,400,base,dir,"Ok");
	JXSetOption(display,cmdfilelist,HighlightCol,red,0);
	}
else if (cmdwin && cmdfilelist && JXGMatch(cmdfilelist,event)) {
	/* Ok/Cancel clicked in upload file browser
	 */
	if (DataFromEvent(event) == ClickEvent) {
		char *filen = FilenameStringFromFile(cmdfilelist);
		char *filep = FilepathStringFromFile(cmdfilelist);
		char fullpath[256];

		sprintf(fullpath,"%s/%s",filep,filen);
		JXSetOption(display,cmdfile,CurrentString,fullpath,0);
		}
	JXFreeGadget(display,cmdfilelist);
	cmdfilelist = NULL;
	}
else if (cmdwin && cmdmode == 2 && JXGMatch(cmdrm,event)) {
	/* Remove a data file on the server
	 */
	client_request(fd,req_rm,StringFromCombo(cmddata));
	close_command_requestor();
	}
else if (cmdwin && cmdmode == 3 && JXGMatch(cmdavset,event)) {
	/* Set avatar clicked in command requestor
	 */
	int afd;
	char *afile = StringFromString(cmdfile);

	if ((afd = open(afile,O_RDONLY)) == -1)
		display_message("Avatar file %s not found",afile);
	else {
		int cppafd = pipethrough(afd,NULL,CPPCMD);

		client_request(fd,req_avatar);
		copy_to_eof(cppafd,fd);
		fdprintf(fd,"\nend\n");
		pipeclose(cppafd);
		close(afd);
		}
	close_command_requestor();
	}
else if (cmdwin && cmdmode == 4 && JXGMatch(cmdpdset,event)) {
	/* Set new player details
	 */
	client_request(fd,req_fullname,StringFromString(cmdfullname));
	client_request(fd,req_email,StringFromString(cmdemail));
	close_command_requestor();
	}
else if (cmdwin && cmdmode == 5 && JXGMatch(cmdsuicide,event)) {
	/* Suicide clicked in command requestor
	 */
	client_request(fd,req_suicide);
	close_command_requestor();
	}

/* Gadgets in prefs requestor
 */
else if (prefswin && (JXGMatch(prefssave,event) || JXGMatch(prefsuse,event))) {
	/* Save or use clicked in prefs window
	 */
	userprefs np;
	char *newpass = StringFromString(prefspass);
	int old_3d;

	get_prefs(&np);
	old_3d = np.default_3d;
	np.default_3d  = ChoiceFromCombo(prefsd3d);
	np.recved_data = ChoiceFromCombo(prefsrecv);
	np.move_speed  = NPosFromScroller(prefsms)*5;
	strcpy(np.password,newpass);
	client_request(fd,req_setpass,newpass);
	np.display_transfer = StateFromToggle(prefsdtrans);
	np.transfer_block   = atoi(StringFromString(prefstblock));
	put_prefs(&np);
	if (JXGMatch(prefssave,event)) {
		if (save_prefs() == 0)
			display_message("Couldn't write prefs");
		}
	JXFreeGadget(display,prefswin);
	prefswin = NULL;
	if (np.default_3d != old_3d)
		new_3d_level();
	}
else if (prefswin && JXGMatch(prefscan,event)) {
	/* Cancel clicked in prefs window
	 */
	JXFreeGadget(display,prefswin);
	prefswin = NULL;
	}

/* Gadgets in map requestor
 */
else if (mapwin && JXGMatch(mapclose,event)) {
	/* Map requestor closed
	 */
	JXFreeGadget(display,mapwin);
	mapwin = NULL;
	}
}


static double flen;		/* time the last frame took to draw */

/* draw_frame
 * Draw the current 3d view.
 */
void draw_frame(player *p, plview *plv)
{
double tbefore = current_time();

if (render_3d(plv)) {
	char *headn[8] = {"north","northeast","east","southeast","south",
			  "southwest","west","northwest" };
	int hn;

	/* The 3d view was redrawn, so redraw the map as well
	 */
	create_player_map(display,plv);
	if (mapwin)
		JXRedrawGadget(display,PaneFromScrollpane(mappane));

	/* Change the compass string
	 */
	hn = fixang(plv->ang + 22.5 + 90) / 45;
	JXSetOption(display,compass,CurrentString,headn[hn],0);
	}
flen = current_time() - tbefore;
}


/* last_frame_len
 * Returns the time the last frame took to draw
 */
double last_frame_len(void)
{
return flen;
}


/* display_message
 * Add a message to the user interface's messages list
 */
void display_message(char *fmt, ...)
{
char buf[256];
va_list ap;

va_start(ap,fmt);
vsprintf(buf,fmt,ap);
va_end(ap);
JXAddListText(display,mlist,buf);
JXSetOption(display,mlist,ListPos,1.0,0);
JXRedrawGadget(display,mlist);
}


/* display_messagebox
 * Display a message in a small popup window
 */
void display_messagebox(char *fmt, ...)
{
char buf[1024];
va_list ap;
char *line[30];
int lc = 0;
JXgadget *mbox;
int px,py;

/* Create lines of text
 */
va_start(ap,fmt);
vsprintf(buf,fmt,ap);
va_end(ap);
line[lc] = strtok(buf,"\n");
while((line[++lc] = strtok(NULL,"\n")))
	;

get_pointer_pos(display,&px,&py);
mbox = JXNewGadget(display,win,white,black,grey,message,px,py,line,lc);
JXSetOption(display,mbox,HighlightCol,red,0);
}


/* fill_here_list
 * If the items list is displaying items at the player's location, re-fill it.
 */
void fill_here_list(person *pr)
{
if (!invent)
	fill_hereinv_list(pr);
}


/* fill_inventory_list
 * If the items list is in inventory mode, re-fill it.
 */
void fill_inventory_list(person *pr)
{
if (invent)
	fill_hereinv_list(pr);
}


/* fill_items_list
 * Clear out the items list and fill it with either items at this
 * location, or item's in this person's inventory, depending on the current
 * mode.
 */
void fill_hereinv_list(person *pr)
{
JXClearListText(display,ilist);
if (invent) {
	/* Fill from inventory
	 */
	int i;

	for(i=0; i<pr->invc; i++)
		JXAddListText(display,ilist,pr->inv[i]);
	}
else {
	/* Fill from objects at this node
	 */
	object *ob;

	reset_object();
	while((ob = next_object()))
		if (ob->pos == pr->pos)
			JXAddListText(display,ilist,ob->name);
	}
JXRedrawGadget(display,ilist);
}


/* open_do_requestor
 * Open do requestor and populate with gadgets
 */
static void open_do_requestor(void)
{
int px,py;

get_pointer_pos(display,&px,&py);
dowin =     JXNewGadget(display,win,white,black,grey,window,px,py,200,185,
			"Do",True,True);

JXWindowAdd(display,dowin,white,black,grey,stext,5,5,190,20,a_left,
	    "Action to do");
doaction =  JXWindowAdd(display,dowin,white,black,grey,combo,5,30,190,20,
			-1,NULL);
JXSetOption(display,doaction,LimitToList,True,0);
JXWindowAdd(display,dowin,white,black,grey,stext,5,55,190,20,a_left,"Object");
doobject =  JXWindowAdd(display,dowin,white,black,grey,string,5,80,190,20,
			StringFromString(iname),256);
JXWindowAdd(display,dowin,white,black,grey,stext,5,105,190,20,a_left,
	    "Parameters");
doparams =  JXWindowAdd(display,dowin,white,black,grey,string,5,130,190,20,
			"",256);
dook =      JXWindowAdd(display,dowin,white,black,grey,boolean,5,155,90,20,
			"Ok");
docan =     JXWindowAdd(display,dowin,white,black,grey,boolean,105,155,90,20,
			"Cancel");
JXSetOption(display,dowin,HighlightCol,red,0);
}


/* fill_do_requestor
 * When the do requestor is opened, a request is send to the server asking
 * for possible actions for that object. This function gets called when
 * the server replies.
 */
void fill_do_requestor(char *oname, char *acts[], int actc)
{
if (strequ(oname,StringFromString(doobject))) {
	int i;

	for(i=0; i<actc; i++)
		JXAddMenuItem(display,doaction,acts[i]);
	if (actc != 0)
		JXSetOption(display,doaction,CurrentString,acts[0],0);
	}
}


/* get_pointer_pos
 * Returns the current pointer position, relative to the root window
 */
static void get_pointer_pos(Display *d, int *px, int *py)
{
Window proot,pchild;
int pwx,pwy;
unsigned int pmask;

XQueryPointer(d,win,&proot,&pchild,px,py,&pwx,&pwy,&pmask);
}


/* fill_send_requestor
 * Called when the server replies with a list of data files available to
 * this player, from a req_ls generated by the send data dialog box code.
 */
void fill_send_requestor(char **dats, int datc)
{
int i;

if (cmdwin && (cmdmode == 0 || cmdmode == 2)) {
	for(i=0; i<datc; i++)
		JXAddMenuItem(display,cmddata,dats[i]);
	if (datc != 0)
		JXSetOption(display,cmddata,CurrentString,dats[0],0);
	}
}


/* close_command_requestor
 * Close and free command requestor
 */
static void close_command_requestor(void)
{
JXFreeGadget(display,cmdfilelist);
cmdfilelist = NULL;
JXFreeGadget(display,cmdwin);
cmdwin = NULL;
}


/* resize_gadgets
 * Reposition ui gadgets after a window resize
 */
static void resize_gadgets(void)
{
userprefs *cp = get_prefs(NULL);
int w = cp->w, h = cp->h, lw = (w-100)/2;

/* Scroll bars and movement buttons
 */
JXMoveGadget(display,horiz,0,h-280,w,20);
JXMoveGadget(display,vert,w-20,0,20,h-280);
JXMoveGadget(display,left,0,h-260,w/5,20);
JXMoveGadget(display,forward,w/5,h-260,w/5,20);
JXMoveGadget(display,compass,(w/5)*2,h-260,w/5,20);
JXMoveGadget(display,look,(w/5)*3,h-260,w/5,20);
JXMoveGadget(display,right,(w/5)*4,h-260,w/5,20);

/* Lists
 */
JXMoveGadget(display,mlist,0,h-240,lw,220);
JXMoveGadget(display,ilist,lw,h-240,lw,200);
JXMoveGadget(display,iname,lw,h-40,lw,20);

/* Gadgets along right side
 */
JXMoveGadget(display,hereinv,w-100,h-240,100,20);
JXMoveGadget(display,dropget,w-100,h-220,100,20);
JXMoveGadget(display,examine,w-100,h-200,100,20);
JXMoveGadget(display,give,w-100,h-180,100,20);
JXMoveGadget(display,doo,w-100,h-160,100,20);
JXMoveGadget(display,command,w-100,h-140,100,20);
JXMoveGadget(display,wizard,w-100,h-120,100,20);
JXMoveGadget(display,prefs,w-100,h-100,100,20);
JXMoveGadget(display,help,w-100,h-80,100,20);
JXMoveGadget(display,mapp,w-100,h-60,100,20);
JXMoveGadget(display,quit,w-100,h-40,100,20);

/* Message gadgets
 */
JXMoveGadget(display,mesg,0,h-20,w-100,20);
JXMoveGadget(display,saymsg,w-100,h-20,50,20);
JXMoveGadget(display,yellmsg,w-50,h-20,50,20);
}


/* handle_key_press
 * Handle a keypress event
 */
static void handle_key_press(XEvent *event, int fd, player *pl, person *pr,
			     plview *plv)
{
KeySym key = XLookupKeysym(&event->xkey,0);

switch(key) {
case XK_Left:
case XK_Right: {
	/* Turn left and right
	 */
	if (key == XK_Left) plv->ang = fixang(plv->ang - 20);
	else                plv->ang = fixang(plv->ang + 20);
	JXSetOption(display,horiz,SliderNPos,plv->ang/360.0,0);
	JXRedrawGadget(display,horiz);
	client_request(fd,req_turn,plv->ang);
	need_redraw();
	break;
	}

case XK_Up:
case XK_Down: {
	/* Tilt up and down
	 */
	if (key == XK_Up) plv->tilt = plv->tilt + 20;
	else              plv->tilt = plv->tilt - 20;
	if (plv->tilt > 89)  plv->tilt = 89;
	if (plv->tilt < -89) plv->tilt = -89;
	JXSetOption(display,vert,SliderNPos,0.5-(plv->tilt/179.0),0);
	JXRedrawGadget(display,vert);
	need_redraw();
	break;
	}

case XK_a: {
	move_along(fd,pr,plv,1);
	break;
	}

case XK_z: {
	move_along(fd,pr,plv,0);
	break;
	}

	} /* end switch */
}


/* fill_command_requestor
 * Create gadgets in the command requestor, depending on the position of the
 * command mode combo box (a la netscape prefs)
 */
static void fill_command_requestor(int fd, person *pr)
{
char cdirbuf[256];
person *pp;

/* Clear window gadgets
 */
JXWindowClear(display,cmdwin);
if (cmdfilelist) {
	JXFreeGadget(display,cmdfilelist);
	cmdfilelist = NULL;
	}

/* Create gadgets that are always there
 */
cmdcombo = JXWindowAdd(display,cmdwin,white,black,grey,combo,20,10,160,20,
		       cmdmode,"Send to player","Upload to server",
		       "Remove data","Set avatar","Player details",
		       "Destroy player",NULL);
JXSetOption(display,cmdcombo,LimitToList,True,0);
cmdclose = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,225,100,20,
		       "Close");
JXWindowAdd(display,cmdwin,white,black,grey,bevelbox,2,40,196,3,1);

/* Create mode-specific gadgets
 */
switch(cmdmode) {
case 0:
	/* Create send gadgets
	 */
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,50,190,20,a_left,
		    "Data to send");
	cmddata = JXWindowAdd(display,cmdwin,white,black,grey,combo,5,75,
			      190,20,-1,NULL);
	JXSetOption(display,cmddata,LimitToList,True,0);
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,100,190,20,a_left,
		    "Person to send to");
	cmdperson = JXWindowAdd(display,cmdwin,white,black,grey,combo,5,125,
				190,20,-1,NULL);
	JXSetOption(display,cmdperson,LimitToList,True,0);
	cmdsend = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,150,
			      190,20,"Send");
	reset_person();
	while((pp = next_person())) {
		if (close_node(pp->pos,pr->pos)) {
			JXAddMenuItem(display,cmdperson,pp->name);
			}
		}
	client_request(fd,req_ls);
	break;

case 1:
	/* Create upload gadgets
	 */
	getcwd(cdirbuf,256);
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,50,190,20,a_left,
		    "File to upload");
	cmdfile = JXWindowAdd(display,cmdwin,white,black,grey,string,5,75,
			      160,20,cdirbuf,255);
	JXSetOption(display,cmdfile,StringAlign,a_right,0);
	cmddir    = JXWindowAdd(display,cmdwin,white,black,grey,boolean,165,
				75,30,20,"Dir");
	cmdupload = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,100,
				190,20,"Upload");
	break;

case 2:
	/* Create rm gadgets
	 */
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,50,190,20,a_left,
		    "Data to remove");
	cmddata = JXWindowAdd(display,cmdwin,white,black,grey,combo,5,75,
			      190,20,-1,NULL);
	JXSetOption(display,cmddata,LimitToList,True,0);
	cmdrm = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,100,
			    190,20,"Remove");
	client_request(fd,req_ls);
	break;

case 3:
	/* Create change avatar gadgets
	 */
	getcwd(cdirbuf,256);
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,50,190,20,a_left,
		    "File for avatar");
	cmdfile = JXWindowAdd(display,cmdwin,white,black,grey,string,5,75,
			      160,20,cdirbuf,255);
	JXSetOption(display,cmdfile,StringAlign,a_right,0);
	cmddir    = JXWindowAdd(display,cmdwin,white,black,grey,boolean,165,
				75,30,20,"Dir");
	cmdavset = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,100,
			       190,20,"New avatar");
	client_request(fd,req_ls);
	break;

case 4:
	/* Create player details gadgets
	 */
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,50,190,20,a_left,
		    "Full name");
	cmdfullname = JXWindowAdd(display,cmdwin,white,black,grey,string,5,75,
				  190,20,pr->fullname ? pr->fullname : "",256);
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,100,190,20,a_left,
		    "Email address");
	cmdemail = JXWindowAdd(display,cmdwin,white,black,grey,string,5,125,
			       190,20,pr->email ? pr->email : "",256);
	cmdpdset = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,150,
			       190,20,"Set details");
	break;

case 5:
	/* Create suicide gadgets
	 */
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,50,190,20,a_left,
		    "Warning!");
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,75,190,20,a_left,
		    "Clicking this button will ");
	JXWindowAdd(display,cmdwin,white,black,grey,stext,5,95,190,20,a_left,
		    "destroy your player.");
	cmdsuicide = JXWindowAdd(display,cmdwin,white,black,grey,boolean,5,120,
				 190,20,"Destroy");
	break;

	} /* end switch */
JXSetOption(display,cmdwin,HighlightCol,red,0);
}


/* normal_cursor
 * Use normal arrow cursor in ui
 */
void normal_cursor(void)
{
XDefineCursor(display,win,arrow_cur);
XFlush(display);
}


/* waiting_cursor
 * Use clock cursor in ui
 */
void waiting_cursor(void)
{
XDefineCursor(display,win,clock_cur);
XFlush(display);
}


/* download_data
 * Ask the user what to do with downloaded data, download it and then display,
 * all in a separate process.
 */
void download_data(char *mtype, int port, int len)
{
if (!sfork()) {
	Display *tdisplay;
	userprefs *cp = get_prefs(NULL);
	int recved_data = cp->recved_data;
	char filename[256];

	reset_fds();
	set_signals(SIG_DFL);
	if ((tdisplay = XOpenDisplay(NULL)) == NULL) {
		logerror("Failed to open display for transfer requestor?");
		exit(1);
		}
	JXStartup(tdisplay,screen_num);
	if (recved_data == 1) {
		/* Ask whether to save or display
		 */
		choose_download_mode(tdisplay,mtype,&recved_data);
		}
	if (recved_data == 0) {
		/* Ask for a filename
		 */
		choose_download_filename(tdisplay,&recved_data,filename);
		}
	if (recved_data != 3) {
		/* Download the data and display progress
		 */
		char datafile[256];
		int tfd, sfd;
		char *host = get_sprefs(NULL)->server;

		/* Choose a filename
		 */
		if (recved_data == 0) strcpy(datafile,filename);
		else {
			if (strchr(mtype,'/'))
				sprintf(datafile,"%s.%s",tmpnam(NULL),
							 strchr(mtype,'/')+1);
			else
				strcpy(datafile,tmpnam(NULL));
			}
		if ((tfd = creat(datafile,0700)) == -1) {
			logerror("Couldn't create %s file %s",
				 recved_data ? "temp" : "data",datafile);
			exit(1);
			}

		/* Connect to server and transfer data
		 */
		if ((sfd = open_tcp_send(host,port)) == -1) {
			logerror("Couldn't connect to server on %s",host);
			exit(1);
			}
		if (!copy_with_status("Receiving data from server",tdisplay,
				      sfd,tfd,len)) {
			logmessage("Download incomplete");
			unlink(datafile);
			}
		else if (recved_data == 2) {
			display_mime_data(datafile,mtype,len);
			unlink(datafile);
			}
		close(sfd);
		close(tfd);
		}
	exit(0);
	}
}


/* choose_download_mode
 * Ask the user if they want to save or display incoming data
 */
static void choose_download_mode(Display *tdisplay, char *mtype, int *rd)
{
JXgadget *dswin, *dssave, *dsdisplay;
int px,py;

/* Create requestor
 */
get_pointer_pos(tdisplay,&px,&py);
dswin = JXNewGadget(tdisplay,win,white,black,grey,window,px,py,200,90,
		    "Received data",True,True);
JXWindowAdd(tdisplay,dswin,white,black,grey,stext,5,5,190,15,a_left,
	    "Data of type");
JXWindowAdd(tdisplay,dswin,red,black,grey,stext,5,20,190,15,a_left,mtype);
JXWindowAdd(tdisplay,dswin,white,black,grey,stext,5,35,190,15,a_left,
	    "has arrived");
dssave = JXWindowAdd(tdisplay,dswin,white,black,grey,boolean,5,65,90,20,"Save");
dsdisplay = JXWindowAdd(tdisplay,dswin,white,black,grey,boolean,105,65,90,20,
			"Display");
JXSetOption(tdisplay,dswin,HighlightCol,red,TopLevel,True,0);

/* Wait for choice
 */
while(1) {
	XEvent event;
	XNextEvent(tdisplay,&event);
	if (JXGadgetEvent(tdisplay,&event) || event.type != ClientMessage)
		continue;
	if (JXGMatch(dssave,&event)) {
		*rd = 0;
		break;
		}
	if (JXGMatch(dsdisplay,&event)) {
		*rd = 2;
		break;
		}
	}
JXFreeGadget(tdisplay,dswin);
}


/* choose_download_filename
 * Ask the user for a name to save a downloaded file as
 */
static void choose_download_filename(Display *tdisplay, int *rd, char *fn)
{
JXgadget *filewind, *filereq;
char cwd[256];

/* Create window and file requestor
 */
filewind = JXNewGadget(tdisplay,win,white,black,grey,window,100,100,300,400,
	 	       "Save file",True,False);
getcwd(cwd,256);
filereq = JXWindowAdd(tdisplay,filewind,white,black,grey,file,0,0,300,400,
		      "",cwd,"Save");
JXSetOption(tdisplay,filewind,HighlightCol,red,TopLevel,True,0);

/* Wait for choice
 */
while(1) {
	XEvent event;
	XNextEvent(tdisplay,&event);
	if (JXGadgetEvent(tdisplay,&event) || event.type != ClientMessage)
		continue;
	if (JXGMatch(filereq,&event)) {
		if (DataFromEvent(&event) == ClickEvent)
			sprintf(fn,"%s/%s",FilepathStringFromFile(filereq),
					   FilenameStringFromFile(filereq));
		else
			*rd = 3;
		break;
		}
	}
JXFreeGadget(tdisplay,filewind);
}


/* copy_with_status
 * Copy data from one fd to another, while displaying a nice status bar for
 * large transfers.
 */
static int copy_with_status(char *message, Display *tdisplay, int fd1,
			    int fd2, int len)
{
JXgadget *statwin, *statbar, *statcan;
int px,py;
int i=0;
userprefs *cp = get_prefs(NULL);
char *buf = smalloc(cp->transfer_block);
int percent, lastpercent = -1;

/* Small transfer
 */
if (len < BUFSIZ || !cp->display_transfer)
	return copy_data(fd1,fd2,len);

/* Create window and status gauge
 */
get_pointer_pos(tdisplay,&px,&py);
statwin = JXNewGadget(tdisplay,win,white,black,grey,window,px,py,300,110,
		      "Data transfer",True,True);
JXWindowAdd(tdisplay,statwin,white,black,grey,stext,5,5,290,20,a_centre,
	    message);
JXWindowAdd(tdisplay,statwin,white,black,grey,stext,5,35,290,20,a_centre,
	    "Transfer progress");
statbar = JXWindowAdd(tdisplay,statwin,white,black,grey,statusgauge,
		      5,60,290,20,0.0);
statcan = JXWindowAdd(tdisplay,statwin,white,black,grey,boolean,195,85,100,20,
		      "Cancel");
JXSetOption(tdisplay,statwin,HighlightCol,red,TopLevel,True,0);

/* Begin transfer, updating gauge
 */
while(i<len) {
	fd_set fds;
	int xfd = ConnectionNumber(tdisplay);

	FD_ZERO(&fds);
	FD_SET(fd1,&fds);
	FD_SET(xfd,&fds);
	select(max(fd1,xfd)+1,&fds,NULL,NULL,NULL);
	if (FD_ISSET(fd1,&fds)) {
		/* Data to transfer
		 */
		int n = (len-i < cp->transfer_block ? len-i
						    : cp->transfer_block);

		if (fdread(fd1,buf,1,n) == -1) {
			logerror("Error reading data");
			break;
			}
		fdwrite(fd2,buf,1,n);
		i += n;
		percent = i*100/len;
		if (percent != lastpercent) {
			JXSetOption(tdisplay,statbar,PropFrac,
				    (double)i/(double)len,0);
			lastpercent = percent;
			}
		}
	if (FD_ISSET(xfd,&fds)) {
		/* X event
		 */
		XEvent event;
		XNextEvent(tdisplay,&event);
		if (!JXGadgetEvent(tdisplay,&event) &&
		    event.type == ClientMessage) {
			if (JXGMatch(statcan,&event)) 
				break;	/* download cancelled */
			}
		}
	}
free(buf);
JXFreeGadget(tdisplay,statwin);
return i==len;
}


